!! This module implements the VERIFY function to search from any given position in the parent
!! string.  It is capable of searching forwards or backwards.  It complements the existing VERIFY
!! built-in function.
!! Usage:  verify (String, Sub, Position)    for a forward search, commening at position <Position>;
!!    verify (String, Sub, Position, .false. )  for a forward search, commening at position <Position>; &
!!    verify (String, Sub, position, .true. )   for a backward search, commencing from position <Position>.
module VERIFY_MODULE
   implicit none
   public :: verify, VERIFY_sub, VERIFYR_sub, VERIFY_control
   intrinsic  verify

   interface  verify
      module  procedure  VERIFY_sub
      module  procedure  VERIFY_control
   end  interface

contains

   !! This function deals with the 4-argument version, calling VERIFY_sub or VERIFYR_sub
   !! according to the value of the fourth argument Direction being .false. or .true. respectively.
   function VERIFY_control (String, Sub, Position, Direction) result (Pos)
      !! INCOMING: String    = the string to be searched;
      !!           Sub       = contains characters to look for;
      !!           Position  = where to start the search (measured from the left-hand end of
      !!                       String).
      !!           Direction = .false. for a forwards search; .true. for a backwards search.
      character (len=*), intent (in)   :: String, Sub
      integer, intent (in)         :: Position
      logical, intent (in)         :: Direction

      integer                      :: Pos

      if (Direction) then                         !! Do a backwards search.
         Pos = VERIFYR_sub (String, Sub, Position)
      else                                        !! Do a forwards search.
         Pos = VERIFY_sub (String, Sub, Position)
      end  if
   end  function  VERIFY_control

   !! This function searches the first string String for any characters that do
   !! not exist in second string Sub.  If there are any such characters, the function returns
   !! the position of the left-most.
   !! The search is performed from left to right, commencing from character position Position.
   function  VERIFY_sub (String, Sub, Position) result (Pos)
      !! INCOMING: String   = the string to be searched;
      !!           Sub      = contains characters to look for;
      !!           Position = where to start the search (measured from the left-hand end of
      !!                      String).
      character (len=*), intent (in) :: String, Sub
      integer, intent (in)           :: Position

      integer                        :: Pos

      logical, dimension (0:255)     :: Table
      integer                        :: J, K

      if (len (String) == 0) then                 !! There's nothing to search.
         Pos = 0
         return
      end  if
      if (len (Sub) == 0) then                    !! Nothing with which to search.
         Pos = Position
         return
      end  if
      if (len (Sub) == 1) then                    !! Looking for one character is fast.
         do J = Position, len(String)
            if (String(J:J) /= Sub) then
               Pos = J
               return
            end  if
         end  do
         Pos = 0                                  !! Unsuccessful search.
         return
      end  if

      !! SETS UP A LOOK-UP TABLE (which is independent of character set).
      Table = .true.                              !! All entries are true.
      do J = 1, len (Sub)
         K = ichar (Sub(J:J))
         Table (K) = .false.                      !! Table(k) is false for each character in Sub.
      end  do

      !! CONDUCT THE SEARCH (forwards).
      do J = Position, len(String)
         K = ichar (String(J:J))
         if (Table(K)) then                       !! True when a Sub character does not match one
            Pos = J                               !! in String.
            return
         end  if
      end  do

      Pos = 0                                     !! Unsuccessful search.

   end  function  VERIFY_sub

   !! This function procedure searches the first string String for any characters that do
   !! not exist in second string Sub.  If there are any such characters, the function returns
   !! the position of the right-most.
   !! The search is performed from right to left, commencing from character position Position.
   function  VERIFYR_sub (String, Sub, Position) result (Pos)
      !! INCOMING: String   = the string to be searched;
      !!           Sub      = contains characters to look for;
      !!           Position = where to start the search (measured from the left-hand end of
      !!                      String).
      character (len=*), intent (in) :: String, Sub
      integer, intent (in)           :: Position
      integer                        :: Pos

      logical, dimension (0:255)     :: Table
      integer                        :: J, K

      if (len (String) == 0) then                 !! There's nothing to search.
         Pos = 0
         return
      end  if
      if (len (Sub) == 0) then                    !! Nothing with which to search.
         Pos = Position
         return
      end  if
      if (len (Sub) == 1) then                    !! Looking for one character is fast.
         do J = Position, 1, -1
            if (String(J:J) /= Sub) then
               Pos = J
               return
            end  if
         end  do
         Pos = 0                                  !! Unsuccessful search.
         return
      end  if

      !! SETS UP A LOOK-UP TABLE (which is independent of character set).
      Table = .true.                              !! All entries are true.
      do J = 1, len (Sub)
         K = ichar (Sub(J:J))
         Table (K) = .false.                      !! Table(k) is false for each character in Sub.
      end  do

      !! CONDUCT THE SEARCH (backwards).
      do J = Position, 1, -1
         K = ichar (String(J:J))
         if (Table(K)) then                       !! True when a Sub character does not match one
            Pos = J                               !! in String.
            return
         end  if
      end  do

      Pos = 0                                     !! Unsuccessful search.

   end  function  VERIFYR_sub

end  module  VERIFY_MODULE
