module  Define_Type

   type, public                    :: Personage
      character (len=20)           :: Name
      character (len=20)           :: Street
      character (len=20)           :: Suburb
      character (len=4)            :: Postcode
   end  type  Personage

end  module  Define_Type

module  GROUP
   use Define_Type
   implicit  none
   private
   public                          :: SUBURB_SORT, NAME_SORT

contains

   !! This subroutine sorts an array by sorting continually changing subgroups.
   !! Elements between Person(M) and Person(N) are persons who live in the same suburb, and are
   !! sorted into order, based on NAME.

   subroutine NAME_SORT (Person, M, N)
      !! INCOMING:   Person = an array whose elements Person(M) to Person(N) are to be sorted.
      !!       M    = lower subscript of array PERSON;
      !!       N    = upper subscript of array PERSON.
      !! OUTGOING:   Person = the sorted array.  Elements Person(M) to Person(N) are in
      !!               ascending order.

      type (Personage), dimension ( : ), intent (in out) :: Person
      integer, intent (in)         :: M, N

      type (Personage)             :: Temp
      integer                      :: J, K, Gap

      Gap = N/2                                   !! The initial gap is half the number of elements.
      do
         if (Gap < 1) then                        !! We are done after the gap is 1, when all
            exit                                  !! elements are examined in the next loop:
         end  if
         !! Perform a Ranking Sort on selected elements.
         do  K = M, N-Gap
            if (Person(K)%Name > Person(K+Gap)%Name) then
                                                  !! Select elements every Gap positions apart . .
               Temp = Person(K+Gap)
               do  J = K, 1, -Gap
                  if (Person(J)%Name < Temp%Name) then
                     exit
                  end  if
                  Person(J+Gap) = Person(J)       !! Leapfrog by Gap locations.
               end  do
               Person(J+Gap) = Temp
            end  if
         end  do
         Gap = Gap/2                              !! Halve the gap for the next Ranking Sort.
      end  do
   end  subroutine  NAME_SORT

   !! This subroutine sorts an array by sorting continually-changing sub-groups.

   subroutine SUBURB_SORT (Person)
      !! INCOMING:   Person = an array whose elements are to be sorted.
      !! OUTGOING:   Person = the sorted array.  Elements are in ascending order.

      type (Personage), dimension ( : ), intent (in out) :: Person

      type (Personage)             :: Temp
      integer                      :: J, K, N, Gap

      N = ubound(Person, 1)                       !! The number of elements in the array Person.

      Gap = N/2                                   !! The initial gap is half the number of elements.
      do
         if (Gap < 1) then                        !! We are done after the gap is 1, when all
            exit                                  !! elements are examined in the next loop:
         end  if
         !! Perform a Ranking Sort on selected elements.
         do  K = 1, N-Gap
            if (Person(K)%Suburb > Person(K+Gap)%Suburb) then
                                                  !! Select elements every Gap positions apart . .
               Temp = Person(K+Gap)
               do  J = K, 1, -Gap
                  if (Person(J)%Suburb < Temp%Suburb) then
                     exit
                  end  if
                  Person(J+Gap) = Person(J)       !! Leapfrog by Gap locations.
               end  do
               Person(J+Gap) = Temp
            end  if
         end  do
         Gap = Gap/2                              !! Halve the gap for the next Ranking Sort.
      end  do
   end  subroutine  SUBURB_SORT

end  module  GROUP

program  SORT
   use  Define_Type
   use  GROUP
   implicit  none
   integer                         :: N

   type (Personage), dimension ( : ), allocatable :: Person
   integer                         :: Start, Finish
   integer                         :: J

   print *, "How many records are there to be sorted?"
   read *, N

   allocate (Person(N))

   do J = 1, N
      print *, "Please type a person's Name, Street, Suburb, and Postcode as strings:"
      read *, Person(J)
   end  do

   call SUBURB_SORT (Person)

   Start = 1
   do
      if (Start >= N) then
         exit
      end  if
      !! Find the start and finish of the names of persons living in the same suburb.
      do Finish = Start, N
         if (Person(Start)%Suburb /= Person(Finish)%Suburb) then
            exit
         end  if
      end  do
      Finish = Finish - 1                         !! Because we went too far.
      !! At this point, the subscripted variables Name(Start) to Name(Finish) contain the same
      !! suburb.
      call NAME_SORT (Person, Start, Finish)      !! to sort those records into name-order.
      Start = Finish+1                            !! To point at the next suburb.
   end  do

   print *, "The sorted data are:"
   do J = 1, N
      print *, Person(J)
   end  do

end  program  SORT

!! Figure 1.13  The main program for sorting structures
