module GROUP
   implicit  none
   public :: MAKE_HEAP, HEAP_SORT

contains

!! This subroutine forces the pyramid, whose top is at Apex, to obey the heap rule.  (A heap is
!! a collection of values arranged as a pyramid.)
!! The heap rule is that in a pyramid (or sub-pyramid), the value at the apex is always
!! greater than or equal to the two nodes directly beneath it.  (It follows that the node at
!! the apex is always greater than or equal to all nodes beneath it.)
!! When this procedure is entered, the two sub-pyramids of the node at Apex already obey the
!! heap rule.  However, the element at the apex may not obey the heap rule.

!! Method: beginning at node Apex, this subroutine determines which of the three elements (apex
!! and the two nodes directly beneath it) is the largest, and brings that element into the
!! apex position, and then moves the original apex element into the node just vacated.  The
!! action then transfers to the sub-pyramid where that element now exists, and works on the
!! three nodes (sub-apex and its two sub-nodes), bringing the largest to the sub-apex position.
!! The original apex element has now moved down another level, and the action continues down
!! there, and so on, until a sub-pyramid is found that already conforms to the heap rule
!! (or the base of the pyramid is reached).

subroutine MAKE_HEAP (A, Apex, No_Elements)
   !! INCOMING: A           = the array whose elements are to be sorted.
   !!           Apex        = the number of the apex element of a sub-pyramid.
   !!           No_Elements = the number of elements in the pyramid.
   !! OUTGOING: A           = array whose sub-pyramid at Apex obeys the heap rule.

   integer, dimension ( : ), intent (in out) :: A
   integer, intent (in)            :: Apex, No_Elements

   integer                         :: Temp        !! For holding one value of the pyramid.
   integer                         :: Node        !! The number of the node under consideration.

   Temp = A(Apex)                                 !! Save the apex element, as it may have to
                                                  !! make way for the largest.
   Node = 2*Apex                                  !! Start at the apex of the left-hand sub-pyramid.
   do
      if (Node > No_Elements) then
         exit
      end  if
      if (Node < No_Elements) then                !! If there is a right-sub-pyramid . . .
         if (A(Node) < A(Node+1)) then
                                                  !! The left sub-node is less than the right.
            Node = Node + 1                       !! Select the apex of the right sub-pyramid.
         end  if
      end  if
      if (Temp >= A(Node) ) then                  !! The apex is already the largest, . . .
         exit
      end  if                                     !! . . . so finish.
      A(Node/2) = A(Node)                         !! Move the larger element up one level of the
                                                  !! pyramid.
      Node = 2*Node                               !! Move one level down the pyramid.
   end  do

   A(Node/2) = Temp                               !! Install the original apex in its correct location.
end  subroutine  MAKE_HEAP

!! This procedure performs the overall tasks of (a) initial setup by organizing the incoming
!! array A as a pyramid obeying the heap rule, and (b) overseeing the extraction of successive
!! next-largest elements from the pyramid, and writing them back in order into the array A.
subroutine HEAP_SORT (A)
   !! INCOMING:   A = the array whose elements are to be sorted.
   !! OUTGOING:   A = array of sorted elements.  Elements are in ascending order.
   integer, dimension ( : ), intent (in out) :: A

   integer                         :: Temp        !! The same type as A.
   integer                         :: K, N, Apex

   N = ubound(A, 1)                               !! The number of elements.

   !! Section to structure the array elements as a pyramid obeying the heap rule.
   !! We start at the second-lowest level of the pyramid, and work towards the top.
   do Apex = N/2, 1, -1
      call MAKE_HEAP (A, Apex, N)
   end  do

   !! Section to sort.
   !! Each time around this loop removes the apex from the tree, places the element in an
   !! output array, and forces the pyramid still remaining to obey the heap rule.
   do K = N, 2, -1
      Temp = A(K)                                 !! Remove the apex element at A(1), and write
      A(K) = A(1)                                 !! it to the output array at A(K), by swapping
      A(1) = Temp                                 !! it with the element at A(K).
                                                  !! Placing a small element at A(1) violates the
                                                  !! heap rule.
                                                  !! Elements K through N are now sorted.
      call MAKE_HEAP (A, 1, K-1)                  !! Makes the remaining K-1 elements conform
   end  do                                        !! to the heap rule.

end  subroutine  HEAP_SORT

!! Figure 1.5

end  module  GROUP
