module GROUP
   implicit  none
   public :: HASH_SORT
   integer, parameter, public :: Double = selected_real_kind (14)

contains

!! This procedure sorts an array into increasing order, using an address-calculation sort.

subroutine HASH_SORT (A, N)
   !! INCOMING: A = an array whose elements are to be sorted.  The N elements are
   !!               stored in positions A(1) through A(N).  Elements A(-N) through
   !!               A(0) are used as workspace.
   !!           N = the number of elements to be sorted.
   !!             = the upper bound of array A; the lower bound is -N.
   !! OUTGOING: A = an array containing the sorted elements in A(1) through A(N).

   integer, intent (in)             :: N
   integer, dimension (-N: ), intent (in out)   :: A

   integer                         :: Temp
   integer                         :: H, J, K
   real (kind=Double), parameter   :: Factor = 2.0_Double
   real (kind=Double)              :: Divisor
   integer                         :: Largest, Smallest

   !! FIND THE SMALLEST & LARGEST ELEMENTS.
   Smallest = minval (A(1:N))
   Largest = maxval (A(1:N))

   if (Largest == Smallest)  then                 !! The elements are already in order.
      return
   end  if

   Divisor = real(Largest, kind=Double) - real(Smallest, kind=Double)
                                                  !! Double precision, to avoid fixed-point overflow,
                                                  !! maintaining high precision.
   !! INITIALIZE THE WORK AREA.
   A(-N:0) = Smallest

   !! DISTRIBUTE THE SMALLER ELEMENTS to the top of the array (the work area).
   do J = 1, N
      if (A(J) /= Smallest) then
         H = Factor*N*(real(A(J), kind=Double) - Smallest)/Divisor
         H = H - N                                !! Hash the value A(J).
         if (H < 0) then                          !! It can be moved.
                                                  !! A linear search for an empty location.
            do
               if (A(H) == Smallest) then
                  exit
               end  if
               H = H + 1
            end  do
            A(H) = A(J)                           !! Move it to the empty location.
            A(J) = Smallest                       !! Mark the old position <empty>.
         end  if
      end  if
   end  do
   !! Elements in A(-N) to A(-1) are nearly sorted.

   !! COMPRESS ALL ELEMENTS TO THE TOP OF THE ARRAY (that is, towards A(-N)).
   K = -N
   do J = -N, N
      if (A(J) /= Smallest) then
         if (J /= K) then
            A(K) = A(J)                           !! Move one value towards the bottom.
            A(J) = Smallest                       !! Mark the old position <empty>.
         end  if
         K = K + 1
      end  if
   end  do

   !! DISTRIBUTE THE LARGER ELEMENTS to the bottom of the array A(1) to A(N).
   do J = -N, -1
      if (A(J) /= Smallest) then
         H = Factor*N*(real(A(J), kind=Double) - Smallest)/Divisor
         H = H - N                                !! Hash the value A(J).
         if (H >= 0) then
            do
               if (A(H) == Smallest) then
                  exit
               end  if
               H = H - 1
            end  do
            A(H) = A(J)
            A(J) = Smallest
         end  if
      end  if
   end  do

   !! COMPRESS ALL ELEMENTS TO THE BOTTOM OF THE ARRAY, that is, towards A(N).
   K = N
   do J = N, -N, -1
      if (A(J) /= Smallest) then
         if (J /= K) then
            A(K) = A(J)
            A(J) = Smallest
         end  if
         K = K - 1
      end  if
   end  do

   !! THE ELEMENTS ARE AGAIN IN A(1) THROUGH A(N), BUT ARE NOT PERFECTLY SORTED.
   !! A RANKING SORT IS PERFORMED TO PUT THESE FEW ELEMENTS IN ORDER.
   do J = 1, N-1
      if (A(J) > A(J+1)) then
         Temp = A(J+1)
         do K = J, 1, -1
            if (A(K) < Temp) then
               exit
            end  if
            A(K+1) = A(K)
         end  do
         A(K+1) = Temp
      end  if
   end  do
                                                  !! AND THEY ARE SORTED.
end  subroutine  HASH_SORT

!! Figure 1.8  The Hash Sort

end  module  GROUP
