module GROUP
   implicit  none
   public :: HASH_SORT2

contains

!! This procedure sorts an array into increasing order, using an address-calculation sort.
!! An auxiliary parallel array P is used to indicate:
!! (1) whether a given location in array A is empty, and
!! (2) whether there has been a collision.  In this case, the pointer contains the location
!! of a likely free location.  This version provides for many equal elements and for cases
!! where many elements hash to the same location.

subroutine HASH_SORT2 (A, N)                      !!{XE "algorithm: Hash Sort"}
   !! INCOMING:   A = an array whose elements are to be sorted.  The elements are
   !!             stored in positions A(1) through A(N).  Elements A(-N) through
   !!             A(0) are used as workspace.
   !!       N = the number of elements to be sorted.
   !!          = the upper bound of array A; the lower bound is -N.
   !! OUTGOING:   A = an array containing the sorted elements in A(1) through A(N).

   integer, intent (in)             :: N
   integer, dimension (-N: ), intent (in out)   :: A

   integer, dimension (-N:N)       :: P           !! Pointers for resolving collisions, parallel to array A.
   integer                         :: Temp
   integer                         :: H, J, K
   integer                         :: Initial_Location
   integer, parameter              :: Double = selected_real_kind (14)
   real (kind=Double), parameter   :: Factor = 2.0_Double
   real (kind=Double)              :: Divisor
   integer                         :: Largest, Smallest
   integer                         :: Empty

   Empty = lbound(A, 1) - 1

   !! FIND THE SMALLEST & LARGEST ELEMENTS.
   Smallest = minval (A(1:N))
   Largest = maxval (A(1:N))

   if (Largest == Smallest) then                  !! The elements are already in order.
      return
   end  if

   Divisor = real(Largest, kind=Double) - Smallest!! Extra precision, to avoid fixed-point overflow.

   !! INITIALIZE THE WORK AREA.
   A(-N:0) = Smallest

   P = Empty                                      !! Initialize the pointers.

   !! DISTRIBUTE THE SMALLER ELEMENTS to the top of the array (the work area).
   do J = 1, N
      H = Factor*N*(real(A(J), kind=Double) - Smallest)/Divisor
      H = H - N                                   !! Hash the value A(J).
      if (H <= 0) then                            !! It can be moved.
                                                                                !! A linear search for an empty location.
         Initial_Location = H
         if (P(H) /= Empty) then                  !! Use the pointer to get
            H = P(H)                              !! an empty location quickly.
         end  if
         do
            if (P(H) == Empty) then
               exit
            end  if
            H = P(H)
         end  do
         A(H) = A(J)                              !! Move it to the empty location.
         A(J) = Smallest                          !! Just to show that it's unoccupied.
         P(H) = H + 1                             !! Link to the next location, hopefully empty.
         P(Initial_Location) = H + 1
                                                                                !! Link to the next location, hopefully empty.
         P(J) = Empty                             !! Mark the old position <empty>.
      else
         P(J) = J                                 !! Mark the element eligible for compression.
      end  if
   end  do
   !! Elements in A(-N) to A(-1) are nearly sorted.

   !! COMPRESS ALL ELEMENTS TO THE TOP OF THE ARRAY (that is, from A(-N)).
   K = -N
   do J = -N, N
      if (P(J) /= Empty) then
         if (J /= K) then
            A(K) = A(J)                           !! Move one value towards the bottom.
            A(J) = Smallest                       !! Mark the old position <empty>.
         end  if
         K = K + 1
      end  if
   end  do

   P = Empty                                      !! Initialize the pointers.

   !! DISTRIBUTE THE LARGER ELEMENTS to the bottom of the array A(1) to A(N).
   do J = -N, -1
      H = Factor*N*(real (A(J), kind=Double) - Smallest)/Divisor
      H = H - N                                   !! Hash the value A(J).
      if (H > 0) then
         Initial_Location = H
         if (P(H) /= Empty) then                  !! Use the pointer to get
            H = P(H)                              !! an empty location quickly.
         end  if
         do
            if (P(H) == Empty) then
               exit
            end  if
            H = P(H)
         end  do
         A(H) = A(J)
         A(J) = Smallest
         P(H) = H - 1
         P(Initial_Location) = H - 1              !! Link to previous location, hopefully empty.
         P(J) = Empty
      else
         P(J) = J                                 !! To mark the element for compression.
      end  if
   end  do

   !! COMPRESS ALL ELEMENTS TO THE BOTTOM OF THE ARRAY, that is, towards A(N).
   K = N
   do J = N, -N, -1
      if (P(J) /= Empty) then
         if (J /= K) then
            A(K) = A(J)
            A(J) = Smallest
         end  if
         K = K - 1
      end  if
   end  do

   !! THE ELEMENTS ARE NOW IN A(1) THROUGH A(N) AGAIN, BUT ARE NOT PERFECTLY SORTED.
   !! A RANKING SORT IS PERFORMED TO PUT THESE FEW ELEMENTS IN ORDER.
   do J = 1, N-1
      if (A(J) > A(J+1)) then
         Temp = A(J+1)
         do K = J, 1, -1
            if (A(K) < Temp) then
               exit
            end  if
            A(K+1) = A(K)
         end  do
         A(K+1) = Temp
      end  if
   end  do
                                                                             !! AND THEY ARE SORTED.
end  subroutine  HASH_SORT2

!! Figure 1.9  The Hash Sort, using pointers

end  module  GROUP