module M7
   implicit  none
   public :: CREATE, SEARCH

contains

!! This subroutine creates a hashed table T from the elements read in.  Quadratic Quotient
!! Searching is used to resolve collisions.
subroutine  CREATE (T)
      !! OUTGOING:  T = an array, initialized with values read in.

      integer, dimension (0: ), intent (out) :: T !! Defined as dimension(0:N), where N+1 is prime.

      integer                      :: Element
      integer                      :: M
      integer, parameter           :: Empty = -1
      integer                      :: A, Q
      integer                      :: J, H, H1, V
      integer                      :: K

      T = Empty
      M = size (T, 1)                             !! The number of elements in T (M is a prime
                                                  !! number).

      do J = 1, 10
         print *, "Please type a value for insertion in the hash table:"
         read *, Element
         H = modulo(Element, M)
         if (T(H) == Empty) then                  !! The location is empty.
            T(H) = Element
         else                                     !! Perform a quadratic quotient search for an
Probe:         do K = 1, 1                        !! empty location.
               Q = Element/M
               A = 1
               H1 = H
               V = Q
               do
                  H1 = H1 + A + V                 !! Forms Modulo(Element,M) + a*i + Q*i**2.
                  H1 = modulo(H1, M)              !! Forces H1 to conform to 0 <= H1 < M.
                  if (T(H1) == Empty) then        !! We found an empty location.
                     exit
                  end  if
                  if (H1 == H) then               !! We are back where we started; no empty location.
                     print *,  "Unable to find an empty location for ", &
                        Element
                     print *,  "The element has been ignored."
                     exit  Probe
                  end  if
                  V = V + Q+Q
               end  do
               T(H1) = Element                    !! Insert the element.
            end  do  Probe
         end  if
      end  do

end  subroutine  CREATE

!! This function performs a hash search on the table T to ascertain whether the
!! element Key exists within it.  If it does exist, the position of the element is returned,
!! otherwise negative is returned.
!! Quadratic quotient searching is performed when two or more keys have the same hash value.
function  SEARCH (T, Key) result (Position)
      !! INCOMING: T   = an array to be searched.
      !!           Key = the element for which the search is to be conducted.
      !! RETURNS:  the position of the element in the array, if it exists;
      !!           or negative, indicating the element does not exist in the array.

      integer, dimension (0: ), intent (in)  :: T !! Defined as dimension(0:N), where N+1 is prime.
      integer, intent (in)         :: Key
      integer                      :: Position

      integer                      :: H, H1
      integer, parameter           :: Empty = -1
      integer                      :: V, Q, A
      integer                      :: M

      M = size (T, 1)                             !! The number of elements in T (M is a prime
                                                  !! number).

      H = modulo (Key, M)                         !! Hash the key.
      if (Key == T(H) ) then                      !! The search was successful.
         Position = H
         return
      end  if
      if (T(H) == Empty) then                     !! The search was unsuccessful.
         Position = Empty
         return
      end  if

      !! Perform a quadratic search, using the formula H1 = hash(key) + a*i + Q*i**2.
      Q = Key/M
      A = 1
      H1 = H
      V = Q
      do
         H1 = H1 + A + V                          !! Forms Modulo(Key,M) + a*i + Q*i**2.
         H1 = modulo(H1, M)                       !! Forces H1 to conform to 0 <= H1 < M.
         if (T(H1) == Key) then                   !! The search was successful.
            Position = H1
            return
         end  if
         if (T(H1) == Empty) then                 !! The key does not exist in the table.
            Position = Empty
            return
         end  if
         if (H1 == H) then                        !! We are back where we started; no empty location.
            Position = Empty
            return
         end  if
         V = V + Q+Q
      end  do

      Position = Empty                            !! The key does not exist in the table.
end  function  SEARCH

!! Figure 10.7  The Quadratic Search algorithm

end  module  M7
