module M4
   implicit  none
   public :: ROMBERG

contains

!! This subroutine implements Romberg Integration

!! Method: A modified Trapezoidal Method is used to compute the area between
!! X = A and X = B for 1, 2, 4, 8, 16, . . . 2**(N-1) strips in succession.
!! However, to avoid repeating unnecessarily the evaluation of the function
!! each time the number of strips is doubled, we take the previously-computed
!! Trapezoidal Method sum, and evaluate the function F at only the new points.
!! For example, suppose that the Trapezoidal Method has been applied to 4 strips.
!! The function F already has been evaluated at the points a+h, a+2h, a+3h (as
!! well as the end points x = a and x = b).  For 8 strips, we only need to
!! evaluate the function F at the points x = a+h/2, a+3h/2, a+5h/2, a+7h/2, and to
!! add their contributions to one-half of the integral for 4 strips, and voila!, we have the
!! sum for 8 strips.
!! The Trapezoidal Method sums for 1, 2, 4, 8, ... strips are stored in the first
!! column of matrix R, in R(1,1), R(2,1), R(3,1), R(4,1), ... respectively.
!! Richardson Extrapolation is first applied to the integral calculated using 1 and 2
!! strips, and the extrapolated integral is deposited in R(2,2).
!! The integral using 4 strips is then formed.  With this and the integral obtained
!! for 2 strips and the extrapolated integral, a second extrapolation
!! is computed and stored in R(3,2).  Using again the previous extrapolations,
!! a new extrapolation is computed and stored in R(3,3).  And so on.

subroutine  ROMBERG (F, A, B, N, Integral)
   !! INCOMING: F        is the function whose integral is to be approximated;
   !!           A        = the left-hand end of the interval for integration;
   !!           B        = the right-hand end of the interval for integration;
   !!           N        = the number of sets for which the Trapezoidal method is
   !!                      to be employed.  For n = 6, for example,
   !!                      the number of strips is 1, 2, 4, 8, 16, and 32.
   !! OUTGOING: Integral = the extrapolated integral.
   interface                                      !! for the F/Fortran function defining the function to be integrated.
      function F (X) result (Value)
         real, intent (in)         :: X
         real                      :: Value
      end  function  F
   end  interface

   real, intent (in)               :: A, B
   integer, intent (in)            :: N
   real, intent (out)              :: Integral

   real, dimension (N,N)           :: R           !! Used for storing the integral computed
                                                  !! by the Trapezoidal Method, and the
                                                  !! extrapolations.  Trapezoidal Method sums
                                                  !! are stored in the first column.
   real                            :: H, Trap_Sum
   integer                         :: I, J

   H = B-A
   R = 0
   R(1,1) = H*(F(A) + F(B))/2                     !! Area of one strip.

   do I = 2, N
      !! Computes the Trapezoidal Method sum, using only the additional areas.
      !! Evaluates the function F at only the mid-points of the previous strips.
      !! There are 2**(I-2) strips, but only 2**(I-1) evaluations.
      H = H/2                                     !! Halve the interval (doubling the number of strips).

      Trap_Sum = 0
      do J = 1, 2**(I-1), 2                       !! At only the odd-numbered positions.
         Trap_Sum = Trap_Sum + F(A + J*H)
      end  do
      Trap_Sum = Trap_Sum * H                     !! Compute the additional areas.

      R(I,1) = R(I-1,1)/2 + Trap_Sum              !! Take half the sum for I-1 strips, and
                                                  !! add the sum of the areas at the mid-points.
                                                  !! R(I,1) now holds the Trapezoidal Method sum for
                                                  !! 2**(I-2) strips.
      do J = 2, I                                 !! Apply extrapolation.
         R(I,J) = (4**(J-1) * R(I, J-1) - R(I-1,J-1) ) / &
                                 (4**(J-1) - 1)
      end  do

   end  do

   !! Show the integrals computed by the Trapezoidal Method (first column of R)
   !! and the interpolated values:
   print *, "       Trapezoidal"
   print *, "Strips   Method    Interpolated Integral"
   print *, "        Integral"
   do J = 1, N
      print *, "  ", 2**(J-1), R(J, 1:J)
   end  do

   Integral = R(N,N)

end  subroutine  ROMBERG

!! Figure 11.4  Integration using the Romberg Method.

end  module M4
