module M5a
   implicit  none
   public :: ADAPTIVE

contains

!! This recursive function implements adaptive quadrature.
!! Method.  The area of the strip or interval [A,B] is approximated by two
!! methods: (1) the area of the strip treated as a trapezoid, and
!! (2) the area of the strip as computed by the Trapezoidal Method with
!! two strips.
!! If the error in approximating the area of the interval [A,B] meets the
!! criterion, the area of the trapezoid is returned (which approximates
!! the area of this strip).
!! If it does not meet the criterion, the strip is broken into two strips of
!! equal width, and the function is called recursively to form the integral
!! of each half-strip.  The areas of the two half-strips are then added to
!! give the area of the whole strip, which is then returned.
recursive function ADAPTIVE (F, A, B, Eps) result (Integral)
   !! INCOMING: F     is the function to be integrated;
   !!           A   = the left-hand end of the interval to be integrated;
   !!           B   = the right-hand end of the interval to be integrated;
   !!           Eps = the error level required (e.g., 1e-5).
   !! RETURNS:  An approximation to the integral of function F in the
   !!           interval [A, B].
   interface                                      !! for the F/Fortran function defining the function to be integrated.
      function F (X) result (Y)
         real, intent (in)         :: X
         real                      :: Y
      end  function F
   end  interface
   real, intent (in)               :: A, B, Eps

   real                            :: Integral

   real                            :: X_Mid, S1, S2, H

   H = B - A                                      !! Initially, assume one strip.
   X_Mid = A + H/2
   S1 = (F(A) + F(B))*H/2                         !! The area of the trapezoid approximating the whole
                                                  !! strip.
   S2 = (S1 + H*F(X_Mid) )/2                      !! The area of the same strip using the Trapezoidal
                                                  !! Method, but as two half-strips.
   if (abs (S2 - S1)/3 <= Eps) then               !! The error criterion was met.
      Integral = S2                               !! Accept the area of this trapezoid as approximating
                                                  !! the area of this strip.
   else                                           !! The error criterion was not met.
                                                  !! Break the interval into two halves
                                                  !! of equal width, and integrate each half.
      Integral = ADAPTIVE (F, A, X_Mid, Eps/2) +   &  !! The area of the left-hand trapezoid.
            ADAPTIVE (F, X_Mid, B, Eps/2)         !! The area of the right-hand trapezoid.
   end  if

end  function ADAPTIVE

!! Figure 11.5a  Recursive function for adaptive quadrature.

end  module  M5a
