module M5b
   implicit  none
   public :: ADAPTIVE

contains

!! This function implements adaptive quadrature.
!! Method.  If the error for the area of the interval or strip [A,B] meets
!! the criterion, the area of the trapezoid is returned (which approximates
!! the area of this strip).  It it does not meet the criterion, the strip is broken
!! into two equal strips and the criterion is applied to the left-hand strip.
!! If still unsuccessful, the left-hand half is halved again, and so on until the
!! strip meets the criterion.  If successful, the area of the left-hand strip
!! is computed and added to a partial sum; then the strip (of the same width)
!! immediately to the right is tested to see whether it meets the criterion,
!! and the width increased or decreased according to whether the criterion
!! is met or not.  If met, the width of the strip is doubled until the criterion
!! is not met, then decreased once.  The area of that strip is then taken and
!! added to the partial sum.  And so on.
function  ADAPTIVE (F, A, B, Epsilon_Dummy) result (Integral)
   !! INCOMING: F         is the function to be integrated;
   !!           A       = the left-hand end of the interval to be integrated;
   !!           B       = the right-hand end of the interval to be integrated;
   !!           Epsilon = the error level required.
   !! RETURNS:  An approximation to the integral of function F in the
   !!           interval [A, B].

   interface                                      !! for the F/Fortran function defining the function to be integrated.
      function F (X) result (Y)
         real, intent (in)         :: X
         real                      :: Y
      end  function  F
   end  interface
   real, intent (in)               :: A, B, Epsilon_Dummy

   real                            :: Integral

   real                            :: X_Mid, F_Mid, S1, S2, H, L, R, Orig_H, Eps

   Eps = Epsilon_Dummy
   H = B - A                                      !! Initially, assume one strip.
   Orig_H = H
   L = A
   R = B
   Integral = 0

   !! In this section, we decrease the width of the interval until the
   !! criterion is met.
   do
      X_Mid = L + H/2
      F_Mid = F(X_Mid)
      S1 = (F(L) + F(R))*H/2                      !! The area of the trapezoid approximating the whole strip.
      S2 = (S1 + H*F_Mid)/2                       !! The sum of two trapezoids obtained by
                                                  !! splitting the strip in two.
      if (abs (S2 - S1)/3 > Eps) then             !! The error criterion was not met.
                                                  !! Break the interval into two halves
                                                  !! of equal width, and integrate each half.
         if (H/(2*Orig_H) > 1e-5) then
            R = (L + R)/2                         !! Halve the interval.
            H = R - L
            Eps = Eps/2
            cycle                                 !! And try again.
         end  if
      end  if

      Integral = Integral + S2

      if (R >= B) then
         return
      end  if

      L = R                                       !! Advance to the adjacent interval . . .
      R = R + H                                   !! . . . and start with the same initial width.

      !! In this section, we increase the width of the interval H until the
      !! criterion is not satisfied.
      do
         X_Mid = L + H/2
         F_Mid = F(X_Mid)
         S1 = (F(L) + F(R))*H/2                   !! The area of the trapezoid approximating the whole strip.
         S2 = (S1 + H*F_Mid)/2                    !! The sum of two trapezoids obtained by
                                                  !! splitting the strip in two.
         if (abs (S2 - S1)/3 <= Eps) then         !! The error criterion was met.
            R = R + H                             !! Double the interval.
            H = R - L
            if (R < B) then
               Eps = 2*Eps
               cycle
            else
               R = B                              !! We changed R, so . . .
               H = R - L                          !! . . . recompute H.
            end  if
         end  if
         exit
      end  do
      exit
   end  do

end  function  ADAPTIVE

!! Figure 11.5b  Non-recursive function for adaptive quadrature.

end  module  M5b
