!! This program plots a curve along with the axes.
!! This version uses a picture image of 1024 x 1024 elements.
!! Uses a white background, with black or color graphs and text.

!! On a system that does not have Salford Graphics, comment out those lines
!! marked "Salford Graphics".
!! It may also be necessary to change the integer KIND=2 to another kind.

!! This program uses MAKE_PCX_FILE (modified so that black is color 7),
!! and DRAWLINB, which is a modified version of DRAWLINE to allow color
!! lines.
!!==============================================================================

module COLORS                                     !! Colors for PCX file rendition (and for printing).
   integer, parameter, public      :: SHORT = selected_int_kind(2)  !! For Salford.
   integer (kind=SHORT), parameter, public    :: White=0, Red=1, Yellow=2, Purple=3,  &
                                      Green=4, Light_Blue=5, Dark_Blue=6, Black=7
end  module  COLORS

module BYTE
   implicit  none
   integer, parameter, public      :: SHORT = selected_int_kind(2)  !! For Salford.
end  module BYTE

module  IMAGE
   implicit  none

   character (len=1), dimension (0:1023, 0:1023), public :: Pixel_Image
   integer, public                         :: WBP, Rec_No
   character (len=128), public             :: Buffer

end  module  IMAGE

module GROUP
   implicit none
   public                          :: DRAW_GRAPH, DRAW_LINE, DRAW_TEXT, SET_PIXEL, RUNGE_KUTTA, F
   public                          :: CIRCLE, DISPLAY_CHARACTER
!!   private                       :: WRITE_BYTE_INTO_BUFFER, COMPRESS

contains

   function F (X, Y) result (Returned_Value)    !! A function to evaluate the equation to be solved.
      !! INCOMING:   X = the X-coordinate ) of the point at which the equation is to be evaluated..
      !!       Y = the Y-coordinate )
!!      implicit  none

      real, intent (in)            :: X, Y
      real                         :: Returned_Value

      Returned_Value = Y*Y - X
      return
   end  function  F

!! This procedure graphs a given function between specified X and Y values.
!! It displays the X-axis and Y-axis, provided that they are in the chosen X and Y values.
subroutine  DRAW_GRAPH (X_Lower, X_Upper, Y_Lower, Y_Upper, &
                        X_axis_Annotation, Y_axis_Annotation, &
                        Graph_Color, Axes_Color)
   !! INCOMING: Function =          the given function, provided as a function of a single argument;
   !!           X_Lower =           the left-hand value to start the graph;
   !!           X_Upper =           the right-hand value to finish the graph;
   !!           Y_Lower =           the bottom (minimum) Y-value; and
   !!           Y_Upper =           the top (maximum) Y-value.
   !!           X_axis_Annotation = annotation for the X-axis.
   !!           Y_axis_Annotation = annotation for the Y-axis.
   !!           Graph_Color =       Color of graph.
   !!           Axes_Color =        Color of the axes.
   use COLORS
   use  IMAGE
!!   implicit  none
   real, intent (in)               :: X_Lower, X_Upper, Y_Lower, Y_Upper
   character (len=*), intent (in)  :: X_axis_Annotation, Y_axis_Annotation
   integer (kind=SHORT), intent (in)   :: Graph_Color, Axes_Color

   real                            :: X_Range, Y_Range
   real                            :: X, Y
   real                            :: X_increment, Y_increment
   real                            :: X_scale, Y_scale
   real                            :: X_offset, Y_offset
   real, dimension (0:100)         :: X_CO_ORD, Y_CO_ORD
   real                            :: Initial_Y
   integer                         :: X_OLD, Y_OLD
   character (len=4)               :: Text
   integer (kind=SHORT)            :: J
   integer                         :: E_Screen_Height, E_Screen_Width
   character (len=11)              :: Annotation
   integer, parameter              :: Screen_Width = ubound(Pixel_Image,1), &
                                      Screen_Height = ubound(Pixel_Image, 2)
   integer (kind=SHORT)            :: IX, IY      !! Integer X- and Y-coordinates respectively.

   Pixel_Image = char(White)                      !! Initialize the background color.

   E_Screen_Height = Screen_Height - 45
   E_Screen_Width = Screen_Width - 35

   X_Range = abs (X_Upper - X_Lower)
   Y_Range = abs (Y_Upper - Y_Lower)

   !! Prepare X and Y origins, and scales for the axes.
   X_increment = X_Range/E_Screen_Width
   Y_increment = Y_Range/E_Screen_Height
   X_scale = 1/X_increment
   Y_scale = 1/Y_increment
   X_offset = -X_Lower*X_scale
   Y_offset = -Y_Lower*Y_scale

   !! Draw the X_axis.
   if (Y_Upper*Y_Lower < 0) then                  !! The X-axis can be shown on the screen.
      !! Draw the X-axis.
      call  DRAW_LINE (0,  &
         E_Screen_Height - int(-Y_Lower*Y_scale),  &
         E_Screen_Width,         &
         E_Screen_Height - int(-Y_Lower*Y_scale), int(Axes_Color) )

      !! Attach an arrowhead to the X-Axis:
      IX = E_Screen_Width - 5
      IY = E_Screen_Height - int(-Y_Lower*Y_scale) - 3
      Pixel_Image (IY, IX) = char (Axes_Color)
      Pixel_Image (IY+1, IX+1:IX+2) = repeat (char (Axes_Color), 2)
      Pixel_Image (IY+2, IX+1:IX+4) = repeat (char (Axes_Color), 4)
      Pixel_Image (IY+3, IX+1:IX+5) = repeat (char (Axes_Color), 5)
      Pixel_Image (IY+4, IX+1:IX+4) = repeat (char (Axes_Color), 4)
      Pixel_Image (IY+5, IX+1:IX+2) = repeat (char (Axes_Color), 2)
      Pixel_Image (IY+6, IX) = char (Axes_Color)

      !! Draw "distance marks" on the X_axis.
      do J = ceiling(X_Lower), int(X_Upper)
         call  DRAW_LINE ( int(J*X_scale+X_offset),   &
            E_Screen_Height -int(-Y_Lower*Y_scale+5), &
            int(J*X_scale+X_offset),   &
            E_Screen_Height - int(-Y_Lower*Y_scale-5), int(Axes_Color) )

         !! Put measurements alongside the "distance marks" on the X-axis.
         IX = J*X_scale+X_offset+2
                                    IY = E_Screen_Height - (-Y_Lower*Y_scale-15)
         write (unit=Annotation, fmt="(I11)" ) J

         call  DRAW_TEXT (trim(adjustl(Annotation)), int(IX), int(IY), int(Axes_Color))
      end  do
   end  if

   !! Display X-axis annotation.
   if (len_trim(X_axis_Annotation) > 0) then
      IX = X_Upper*X_scale-80
                               IY = E_Screen_Height - (-Y_Lower*Y_scale+15)
      call  DRAW_TEXT (trim (X_axis_Annotation), int(IX), int(IY), int(Axes_Color) )
   end  if

   !! Draw the Y-axis.
   if (X_Upper*X_Lower < 0) then                  !! The Y-axis can be shown on the screen.
      !! Draw the Y-Axis.
      call  DRAW_LINE ( int (-X_Lower*X_scale), 0, &
         int (-X_Lower*X_scale), int(E_Screen_Height), int(Axes_Color) )

      !! Attach an arrowhead to the Y-Axis:
      IX = -X_Lower*X_scale - 3
      IY = 3
      Pixel_Image (IY, IX+3) = char (Axes_Color)
      Pixel_Image (IY+1, IX+2:IX+4) = repeat (char (Axes_Color), 3)
      Pixel_Image (IY+2, IX+2:IX+4) = repeat (char (Axes_Color), 3)
      Pixel_Image (IY+3, IX+1:IX+5) = repeat (char (Axes_Color), 5)
      Pixel_Image (IY+4, IX+1:IX+5) = repeat (char (Axes_Color), 5)
      Pixel_Image (IY+5, IX) = char (Axes_Color)
      Pixel_Image (IY+5, IX+6) = char (Axes_Color)

      !! Draw "distance marks" on the Y_axis.
      do J = ceiling(Y_Lower), int(Y_Upper)
         call  DRAW_LINE (int (-X_Lower*X_scale+5),   &
            int (E_Screen_Height-(J*Y_scale+Y_offset-1)),     &
            int (-X_Lower*X_scale-5),     &
            int (E_Screen_Height-(J*Y_scale+Y_offset-1)), int(Axes_Color) )

         !! Put measurements alongside the "distance marks" on the Y-axis.
         IX = -X_Lower*X_scale+10
                                    IY = E_Screen_Height - (J*Y_scale+Y_offset-2)
         write (unit=Annotation, fmt="(I11)" ) J
         call  DRAW_TEXT (trim(adjustl(Annotation)), int(IX), int(IY), int(Axes_Color) )
      end  do
   end  if

   !! Display Y-axis annotation.
   if (len_trim(Y_axis_Annotation) > 0) then
      IX = -X_Lower*X_scale+20
                                IY = 20          !! Y_Upper*Y_scale+Y_offset+2
      call  DRAW_TEXT (trim (Y_axis_Annotation), int(IX), int(IY), int(Axes_Color))
   end  if

   Initial_Y = 0.2
   do                                             !! This loop changes the initial Y value,
                                                  !! thus allowing multiple graphs.
                                                  !! For one graph, omit this loop.
      !! Annotate the Graph.
      write (unit=Text, fmt="(F4.1)" ) Initial_Y
      call DRAW_TEXT ("Y=" // Text, 10, E_Screen_Height-int(Initial_Y*Y_scale+Y_offset), int(Graph_Color))

      !! Generate the solution points, saving them in arrays X_CO_ORD & Y_CO_ORD.
      call RUNGE_KUTTA (F, 0.1, Initial_Y, 0.04, ubound(X_CO_ORD, 1),   &
                        X_CO_ORD, Y_CO_ORD)

      !! Draw the graph, joining together the solution points.
      X_OLD = int(X_CO_ORD(0)*X_scale+X_offset)
      Y_OLD = E_Screen_Height-int(Y_CO_ORD(0)*Y_scale+Y_offset)
      do J = 1, ubound (X_CO_ORD, 1)              !! Use a smaller increment to handle steep gradients.
         X = X_CO_ORD (J)
         Y = Y_CO_ORD (J)
         call  SET_PIXEL ( int(X*X_scale+X_offset), E_Screen_Height-int(Y*Y_scale+Y_offset), int(Graph_Color) )
         call  DRAW_LINE (X_OLD,  &
               Y_OLD,  &
               int(X*X_scale+X_offset),         &
               E_Screen_Height-int(Y*Y_scale+Y_offset), int(Axes_Color) )
         X_OLD = int(X*X_scale+X_offset)
         Y_OLD = E_Screen_Height-int(Y*Y_scale+Y_offset)
      end  do

      Initial_Y = Initial_Y + 0.1
      if (Initial_Y > 0.8) then
         exit
      end  if
   end  do                                        !! end of loop to produce multiple graphs.

   return
end  subroutine  DRAW_GRAPH

subroutine SET_PIXEL (Ix, Iy, Color)
   use  IMAGE
!!   implicit  none
   integer, intent (in)                  :: Ix, Iy, Color

   Pixel_Image(Iy,Ix:Ix) = char(Color)

end  subroutine  SET_PIXEL

!! This subroutine solves a differential equation using the Runge-Kutta method.
subroutine  RUNGE_KUTTA (F, Initial_x, Initial_y, h, N, X_CO_ORD, Y_CO_ORD)
   !! INCOMING:   F = a function to evaluate the equation to be solved;
   !!       x = the x-coordinate ) of the initial point.
   !!       y = the y-coordinate )
   !!       h = the small increment in x at which the function is to be evaluated.
   !!  N = Number of points.
!!   implicit  none
   interface
      function F (X, Y) result (F_result)                     !! A function to evaluate the equation to be solved.
      !! INCOMING:   X = the X-coordinate ) of the point at which the equation is to be evaluated..
      !!       Y = the Y-coordinate )
!!      implicit  none
      real, intent (in)                :: X, Y
      real :: F_result
      end  function  F
   end  interface

!! real                            :: F
   real, intent (in)               :: Initial_x, Initial_y, h
   integer, intent (in)            :: N
   real, dimension (0: ), intent (out) :: X_CO_ORD, Y_CO_ORD

   real                            :: x, y
   real                            :: k1, k2, k3, k4
   integer                         :: K

   x = Initial_x
   y = Initial_y

   X_CO_ORD (0) = x
   Y_CO_ORD (0) = y

   do K = 1, N
      k1 = h * F(x, y)
      k2 = h * F(x + h/2, y + k1/2)
      k3 = h * F(x + h/2, y + k2/2)
      k4 = h * F(x + h, y + k3)
      y = y + (k1 + 2*k2 + 2*k3 + k4)/6           !! Calculate the new value of Y at X = X + deltaX
      x = x + h                                   !! Don't forget to increment X.

      X_CO_ORD (K) = x                            !! Save the co-ordinates.
      Y_CO_ORD (K) = y
   end  do

   return
end  subroutine  RUNGE_KUTTA

!! This subroutine displays <Text> on the screen at location (Ix, Iy),
!! where Ix and Iy are pixel positions measured with respect to the top
!! left-hand corner of the screen, which is position (0, 0).
!! These positions are absolute.  They do NOT change according to the
!! position of the axes.

!! The text is displayed with color <Color>.

subroutine  DRAW_TEXT (Text, Ix, Iy, Color)
   !! INCOMING: Text =  The text to be displayed;
   !!           Ix =    The X co-ordinate of the start of the text;
   !!           Iy =    The Y co-ordinate of the start of the text.
   !!                   Note that co-ordinates are measured from the top
   !!                   left-hand corner of the image which is (0, 0).
   !!           Color = The color of the text to be displayed.
   use  IMAGE
!!   implicit none
   character (len=*), intent (in)  :: Text
   integer, intent (in)            :: Ix, Iy, Color

   integer                         :: I, X, Y

   X = Ix
   Y = Iy

   do I = 1, len(Text)

      !! Check that the character will lie within the picture area.
      if (X < 0 .or. X+8 > ubound (Pixel_Image, 2) ) then
         return
      end  if
      if (Y < 0 .or. Y+15 > ubound (Pixel_Image, 1) ) then
         return
      end  if

      select case ( Text(I:I) )
         case (".")
            Pixel_Image (Y+7:Y+9,X+3:X+5) = repeat (char(Color), 3)
            X = X + 8                             !! Advance to the next character position.
         case ("0")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 4, 6, 6, 4, 3, 1, 1 /), &
               (/ 2, 0, 0, 2, 7, 9, 9, 7, 2 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("1")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 2, 2, 1, 3 /),       &
               (/ 1, 0, 9, 9, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("2")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 2, 1, 1, 2, 4, 5, 5, 4, 2, 1, 1, 5 /), &
               (/ 1, 2, 2, 1, 0, 0, 1, 3, 4, 6, 7, 9, 9 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("3")
            call DISPLAY_CHARACTER (X, Y,          &
               (/ 3, 3, 2, 3, 5, 6, 6, 5, 3, 5, 6, 6, 5, 3, 2, 3, 3 /), &
               (/ 0, 1, 1, 0, 0, 1, 3, 4, 4, 4, 5, 7, 9, 9, 7, 7, 8 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("4")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 4, 1, 6 /),       &
               (/ 9, 0, 5, 5 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("5")
            call DISPLAY_CHARACTER (X, Y,          &
               (/ 5, 2, 2, 3, 4, 5, 6, 6, 5, 4, 3, 2, 1 /), &
               (/ 0, 0, 4, 3, 3, 4, 5, 7, 8, 9, 9, 8, 7 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("6")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 3, 2, 1, 1, 2, 4, 5, 6, 6, 5, 2, 1 /), &
               (/ 0, 1, 2, 4, 7, 9, 9, 8, 7, 5, 4, 4, 5 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("7")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 6, 4, 2 /),       &
               (/ 0, 0, 4, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("8")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 3, 2, 2, 3, 4, 5, 6, 6, 6, 5, 4, &
               3, 2, 1, 1, 1, 2, 3, 4, 5, 5, 4, 3 /), &
               (/ 0, 1, 2, 3, 3, 4, 5, 6, 7, 8, 9, &
               9, 8, 7, 6, 5, 4, 3, 3, 2, 1, 0, 0 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("9")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 3, 6, 6, 6, 5, 4, 3, 2, 1, 1, 2, &
               3, 4, 5, 6 /),          &
               (/ 9, 4, 3, 2, 1, 0, 0, 1, 2, 3, 4, &
               5, 5, 5, 4 /),          &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("+")
            Pixel_Image (Y+4:Y+4,X+1:X+5) = repeat (char(Color), 5)  !! Horizontal.
            Pixel_Image (Y+2:Y+6,X+3:X+3) = repeat (char(Color), 5)  !! Vertical.
            X = X + 8                             !! Advance to the next character position.
         case ("-")
            Pixel_Image (Y+4:Y+4,X+1:X+5) = repeat (char(Color), 5)  !! Horizontal.
            X = X + 8                             !! Advance to the next character position.
         case ("A")
            call DISPLAY_CHARACTER (X, Y,       &
            (/ 1, 3, 4, 5, 2, 5, 6 /),       &
            (/ 9, 0, 0, 4, 4, 4, 9 /),       &
            Color)
            X = X + 8                             !! Advance to the next character position.
         case ("B")
            call DISPLAY_CHARACTER (X, Y,    &
               (/ 1, 3, 4, 4, 4, 3, 1, 3, &
               4, 5, 6, 6, 5, 4, 1, 1 /), &
               (/ 0, 0, 1, 2, 3, 4, 4, 4, &
               4, 5, 6, 7, 8, 9, 9, 0 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("C")
            call DISPLAY_CHARACTER (X, Y,          &
               (/ 6, 5, 4, 3, 2, 1, 1, 2, 3, 4, 5, 6 /), &
               (/ 2, 1, 0, 0, 1, 2, 7, 8, 9, 9, 8, 7 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("D")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 4, 5, 6, 6, 5, 4, 1, 1 /), &
               (/ 0, 0, 1, 2, 7, 8, 9, 9, 0 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("E")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 1, 1, 4, 1, 1, 6 /),    &
               (/ 0, 0, 4, 4, 4, 9, 9 /),    &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("F")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 1, 1, 4, 1, 1 /),       &
               (/ 0, 0, 4, 4, 4, 9 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("G")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 6, 5, 4, 3, 2, 1, 1, 2, 3, 4, 5, 6, 6, 4 /), &
               (/ 2, 1, 0, 0, 1, 2, 7, 8, 9, 9, 8, 7, 5, 5 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("H")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 1, 6, 6, 6 /),       &
               (/ 0, 9, 4, 4, 0, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("I")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 4, 3, 3, 2, 4 /),       &
               (/ 0, 0, 0, 9, 9, 9 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("J")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 6, 5, 5, 4, 3, 2, 1 /),    &
               (/ 0, 0, 0, 7, 8, 9, 8, 7 /),    &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("K")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 1, 5, 1, 6 /),       &
               (/ 0, 9, 4, 0, 4, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("L")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 6 /),          &
               (/ 0, 9, 9 /),          &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("M")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 4, 7, 7 /),       &
               (/ 9, 0, 4, 0, 9 /),       &
               Color)
            X = X + 9                             !! Advance to the next character position.
         case ("N")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 6, 6 /),       &
               (/ 9, 0, 9, 0 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("O")
            call DISPLAY_CHARACTER (X, Y,          &
               (/ 6, 5, 4, 3, 2, 1, 1, 2, 3, 4, 5, 6, 6 /), &
               (/ 2, 1, 0, 0, 1, 2, 7, 8, 9, 9, 8, 7, 2 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("P")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 1, 3, 4, 4, 4, 3, 1 /), &
               (/ 9, 0, 0, 0, 1, 2, 3, 4, 4 /), &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("Q")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 6, 5, 6, 6, 6, 5, 4, 3, 2, 1, 1, 2, 3, 4, 5, 4 /), &
               (/ 9, 8, 7, 5, 2, 1, 0, 0, 1, 2, 7, 8, 9, 9, 8, 7 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("R")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 1, 3, 4, 4, 4, 3, 1, 3, 5 /), &
               (/ 9, 0, 0, 0, 1, 2, 3, 4, 4, 4, 9 /), &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("S")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 5, 4, 3, 2, 2, 2, 3, 5, 6, 6, 6, 5, 4, 3, 2, 1 /), &
               (/ 2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 9, 8, 7 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("T")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5, 3, 3 /),       &
               (/ 0, 0, 0, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("U")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 2, 3, 4, 5, 6, 6 /),    &
               (/ 0, 7, 8, 9, 9, 8, 7, 0 /),    &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("V")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 5 /),          &
               (/ 0, 9, 0 /),          &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("W")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 2, 4, 6, 7 /),       &
               (/ 0, 9, 5, 9, 0 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("X")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5 /),       &
               (/ 0, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5 /),       &
               (/ 9, 0 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("Y")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 3, 3, 5 /),       &
               (/ 0, 4, 9, 4, 0 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 5, 3, 3 /),       &
               (/ 0, 4, 0, 4, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("Z")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 6, 1, 6 /),       &
               (/ 0, 0, 9, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("a")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 2, 3, 4, 5, 5, 6 /),       &
               (/ 4, 3, 3, 3, 4, 8, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 2, 1, 1, 2, 4, 5 /),       &
               (/ 5, 5, 6, 8, 9, 9, 8 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("b")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 0, 1, 1 /),       &
               (/ 1, 0, 9 /),       &
               Color)
            call CIRCLE (X, Y, Color)
            X = X + 8                             !! Advance to the next character position.
         case ("c")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 5, 3, 2, 1, 1, 2, 3, 5, 6 /),       &
               (/ 5, 4, 4, 5, 6, 7, 8, 9, 9, 8 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("d")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 6, 6 /),       &
               (/ 1, 0, 9 /),       &
               Color)
            call CIRCLE (X, Y, Color)
            X = X + 8                             !! Advance to the next character position.
         case ("e")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 6, 4, 2, 1, 1, 2, 5, 6 /),       &
               (/ 6, 6, 4, 4, 6, 8, 9, 9, 8 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("f")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 4 /),       &
               (/ 6, 6 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 5, 5, 4, 3,  3,  2,  1,  1,  2 /),        &
               (/ 1, 1, 0, 0, 1, 11, 12, 13, 12, 12 /),        &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("g")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 6,  6,  5,  4,  2,  2,  3,  3 /),         &
               (/ 4, 11, 12, 13, 13, 12, 12, 13 /),         &
               Color)
            call CIRCLE (X, Y, Color)
            X = X + 8                             !! Advance to the next character position.
         case ("h")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1 /),       &
               (/ 0, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 2, 4, 5, 6, 6 /),       &
               (/ 5, 4, 4, 5, 6, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("i")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 3, 3, 4, 5 /),       &
               (/ 5, 4, 8, 9, 9 /),       &
               Color)
            Pixel_Image (Y+2, X+3) = char (Color) !! For the dot.
            X = X + 7                             !! Advance to the next character position.
         case ("j")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 3,  3,  2,  1,  1,  1,  2,  2 /),         &
               (/ 5, 4, 11, 12, 13, 13, 12, 12, 13 /),         &
               Color)
            Pixel_Image (Y+2, X+3) = char (Color) !! For the dot.
            X = X + 5                             !! Advance to the next character position.
         case ("k")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1 /),       &
               (/ 0, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 1, 5 /),       &
               (/ 3, 5, 9 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("l")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 3, 3, 4, 5 /),       &
               (/ 1, 0, 8, 9, 8 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("m")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 0, 0 /),       &
               (/ 4, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 0, 2, 3, 3 /),       &
               (/ 5, 4, 5, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 3, 5, 6, 6 /),       &
               (/ 5, 4, 5, 9 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("n")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1 /),       &
               (/ 4, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 4, 5, 5 /),       &
               (/ 5, 4, 4, 5, 9 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("o")
            call CIRCLE (X, Y, Color)
            X = X + 8                             !! Advance to the next character position.
         case ("p")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1,  1,  0 /),        &
               (/ 4, 14, 13 /),        &
               Color)
            call CIRCLE (X, Y, Color)
            X = X + 8                             !! Advance to the next character position.
         case ("q")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 6, 6, 7 /),       &
               (/ 4, 14, 13 /),        &
               Color)
            call CIRCLE (X, Y, Color)
            X = X + 8                             !! Advance to the next character position.
         case ("r")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1 /),       &
               (/ 4, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 4, 3, 3 /),       &
               (/ 5, 4, 5, 5, 4 /),       &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("s")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 3, 3, 4, 3, 2, 1, 1, 4, 4, 3, 2, 1, 2, 2 /),       &
               (/ 4, 5, 5, 4, 4, 5, 6, 7, 8, 9, 9, 8, 8, 9 /),       &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("t")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3 /),       &
               (/ 4, 4 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 2, 3, 4, 5 /),       &
               (/ 2, 8, 9, 9, 8 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("u")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 5 /),       &
               (/ 4, 9 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 1, 2, 3, 4, 5 /),       &
               (/ 4, 8, 9, 9, 9, 8 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("v")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 5 /),       &
               (/ 4, 9, 4 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("w")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 0, 0, 1, 2, 3, 3, 3, 4, 5, 6, 6 /),       &
               (/ 4, 8, 9, 9, 8, 4, 8, 9, 9, 8, 4 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("x")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 2, 3, 3, 2, 1 /),       &
               (/ 5, 4, 5, 8, 9, 8 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 4, 3, 3, 4, 5 /),       &
               (/ 5, 4, 5, 8, 9, 8 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("y")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3 /),          &
               (/ 4, 9 /),          &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5,  2,  1,  1,  2 /),         &
               (/ 4, 12, 13, 12, 12 /),         &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("z")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5, 1, 5 /),       &
               (/ 4, 4, 9, 9 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case (" ")
            !! No character display for a blank.
            X = X + 7                             !! Advance to the next character position.
         case ("(")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 2, 2, 4 /),       &
               (/ 0, 3, 6, 9 /),       &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case (")")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 4, 4, 2 /),       &
               (/ 0, 3, 6, 9 /),       &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("[")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 2, 2, 4 /),       &
               (/ 0, 0, 9, 9 /),       &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("]")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 3, 1 /),       &
               (/ 0, 0, 9, 9 /),       &
               Color)
            X = X + 5                             !! Advance to the next character position.
         case ("*")                               !! First do a cross.
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5 /),       &
               (/ 2, 6 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5 /),       &
               (/ 6, 2 /),       &
               Color)
            !! Now do a "+" sign.
            Pixel_Image (Y+4:Y+4,X+1:X+5) = repeat (char(Color), 5)  !! Horizontal line.
            Pixel_Image (Y+2:Y+6,X+3:X+3) = repeat (char(Color), 5)  !! Vertical line.
            X = X + 7                             !! Advance to the next character position.
         case ("/")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 6, 1 /),          &
               (/ 0, 9 /),          &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("\")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 6 /),          &
               (/ 0, 9 /),          &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("|")                               !! Produces an integral symbol.
            call DISPLAY_CHARACTER (X, Y,       &
               (/  4,  5,  5,  4,  4,  3,  3,  2,  &
                   1,  1,  2,  2 /),         &
               (/ -2, -2, -1, -1, -2, -1, 10, 11,  &
                  11, 10, 10, 11 /),         &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("=")
            Pixel_Image (Y+6:Y+6,X+1:X+5) = repeat (char(Color), 5)  !! Horizontal line.
            Pixel_Image (Y+4:Y+4,X+1:X+5) = repeat (char(Color), 5)  !! Horizontal line.
            X = X + 7                             !! Advance to the next character position.
         case ("!")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 3, 3, 4, 4 /),       &
               (/ 0, 6, 6, 0 /),       &
               Color)
            Pixel_Image (Y+8:Y+9, X+3:X+4) = repeat (char(Color), 2) !! A fullstop.
            X = X + 6                             !! Advance to the next character position.
         case ("""")                              !! Quotation mark.
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 2 /),       &
               (/ 0, 2 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 4 /),       &
               (/ 0, 2 /),       &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("'")                               !! Apostrophe.
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 3, 2 /),       &
               (/ 0, 2 /),       &
               Color)
            X = X + 5                             !! Advance to the next character position.
         case ("`")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 3 /),       &
               (/ 0, 2 /),       &
               Color)
            X = X + 5                             !! Advance to the next character position.
         case (">")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5, 1 /),       &
               (/ 3, 6, 9 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("<")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 1, 5 /),       &
               (/ 3, 6, 9 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("_")
            call DISPLAY_CHARACTER (X, Y,       &
               (/  0,  6,  6,  0 /),         &
               (/ 10, 10, 11, 11 /),         &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case (":")
            Pixel_Image (Y+4:Y+5, X+3:X+4) = repeat (char(Color), 2)
            Pixel_Image (Y+8:Y+9, X+3:X+4) = repeat (char(Color), 2)
            X = X + 6                             !! Advance to the next character position.
         case (";")
            Pixel_Image (Y+4:Y+5, X+3:X+4) = repeat (char(Color), 2)
            Pixel_Image (Y+8:Y+9, X+3:X+4) = repeat (char(Color), 2)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4,  4,  3 /),     &
               (/ 9, 10, 11 /),     &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case (",")
            Pixel_Image (Y+8:Y+9, X+3:X+4) = repeat (char(Color), 2)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4,  4,  3 /),     &
               (/ 9, 10, 11 /),     &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("%")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 6, 1 /),       &
               (/ 1, 8 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 3, 4, 4, 3, 2, 1, 1, 2 /),       &
               (/ 0, 0, 1, 2, 3, 3, 2, 1, 0 /),       &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 5, 6, 6, 5, 4, 3, 3, 4 /),       &
               (/ 6, 6, 7, 8, 9, 9, 8, 7, 6 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("&")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 6, 2, 2, 3, 4, 4, 1, 1, 2, 5, 6, 6 /),       &
               (/ 9, 2, 1, 0, 1, 2, 6, 8, 9, 9, 7, 6 /),       &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("^")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 3, 5 /),       &
               (/ 2, 0, 2 /),       &
               Color)
            X = X + 7                             !! Advance to the next character position.
         case ("$")                               !! Sisplay an "S", and then draw a vertical line thru it.
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 5, 4, 3, 2, 2, 2, 3, 5, 6, 6, 5, 4, 3, 2, 1 /), &
               (/ 2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 8, 7 /), &
               Color)
            !! Now do the vertical line:
            call DISPLAY_CHARACTER (X, Y,       &
               (/  4,  4 /),  &
               (/ -2, 11 /),  &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("#")                               !! Display an "S", and then draw a vertical line thru it.
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2,  2 /),   &
               (/ 0, 10 /),   &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4,  4 /),   &
               (/ 0, 10 /),   &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 0, 6 /), &
               (/ 4, 4 /), &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 0, 6 /), &
               (/ 6, 6 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("@")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 5, 4, 3, 2, 2, 3, 4, 5 /), &
               (/ 4, 3, 3, 4, 8, 9, 9, 8 /), &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 4, 5, 6, 7, 7, 5, 2, 0, 0,  1,  2,  4,  5 /),   &
               (/ 3, 8, 9, 8, 2, 0, 0, 2, 8, 10, 11, 11, 10 /),   &
               Color)
            X = X + 9                             !! Advance to the next character position.
         case ("?")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1,  2,  4, 6, 6, 4, 3, 3 /),  &
               (/ 1, -1, -1, 1, 2, 4, 4, 6 /),  &
               Color)
            !! Display a period.
            Pixel_Image (Y+8:Y+9, X+3:X+4) = repeat (char(Color), 2)
            X = X + 8                             !! Advance to the next character position.
         case ("~")
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 2, 3, 4, 5, 6 /), &
               (/ 4, 2, 2, 4, 4, 2 /), &
               Color)
            X = X + 8                             !! Advance to the next character position.
         case ("{")
            call DISPLAY_CHARACTER (X, Y,       &
               (/  4, 3, 3, 2, 3,  3,  4 /), &
               (/ -1, 0, 4, 5, 6, 10, 11 /), &
               Color)
            X = X + 6                             !! Advance to the next character position.
         case ("}")
            call DISPLAY_CHARACTER (X, Y,       &
               (/  2, 3, 3, 4, 3,  3,  2 /), &
               (/ -1, 0, 4, 5, 6, 10, 11 /), &
               Color)
            X = X + 6                             !! Advance to the next character position.

         case default
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 4 /), &
               (/ 4, 4 /), &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5 /), &
               (/ 5, 5 /), &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5 /), &
               (/ 6, 6 /), &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 1, 5 /), &
               (/ 7, 7 /), &
               Color)
            call DISPLAY_CHARACTER (X, Y,       &
               (/ 2, 4 /), &
               (/ 8, 8 /), &
               Color)
            X = X + 7                             !! Advance to the next character position.
      end  select
   end  do

end  subroutine  DRAW_TEXT

!! Draws a circle associated with the lower-case characters b, d, g, o, p, q.
subroutine  CIRCLE (X, Y, Color)
   !! INCOMING: X =     X co-ordinate of where to display the character;
   !!           Y =     Y co-ordinate of where to display the character;
   !!                   The co-ordinates specify the lower left-hand
   !!                   corner of the character.
   !!           Color = color of the character to be displayed.
!!   implicit  none
   integer, intent (in)            :: X, Y
   integer, intent (in)            :: Color

   call DISPLAY_CHARACTER (X, Y,       &
            (/ 1, 2, 3, 4, 5, 6, 6, 5, 3, 2, 1, 1 /),       &
            (/ 6, 5, 4, 4, 5, 6, 8, 9, 9, 8, 7, 6 /),       &
            Color)

end  subroutine  CIRCLE

!! This subroutine draws a character at position (X, Y).
subroutine  DISPLAY_CHARACTER (X, Y, DOT_X, DOT_Y, Color)
   !! INCOMING: X =     X co-ordinate of where to display the character;
   !!           Y =     Y co-ordinate of where to display the character;
   !!                   The co-ordinates specify the lower left-hand
   !!                   corner of the character.
   !!           DOT_X = X co-ordinates of points representing the outline of the character;
   !!           DOT_Y = Y co-ordinates of points representing the outline of the character;
   !!           Color = color of the character to be displayed.
!!   implicit  none
   integer, intent (in)            :: X, Y
   integer, dimension ( : ), intent (in)  :: DOT_X !! For the X co-ordinates.
   integer, dimension ( : ), intent (in)  :: DOT_Y !! For the corresponding Y co-ordinates.
   integer, intent (in)            :: Color

   integer                         :: J

   do J = 1, ubound(DOT_X, 1) - 1
      call DRAW_LINE (X+DOT_X(J), Y+DOT_Y(J),   &
            X+DOT_X(J+1), Y+DOT_Y(J+1), Color)
   end  do

end  subroutine  DISPLAY_CHARACTER

!! This special version of DRAW_LINE draws color lines.
!! Uses a picture array of 1024 x 1024 elements.
!! Assumes that one pixel is stored per byte.

!! This subroutine draws a straight line from the starting point (Xs,Ys) to the
!! finishing point (Xf,Yf).  It also performs scaling of the artwork according to
!! the value of SCALE, and gives it an OFFSET, for centering on the page.

subroutine  DRAW_LINE (Xs, Ys, Xf, Yf, Color)
      !! INCOMING: (Xs, Ys) are the coordinates of the starting point;
      !!       (Xf. Yf) are the coordinates of the finish point.
      use  IMAGE
!!      implicit  none
      integer, intent (in)         :: Xs, Ys,        &
                                      Xf, Yf
      integer, intent (in)         :: Color

      integer, parameter           :: Scaling = 1
      integer, parameter           :: Offset = 0
      integer                      :: Row, Col
      integer                      :: X_increment,   &
                                      Y_increment
      integer                      :: Round_Off, Itemp

      if (Ys > Yf) then                           !! "Y_increment" is 1 Pixel in the Y-direction.
         Y_increment = -1
      else
         Y_increment = 1
      end  if
      if (Xs > Xf) then                           !! "X_increment" is 1 Pixel in the X-direction.
         X_increment = -1
      else
         X_increment = 1
      end  if

      if (int(Xf*Scaling) == int(Xs*Scaling) ) then   !! Only the Y-value changes (vertical line).
         Col = Xs*Scaling + Offset
         do Row = Ys*Scaling, Yf*Scaling, Y_increment
            Pixel_Image(int(Row),int(Col)+Offset) = char(Color)
         end  do
      else if (abs(Xs-Xf) >= abs(Ys-Yf) ) then    !! The slope of the line is <= 45 degrees.
                                                  !! The X and Y-values change, or just X changes.
         Round_Off = (Xf-Xs)/2
         do Col = Xs*Scaling, Xf*Scaling, X_increment
            Itemp = (Yf-Ys)*(Col-Xs*Scaling)
            Itemp = Itemp + sign(Round_Off, Itemp)
            Row = Itemp/(Xf - Xs) + Ys*Scaling

            Pixel_Image(int(Row), Col+Offset) = char(Color)
         end  do
      else                                        !! The slope of the line is > 45 degrees.
                                                  !! The X and Y-values change, or just X changes.
         Round_Off = (Yf-Ys)/2
         do Row = Ys*Scaling, Yf*Scaling, Y_increment
            Itemp = (Xf-Xs)*(Row-Ys*Scaling)
            Itemp = Itemp + sign(Round_Off, Itemp)
            Col = Itemp/(Yf - Ys) + Xs*Scaling
            Pixel_Image(int(Row), Col+Offset) = char(Color)
         end  do
      end  if

      return
end  subroutine  DRAW_LINE

end  module  GROUP

module  PCX_Header
   !! The next set of declarations constitutes the PCX_Header (which is 128 bytes in size).
   type, public                    :: Header
      character (len=1)            :: Identifier  !! File Identification.
      character (len=1)            :: Version     !! Version Number.
      character (len=1)            :: Encoding_format      !! 0 = no encoding (uncompressed);
                                                                                   !! 1 = run-length encoding (compressed).
      character (len=1)            :: Bits_per_pixel    !! 4 bits per pixel.
                                                                                   !! Co-ordinates for the start of the image:
      character (len=2)            :: X_Start     !! X co-ordinate of start.
      character (len=2)            :: Y_Start     !! Y co-ordinate of start.
                                                                                   !! Co-ordinates for the end of the image:
      character (len=2)            :: X_end       !! X co-ordinate of end.
                                                                                   !! 0:1023 = length of image in pixels.
      character (len=2)            :: Y_end       !! Y co-ordinate of end.
                                                                                   !! 0:1023 = height of image in pixels.
      character (len=2)            :: Horizontal_Resolution   !! in dots per inch.
      character (len=2)            :: Vertical_Resolution  !! in dots per inch.
      character (len=1), dimension (0:47) :: Palette  !! The color palette.
      character (len=1)            :: Reserved
      character (len=1)            :: Num_Bit_Planes
      character (len=2)            :: Bytes_per_line    !! # of bytes per line unencoded.
      character (len=2)            :: Palette_type
      character (len=2)            :: Horizontal_Screen_Size
      character (len=2)            :: Vertical_Screen_Size
      character (len=54)           :: Reserved2
   end  type  Header                              !! End of declarations for the PCX header.

end  module  PCX_Header

module GROUP2
   use IMAGE
   private
   public :: WRITE_BYTE_INTO_BUFFER, COMPRESS, MAKE_PCX_FILE, PACK_CHAR
   implicit none

contains

!! These subroutines create a compressed picture file in PCX format.
!! The record size of the output file is 128 bytes.

!! This subroutine accepts a single byte and stores it in a buffer,
!! writing out the buffer whenever it becomes full.
subroutine  WRITE_BYTE_INTO_BUFFER (Byte_to_Write)
      character (len=*), intent (in)   :: Byte_to_Write

      WBP = WBP + 1
      if (WBP > 128) then
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Buffer
         WBP= 1
      end  if

      Buffer(WBP:WBP) = Byte_to_Write
   end  subroutine  WRITE_BYTE_INTO_BUFFER

!! This subroutine uses run-length encoding to compress the picture image in Pixel_Image.
!! The area of the image to be compressed is 0:X_end, 0:Y_end.
!! The image is written out via subroutine WRITE_BYTE_INTO_BUFFER.
subroutine  COMPRESS (X_end, Y_end)
   integer, intent (in)            :: X_end       !! The address of the RHS column of the image.
   integer, intent (in)            :: Y_end       !! The address of the bottom row of the image.

   character (len=1)               :: Byte, CHR
   integer                         :: I, CP, Start, Ending, Run_Length
   integer                         :: UB4, LB4
   character (len=1)               :: B8

   do I = 0, Y_end                                !! For each of the scan lines.
      CP = 0
      do
         UB4 = ichar(Pixel_Image(I,CP) )          !! Note the pair of pixels at the start of a potential
         LB4 = ichar(Pixel_Image(I,CP+1) )        !! run.
         B8 = char(UB4*16 + LB4)
         Start = CP                               !! The location of the byte at the start of the run.
         do                                       !! Look for the end of the run.
            CP = CP + 2
            if (CP >= X_end) then
               exit
            end  if
            UB4 = ichar(Pixel_Image(I, CP) )
            LB4 = ichar(Pixel_Image(I, CP+1) )
            if (B8 /= char(UB4*16 + LB4) ) then
               exit
            end  if
         end  do
         !! When we get here, CP points at the byte that terminates the run, or is > X_end.
         Ending = CP
         Run_Length = (Ending - Start)/2          !! The length of the run. [Note: not
                                                                                !! Ending-Start+1.].
         Byte = B8                                !! The byte at the start of the run.

         !! This loop breaks very long runs to chunks of 63 bytes.
         do
            if (Run_Length <= 63) then
               exit
            end  if
            call WRITE_BYTE_INTO_BUFFER (char(255) )  !! 'C0'X + 63.
            call WRITE_BYTE_INTO_BUFFER (Byte)
            Run_Length = Run_Length - 63
         end  do

         !! This section handles runs of length 1 (special case) to length 63.
         if (Run_Length == 1) then                !! Just one byte in this run.
            if (Byte <= char(63) ) then           !! '00111111'B
               !! Special case: upper 2 bits are zero.
               call WRITE_BYTE_INTO_BUFFER(Byte)
            else                                  !! Bad luck: one data byte expands to 2 bytes.
               !! The count byte is '11000001'B = 193, for a one-byte run.
               CHR = char(193)
               call WRITE_BYTE_INTO_BUFFER (CHR)
               call WRITE_BYTE_INTO_BUFFER (Byte)
            end  if
         else if (Run_Length > 1) then
            CHR = char (192 + Run_Length)         !! 'C0'X = '11000000'B = 192
                                                                                !! 'C0'X + length of run.
            call WRITE_BYTE_INTO_BUFFER (CHR)
            call WRITE_BYTE_INTO_BUFFER (Byte)
         end  if

         if (CP >= X_end) then                    !! We're past the end of a scan line.
            exit
         end  if
      end  do
   end  do

   do I = WBP+1, len(Buffer)+1                    !! Fill out buffer with zero bytes, and one more,
      call WRITE_BYTE_INTO_BUFFER (char(0) )      !! and write out the final record.
   end  do

end  subroutine  COMPRESS

!! This subroutine writes the header for a picture file in PCX format.
!! It then calls COMPRESS to prepare the graphic image.
!! The record size of the output file is 128 bytes.{XE "PCX file: compressed"}
subroutine  MAKE_PCX_FILE ( )                     !!{XE "algorithm: make PCX file"}{XE "algorithm: graphics: make PCX file"}
   !! A picture file commences with a 128-byte header, which gives video screen
   !! parameters, colors, and size of image.
   !! For details of PCX file formats, see J.D. Murray & W. vanRyper, "Encyclopedia of Graphics
   !! File Formats", 2nd Ed., O'Reilly, Sebastopol, CA, 1996.

   use  PCX_Header

   type (Header)                   :: H

   integer                         :: Bytes_per_line, IX_end, IY_end

   integer                         :: Height, Width

   !! INITIALIZATION SECTION: Initialize parameters in the PCX file header.
               !! The next array holds the color palette, 16 colors in groups of 3.
               !! Each line represents the components of one color: red, green, & blue.
               !! 0 represents none of that color, 255 represents the
               !! full intensity of that color.
               !! The order of the following colors is unimportant, and has been
               !! selected for providing best contrast on paper.
               !! White is color #0, best background for printing on white paper.  For
               !! displaying on screen, black is usually the best background, so exchange
               !! black and white (#0 with #13).
   H%Palette = (/ &  !!  Red     Green       Blue
                     char(255), char(255), char(255), & !!  0 = White.
                     char(255), char(0),   char(0),   & !!  1 = Red.
                     char(255), char(255), char(0),   & !!  2 = Yellow.
                     char(255), char(0),   char(255), & !!  3 = Purple.
                     char(102), char(255), char(102), & !!  4 = Light Green
                     char(0),   char(255), char(255), & !!  5 = Light blue.
                     char(0),   char(0),   char(255), & !!  6 = Dark blue.
                     char(255), char(102), char(0),   & !!  7 = Orange.
                     char(0),   char(255), char(0),   & !!  8 = Green.
                     char(255), char(127), char(255), & !!  9 = Pale purple.
                     char(255), char(255), char(127), & !! 10 = Lemon.
                     char(0),   char(127), char(255), & !! 11 = Medium blue.
                     char(255), char(192), char(0),   & !! 12 = Light brown.
                     char(0),   char(0),   char(0),   & !! 13 = Black.
                     char(0),   char(0),   char(0),   &
                     char(0),   char(0),   char(0)  /)

   H%Identifier = char(10)
   H%Version   = char(5)
   H%Encoding_format = char(1)                    !! Run length encoding (compressed file).
   H%Bits_per_pixel = char(4)
   H%X_Start = PACK_CHAR (0)
   H%Y_Start = H%X_Start
   H%X_End = PACK_CHAR (1023)
   H%Y_End = H%X_End
   H%Horizontal_Resolution = PACK_CHAR (300)
   H%Vertical_Resolution = PACK_CHAR (300)

   H%Reserved = char (0)
   H%Num_Bit_Planes = char (1)
   H%Bytes_Per_Line = PACK_CHAR (512)
   H%Palette_Type = PACK_CHAR (4)
   H%Horizontal_Screen_Size = H%Horizontal_Resolution
   H%Vertical_Screen_Size = H%Vertical_Resolution
   H%Reserved2 = repeat (char(0), len(H%Reserved2) )
   !! End of initialization section.

   !! This section crops the image -- it removes empty borders at the right-hand side and
   !! the bottom.
   !! Determine the height of the actual image.
   do Height = ubound(Pixel_Image, 1), 2, -1
      if ( any(Pixel_Image(Height, : ) /= char(0) ) ) then
         exit
      end  if
   end  do
   !! When we get here, Height is the row number of the bottom of the image.
   IY_end = Height
   H%Y_end = PACK_CHAR (IY_end)

   !! Determine the width of the actual image.
   do Width = ubound(Pixel_Image, 2), 2, -1
      if ( any(Pixel_Image( : , Width) /= char(0) ) ) then
         exit
      end  if
   end  do
   !! When we get here, Width is the column number of the RHS of the image.
   IX_end = Width
   Bytes_per_line = (IX_end+8)/8*ichar(H%Bits_per_pixel)
                                                                                !! +8, not +7, because pixels number from 0.

   if (modulo(Bytes_per_line, 2) /= 0) then
      Bytes_per_line = Bytes_per_line + 1         !! Make sure that the line has an even # of bytes.
   end  if
   H%Bytes_per_line = PACK_CHAR (Bytes_per_line)
   IX_end = Bytes_per_line*8/ichar(H%Bits_per_pixel)-1   !! Recompute the image area.
   H%X_end = PACK_CHAR (IX_end)

   open (unit=9, file="GRAPH.PCX", action="write",  &
      access="direct", form="unformatted", status="new", recl=128)

   Rec_No = 1
   write (unit=9, rec=Rec_No) H                   !! Write out the parameters of the picture file.

   call COMPRESS (IX_end, IY_end)                 !! Write out the compressed image.

end  subroutine  MAKE_PCX_FILE

!! This function converts a binary integer -- containing a positive integer of
!! 16 significant bits -- to a two-byte integer in reverse format, that is,
!! least-significant byte first.  The reverse-byte integer is returned as CHARACTER.
function  PACK_CHAR (Whole_Number) result (Pair)
      integer, intent (in)         :: Whole_Number

      character (len=2)            :: Pair

      Pair(2:2) = char (ishft (Whole_Number, -8) )!! The high-order bits come second.
      Pair(1:1) = char (iand (Whole_Number, 255) )!! The low-order bits come first.

   end  function  PACK_CHAR

end  module  GROUP2

!! Figure 4.17  Procedures to write out a compressed picture file in PCX format


!! include "SALFORD.INC"                            !! Salford graphics interface.  (Refer to Appendix C.)


program DISPLAY
!!   use  SALFORD                                   !! Supply Salford colors and Interfaces.
   use  COLORS                                    !! Colors for the PCX file rendition.
   use  IMAGE
   use  GROUP
   use  GROUP2
   implicit  none

   integer                         :: I, J
   integer (kind=SHORT)                :: Ix, Iy
   integer (kind=SHORT)                :: Key

   WBP = 0
   Rec_No = 1

   call  DRAW_GRAPH (-0.6,    &                   !! X-axis left value;
            5.0,     &                            !! X-axis right value;
           -2.1,     &                            !! Y-axis lower value;
            2.1,     &                            !! Y-axis upper value;
           "X-AXIS", &                            !! a label for the X axis;
           "Y-AXIS", &                            !! a label for the Y axis;
           Black, &                               !! color of the graph; and
           Black)                                 !! color of the axes.

   call DRAW_TEXT ("GRAPH OF dy/dx = y**2 - x", 270, 20, int(Black))

   !! Display the Pixel Image on Salford Graphics:
!!   call VGA@

!!   do J = 1, 479
!!      do I = 1, 639
!!         Ix = I
!!         Iy = J
!!         call SET_PIXEL@ (Ix, Iy,   &
!!            int ( ichar(Pixel_Image(Iy,Ix)), kind=SHORT) )
!!      end  do
!!   end  do

!!   call GET_KEY@(Key)                             !! Pause to admire the display.
!!   call TEXT_MODE@                                !! Press key to proceed.
   !! End of the section to display on Salford Graphics.

   print *, "Completed generating the image.  Now writing the PCX file."
   !! Ensure that no color value is greater than 7.  More causes the picture
   !! to break up with picture file readers.  Mask with 7.
   Pixel_Image = char (iand (ichar(Pixel_Image), 7) )
   call MAKE_PCX_FILE ( )

   print *, "Finished writing the PCX file."

end  program  DISPLAY
