module GROUP
   implicit  none
   public :: QUICK_SORT, SWAP

contains

!! This is a professional version of Quicksort.
recursive  subroutine  QUICK_SORT (A, LLeft, RRight)
   !! This subroutine implements an improved Quicksort, based on the algorithm by
   !! C. A. R. Hoare.
   !! INCOMING: A      = an array whose elements are to be sorted;
   !!           LLeft  = a pointer to the left-hand end of the partition to be sorted;
   !!           RRight = a pointer to the right-hand end of the partition to be sorted;
   !! OUTGOING: A      = an array whose elements A(LLeft) through A(RRight) are sorted.

   integer, dimension ( : ), intent (in out) :: A
   integer, intent (in)            :: LLeft, RRight

   integer                         :: Mid
   integer                         :: Left, Right
   integer                         :: Ref, &      !! To hold the Reference Element.
                                      Temp        !! A temporary for swapping elements,
                                                  !! same type as A.
   integer                         :: L, R, La    !! L = a pointer used when searching
                                                  !! from the left.
                                                  !! R = a pointer, used when searching
                                                  !! from the right.
   integer                         :: No_Swaps, & !! To count the number of swaps.
                                      J
   integer                         :: Rml
   integer, parameter              :: One = 1

   Left = LLeft
   Right = RRight
   if (Right <= Left) then                        !! The partition is empty or contains one element.
      return                                      !! There's no work to do.
   end  if

   !! Section to select the median of three elements, and to move it to the left.
   Mid = (Left + Right)/2
   if (A(Mid)  > A(Right))  then
      call SWAP (A(Mid), A(Right))
   end  if

   if (Left+1 == Right) then                      !! There are 2 elements in the partition,
      return                                      !! & they are now in sort.
   end  if
   if (A(Left) > A(Mid))  then
      call SWAP (A(Left), A(Mid))
   end  if
   if (A(Mid)  > A(Right))  then
      call SWAP (A(Mid), A(Right))
   end  if
   if (Left+ 2 == Right) then                     !! There are 3 elements in the partition,
      return                                      !! & they are now in sort.
   end  if
   if (A(Mid) == A(Right)) then                   !! Some elements are equal!
      Ref = A(Left)                               !! Forces the left partition to omit equal elements.
   else
      call SWAP (A(Left), A(Mid))
      Ref = A(Left)                               !! Select the Reference Element.
   end  if

   L = Left
   R = Right + 1

   !! Partition the elements into three groups.
   No_Swaps = 0
   do
      if (L >= R) then
         exit
      end  if
      do L = L + 1, R - 1                         !! Scan from the left for an element
         if (A(L) > Ref) then                     !! larger than the Reference.
            exit
         end  if
      end  do

      do                                          !! Scan from the right for an element
         R = R - 1
         if (A(R) <= Ref) then                    !! less than or equal to the Reference.
            exit
         end  if
      end  do

      if (L < R) then                             !! Swap two elements that are in the
                                                  !! wrong partitions.
         Temp = A(R)
         A(R) = A(L)
         A(L) = Temp
         No_Swaps = No_Swaps + 1                  !! Count each swap as we go.
      end  if
   end  do
                                                  !! Partitioning is complete.
   if (Left < R) then                             !! Swap the Reference Element into its
                                                  !! final position R in the array.
      A(Left) = A(R)
      A(R)    = Ref
   end  if
   !! At this point, A(R) is in its correct position in the list.  Elements A(Left) to A(R-1)
   !! are less than or equal to A(R), and elements A(R+1) to A(Right) are greater then A(R).

   !! Section to find out why no swaps were performed.
   if (No_Swaps == 0) then                        !! Something funny happened: not one
INCREASING:                   &                   !! element was moved. Investigate cause.
      do
         !! Look for any pre-existing order.
         do J = Left, Right-1
            if (A(J) > A(J+1))  then
               exit  INCREASING
            end  if
         end  do
         return                                   !! The elements are already in order.
      end  do  INCREASING

   !! Section to take a strong hand when the maximum number of elements is swapped.
   !! It's possible that the elements were in decreasing order.  Check it.
   else if (No_Swaps+1 == (Right-Left)/2) then
                                                  !! All possible pairs were swapped.
DECREASING:                   &                   !! Perhaps the elements were in reverse
      do                                          !! order?  Find out why.
         Rml = Right - Left
         if (iand(Rml, One) /= 0) then            !! A partition containing an even number
                                                  !! of elements was disarranged during
                                                  !! partitioning.
            if (Left < R-1)   then
               call SWAP (A(Left), A(R-1))        !! Restore order.
            end  if
         end  if
         do J = Left, Right-1                     !! Check that the elements are sorted.
            if (A(J) > A(J+1)) then
               exit DECREASING
            end  if
         end  do
         return                                   !! The entire sub-list is sorted.
      end  do  DECREASING
   end  if

   do La = R-1, Left+1, -1                        !! Pass over any elements that are equal to Ref.
      if (A(La) /= Ref) then
         exit
      end  if
   end  do
   !! At this point, elements A(La+1) through A(R) are equal.
                                                  !! A(L) is in its correct position too, even
                                                  !! if it is not equal to A(Mid)!
                                                  !! But if Left=La already, the partition
                                                  !! was just lop-sided.

   if (Left < La) then
      call QUICK_SORT (A, Left, La)               !! Partition the left segment.
   end  if
                                                  !! The element at R is in its correct position.
   if (R+1 < Right) then
      call QUICK_SORT (A, R+1, Right)             !! Partition the right segment.
   end  if

end  subroutine  QUICK_SORT

!!  insert the procedure SWAP from Figure 2.8 

!! This subroutine swaps the element Left_Element with Right_Element.
subroutine SWAP (Left_Element, Right_Element)
   integer, intent (in out)        :: Left_Element, Right_Element

   integer                         :: Temp

   Temp     = Left_Element
   Left_Element   = Right_Element
   Right_Element  = Temp
end  subroutine  SWAP

!! Figure 2.10

end  module  GROUP
