!! This program creates a color TIFF file of Sierpinski's curve.
!! Used to create a 2-color filled-in Sierpinski curve for Figure 2.15.
!! Uses a modified version of DRAW_LINE for color images.
module  COMPASS
   implicit  none
   integer, parameter, public      :: North = 0, East = 1, South = 2, West = 3
end  module  COMPASS

!! This module defines intensities for use with the TIFF color palette.
module  INTENSITIES
   implicit  none
   character (len=2), parameter, public ::               &
                        Full = char(255) // char(255),   &
                        Tqtr = char(204) // char(0),     &
                        Half = char(127) // char(255),   &
                        Qtr  = char(128) // char(0),     &
                        Nil  = char(0)   // char(0),     &
                        Special = char(192) // char(0)   !! '110 000 000 000 000'B
end  module  INTENSITIES

module  Directory_Type
   type, public                    :: Image_File_Directory         !! A TIFF Image File Directory.
      integer                      :: Tag_ID
      integer                      :: Data_Type
      integer                      :: Counter
      integer                      :: Offset
   end  type  Image_File_Directory
end  module  Directory_Type

module  RASTER

   character (len=1), dimension(0:1023, 0:1023), public :: Pixel_Image !! Holds the bit-image color picture.
   integer, parameter, public      :: Color_Green = 2, Color_Red = 4, &
                                      Color_Tiff_Red = 1, Color_Tiff_Green = 4
   integer, parameter, public      :: h = 5       !! The step size (length of line to be drawn).

end  module  RASTER

module  GROUP
   use  RASTER
   implicit none
   private
   public :: SIERPINSKI, MOVE, FILL, DRAW_LINE, MAKE_TIFF_FILE, WRITE_BUFFER, MAKE_TAG

contains

!! This procedure draws a Sierpinski curve of order <Order>, one-quarter of the diagram in Figure 2.11.
recursive  subroutine  SIERPINSKI (Direction, Order)
   !! INCOMING: Direction = 0, 1, 2, or 3, representing North, East, South, West respectively,
   !!                       and specifying the direction in which to draw a line of length h pixels.
   !!           Order     = the order (or size) of the Sierpinski curves to be drawn.

   use  COMPASS
   integer, intent (in)            :: Direction, Order
   integer                         :: Dsucc, &    !! A successor to Direction.
                                      Dpred       !! A predecessor to Direction.

   if (Order <= 0) then
      return
   end  if

   if (Direction == West) then
      Dsucc = North
   else
      Dsucc = Direction + 1
   end  if

   if (Direction == North) then
      Dpred = West
   else
      Dpred = Direction - 1
   end  if

   call SIERPINSKI (Direction, Order-1)
                           call MOVE (Direction, Dsucc, h)
   call SIERPINSKI (Dsucc, Order-1)
                           call MOVE (Direction, Direction, 2*h)
   call SIERPINSKI (Dpred, Order-1)
                           call MOVE (Direction, Dpred, h)
   call SIERPINSKI (Direction, Order-1)

end  subroutine  SIERPINSKI

!! This procedure translates the directions North, East, South, and West to specific
!! increments +h or -h in the X and/or Y directions.  It then calls the procedure DRAW_LINE
!! to draw a straight line from the current point (X0, Y0) to the new point (X0+Dx, Y0+Dy).
subroutine  MOVE (X_Direction, Y_Direction, h)
   !! INCOMING: X_Direction = )-are directions (North, East, South, West) that define the direction
   !!           Y_Direction = )-of the next line to be drawn.  When combined, they produce
   !!                         )-one of the eight movements N, NE, E, SE, S, SW, W, NW.
   !!           h           = the number of units to move in the required direction.
   use  COMPASS
   integer, intent (in)            :: X_Direction, Y_Direction
   integer, intent (in)            :: h

   integer, save                   :: X0 = 10, Y0 = 10
   integer                         :: Dx, Dy      !! will hold small increments in the X and Y
                                                  !! directions, respectively, derived from X_direction
                                                  !! & Y_Direction.
                                                  !! X0 & Y0 hold the co-ordinates of the point to
                                                  !! which the previous line was drawn.  When the
                                                  !! next line is drawn, it will start from this point.

   Dx = 0
   Dy = 0
   select case (X_Direction)
      case (North)
         Dy = h
      case (East)
         Dx = h                                   !! Converts East to an increment h in the X direction.
      case (South)
         Dy = -h
      case (West)
         Dx = -h                                  !! Converts West to an increment -h in the X direction.
   end  select

   select case (Y_Direction)
      case (North)
         Dy = h                                   !! Converts North to an increment h in the Y direction.
      case (East)
         Dx = h
      case (South)
         Dy = -h                                  !! Converts South to an increment -h in the Y direction.
      case (West)
         Dx = -h
   end  select

   call DRAW_LINE (X0, Y0, X0+Dx, Y0+Dy)
!! call DRAW_LINE@ (X0, Y0, X0+Dx, Y0+Dy, Color_Green)   !! For Salford Graphics

   X0 = X0 + Dx                                   !! Record the new position . . .
   Y0 = Y0 + Dy                                   !! . . . of the co-ordinates,
                                                  !! which become the starting-point of the next line.
end  subroutine  MOVE

!! This procedure fills in the interior of the Sierpinski curves with solid color.
!! The procedure first checks that the pixel is not already colored.  If it is,
!! the procedure terminates.  If the pixel is not colored, it is colored.
!! The procedure then looks in turn at the nearest pixel to the north, west,
!! east, and south.  If the pixel is not colored, the FILL procedure is recursively
!! called with the co-ordinates of that pixel.  Whether or not that pixel was originally
!! colored, the process is repeated with the remaining adjacent pixels.
!! The procedure then terminates.
!! Remove the comment marks at the beginning of Fortran statements for the Salford compiler.
recursive  subroutine  FILL (X, Y, Color)
   !! INCOMING: X = the X co-ordinate of the pixel to be "filled in";
   !!       Y = the Y co-ordinate of the pixel to be "filled in";
   !!       Color = the color of the pixel to be "filled in".

   integer, intent (in)            :: X, Y, Color

   integer                         :: Get_Color

   !! Section to ensure that we stay within the area of the graphics image.
   if (X <= lbound(Pixel_Image, 2) .or. X >= ubound(Pixel_Image, 2) ) then
      return
   end  if
   if (Y <= lbound(Pixel_Image, 1) .or. Y >= ubound(Pixel_Image, 1) ) then
      return
   end  if

   Get_Color = ichar (Pixel_Image(Y, X) )         !! Find out the color at the point (X, Y)
!! call GET_PIXEL@ (X, Y, Get_Color)              !! Find out the color at the point (X, Y)
!! Get_Color = iand (Get_Color, 15)               !! Extract the color from the low-order 4 bits.
   if (Get_Color == Color_Tiff_Green .or. Get_Color == Color) then
      return                                      !! Pixel (X, Y) already has a color.
   end  if

!! call SET_PIXEL@ (X, Y, Color)                  !! Color the pixel.
   Pixel_Image(Y, X) = char(Color_Tiff_Red)       !! Red, for TIFF file.

   Get_Color = ichar (Pixel_Image(Y-1, X) )       !! Look at the pixel to the north.
!! call GET_PIXEL@ (X, Y-1, Get_Color)            !! Look at the pixel to the north.
!! Get_Color = iand (Get_Color, 15)               !! Extract the color from the low-order 4 bits.

   if (Get_Color /= Color_Tiff_Green .and. Get_Color /= Color) then
      call FILL (X, Y-1, Color)                   !! Go North.
   end  if

   Get_Color = ichar (Pixel_Image(Y, X-1) )       !! Look at the pixel to the west.
!! call GET_PIXEL@ (X-1, Y, Get_Color)            !! Look at the pixel to the west.
!! Get_Color = iand (Get_Color, 15)               !! Extract the color from the low-order 4 bits.

   if (Get_Color /= Color_Tiff_Green .and. Get_Color /= Color) then
      call FILL (X-1, Y, Color)                   !! Go West, young man.
   end  if

   Get_Color = ichar (Pixel_Image(Y, X+1) )       !! Look at the pixel to the east.
!! call GET_PIXEL@ (X+1, Y, Get_Color)            !! Look at the pixel to the east.
!! Get_Color = iand (Get_Color, 15)               !! Extract the color from the low-order 4 bits.

   if (Get_Color /= Color_Tiff_Green .and. Get_Color /= Color) then
      call FILL (X+1, Y, Color)                   !! Go East.
   end  if

   Get_Color = ichar (Pixel_Image(Y+1, X) )       !! Look at the pixel to the south.
!! call GET_PIXEL@ (X, Y+1, Get_Color)            !! Look at the pixel to the south.
!! Get_Color = iand (Get_Color, 15)               !! Extract the color from the low-order 4 bits.

   if (Get_Color /= Color_Tiff_Green .and. Get_Color /= Color) then
      call FILL (X, Y+1, Color)                   !! Go South.
   end  if

end  subroutine  FILL

!! This special version of DRAW_LINE draws color lines.
!! Assumes that one pixel is stored per byte.
!! This subroutine draws a straight line from the starting point (Xs,Ys) to the
!! finishing point (Xf,Yf).  It also performs scaling of the artwork according to
!! the value of Scaling, and gives it an Offset, for centering on the page.
subroutine  DRAW_LINE (Xs, Ys, Xf, Yf)
      !! INCOMING: (Xs, Ys) are the co-ordinates of the starting point;
      !!       (Xf, Yf) are the co-ordinates of the finish point.

      integer, intent (in)         :: Xs, Ys,  &
                                      Xf, Yf

      integer, parameter           :: Scaling = 1
      integer, parameter           :: Offset = 0

      real                         :: Row, Col
      integer                      :: X_Increment,   &
                                      Y_Increment
      if (Ys > Yf) then                           !! "Y_Increment" is 1 Pixel in the Y direction.
         Y_Increment = -1
      else
         Y_Increment = 1
      end  if
      if (Xs > Xf) then                           !! "X_Increment" is 1 Pixel in the X direction.
         X_Increment = -1
      else
         X_Increment = 1
      end  if

      if (int(Xf*Scaling) == int(Xs*Scaling) ) then   !! Only the Y value changes.
         Col = Xs*Scaling + Offset
         Row = Ys*Scaling
!!       do Row = Ys*Scaling, Yf*Scaling, Y_Increment
         do
            if (Y_Increment > 0) then
               if (Row > Yf*Scaling) then
                  exit
               end  if
            else
               if (Row < Yf*Scaling) then
                  exit
               end  if
            end  if

            Pixel_Image(int(Row), int(Col)+Offset) = char(Color_Tiff_Green)
            Row = Row + Y_Increment
         end  do
      else if (abs(Xs-Xf) >= abs(Ys-Yf) ) then    !! The slope of the line is <= 45 degrees.
                                                  !! The X and Y values change, or just X changes.
         Col = Xs*Scaling
!!       do Col = Xs*Scaling, Xf*Scaling, X_Increment
         do
            if (X_Increment > 0) then
               if (Col > Xf*Scaling) then
                  exit
               end  if
            else
               if (Col < Xf*Scaling) then
                  exit
               end  if
            end  if
            Row = (Yf-Ys)*(Col-Xs*Scaling)/(Xf - Xs) + Ys*Scaling

            Pixel_Image(int(Row), int(Col)+Offset) = char(Color_Tiff_Green)
            Col = Col + X_Increment
         end  do
      else                                        !! The slope of the line is > 45 degrees.
         Row = Ys*Scaling
!!       do Row = Ys*Scaling, Yf*Scaling, Y_Increment
         do
            if (Y_Increment > 0) then
               if (Row > Yf*Scaling) then
                  exit
               end  if
            else
               if (Row < Yf*Scaling) then
                  exit
               end  if
            end  if
            Col = (Xf-Xs)*(Row-Ys*Scaling)/(Yf - Ys) + Xs*Scaling

            Pixel_Image(int(Row), int(Col)+Offset) = char(Color_Tiff_Green)
            Row = Row + Y_Increment
         end  do
      end  if

end  subroutine  DRAW_LINE

   !! This subroutine is used by MAKE_TIFF_FILE to check whether the buffer
   !! is full.  When the buffer is full, it is written to the graphics file.
   subroutine WRITE_BUFFER (Rec, Rec_No, I)
      integer, dimension ( : ), intent (in)  :: Rec
      integer, intent (in out)     :: Rec_No, I

      I = I + 1
      if (I > 32) then
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Rec
         I = 1
      end  if
   end  subroutine  WRITE_BUFFER

   !! This subroutine makes a TIFF tag at position "P" in IFD.
   subroutine  MAKE_TAG (P, IFD, TID, DT, Cnt, Off)
      use  Directory_Type
      type (Image_File_Directory), dimension (0: ), intent (in out)  :: IFD
      integer, intent (in)         :: P, TID, DT, Cnt, Off

      IFD(P)%Tag_ID = TID
               IFD(P)%Data_Type = DT
                        IFD(P)%Counter = Cnt
                              IFD(P)%Offset = Off

   end  subroutine  MAKE_TAG

    !! This procedure writes out a TIFF file containing a color image.
    subroutine  MAKE_TIFF_FILE ( )
!!    use  GLOBAL
      use  INTENSITIES
      use  Directory_Type

      type (Image_File_Directory), dimension (0:17) :: IFD

      integer, dimension (32)             :: Rec
      character (len=2), dimension (0:63) :: Colors
      character (len=128)                 :: Line
      character (len=512)                 :: Pixel_Line

      character (len=2), dimension (0:47), parameter  :: Color_Palette = (/   &
               Full, Full, Full,        &          !!  0 = White.
               Full, Nil,  Nil,         &          !!  1 = Red.
               Full, Full, Nil,         &          !!  2 = Yellow.
               Full, Nil,  Full,        &          !!  3 = Purple.
               Tqtr, Full, Tqtr,        &          !!  4 = Light Green
               Nil,  Full, Full,        &          !!  5 = Light blue.
               Nil,  Nil,  Full,        &          !!  6 = Dark blue.
               Full, Tqtr, Nil,         &          !!  7 = Orange.
               Nil,  Full, Nil,         &          !!  8 = Green.
               Full, Half, Full,        &          !!  9 = Pale purple.
               Full, Full, Half,        &          !! 10 = Lemon.
               Nil,  Half, Full,        &          !! 11 = Medium blue.
               Full, Special, Nil,      &          !! 12 = Light brown.
               Nil,  Nil,  Nil,         &          !! 13 = Black.
               Nil,  Nil,  Nil,         &          !! 14 = Black.
               Nil,  Nil,  Nil /)                  !! 15 = Black.

      character (len=2), parameter :: Empty = char(0) // char(0)
      integer                      :: I, J, K, N, Row, Col
      integer                      :: Rec_No
      character (len=128)          :: CP

      print *, "Preparing to write TIFF file."

      open (unit=9, file="TIFF_FILE.TIF", access="direct", action="write",    &
           form="unformatted", status="new", recl=128)

      !! Create the Image File Header (IFH) (8 bytes).
      !! Contains: Byte Ordering:  Intel byte ordering (2 bytes of "II");
      !!       Identification: 2 bytes, integer value 42;
      !!       Address of the first IFD (4 bytes).
      IFD(0)%Tag_ID = ichar("I")*256+ichar("I")
                        IFD(0)%Data_Type = 42
                              IFD(0)%Counter = 10
                                                  !! Header.

      !! Create the Image File Directory (IFD) Header:
      !! 2 bytes: Number of entries in the IFD.
      !! 12 bytes: For each Tag in the IFD: Tag_ID (2 bytes),
      !!          Data Type (2 bytes)
      !!          Number of items (4 bytes)
      !!          Address of item (4 bytes)
      !!             (If the No. of items is 1,
      !!             and the data type is 4 or less,
      !!             this field contains the actual data.)
      !! 4 bytes: Trailer: A link to other IFDs, or zero if there is no more.
      IFD(0)%Offset = 15*65536                    !! No. of IFD tags*65536.

      !! Following should be in increasing order of TAG ID.
      call MAKE_TAG (1, IFD, 254, 4, 1, 1)        !! Tag for new sub-file type.
      call MAKE_TAG (2, IFD, 256, 3, 1, 1024)     !! Tag for LINE WIDTH in pixels.
      call MAKE_TAG (3, IFD, 257, 3, 1, 1024)     !! Tag for IMAGE DEPTH in pixels.
      call MAKE_TAG (4, IFD, 258, 3, 1, 4)        !! Tag for Number of Bits per Sample.
      call MAKE_TAG (5, IFD, 259, 3, 1, 1)        !! Tag for Compression: 1=uncompressed.
      call MAKE_TAG (6, IFD, 262, 3, 1, 3)        !! Photometric Interpretation:
                                                  !! 0 means render pixel 0 as white, pixel 1 as black
                                                  !! 1 means render pixel 0 as black, pixel 1 as white
                                                  !! 3 means that the image is color.
      call MAKE_TAG (7, IFD, 273, 4, 1024, 384)   !! Strip Offsets: This tag points to a table of
                                                  !! addresses.  The table contains 1024 entries, one
                                                  !! for each row of the image (and each entry
                                                  !! occupies 4 bytes).
      call MAKE_TAG (8, IFD, 277, 3, 1, 1)        !! Tag for Samples per pixel.
      call MAKE_TAG (9, IFD, 278, 3, 1, 1)        !! Tag for Rows per strip.
      call MAKE_TAG (10, IFD, 279, 3, 1, 128)     !! Tag for Strip byte count.
      call MAKE_TAG (11, IFD, 282, 5, 1, 196)     !! Tag for address of X resolution.
      call MAKE_TAG (12, IFD, 283, 5, 1, 204)     !! Tag for address of Y resolution.
      call MAKE_TAG (13, IFD, 284, 3, 1, 1)       !! Tag for Planar configuration.
      call MAKE_TAG (14, IFD, 296, 3, 1, 3)       !! Tag for Resolution:
                                                  !! 1 = undefined, 2 = cm, 3 = inches.
      call MAKE_TAG (15, IFD, 320, 3, 48, 256)    !! Tag Points to the Color Palette.

      J = 16
      IFD(J)%Tag_ID = 0
               IFD(J)%Data_Type = 0               !! 4 bytes of zeros to terminate the IFD chain.

      !! The next few entries are parameters of the picture.
      IFD(J)%Counter = 300
               IFD(J)%Offset = 1                  !! X resolution.
      J = J + 1
      IFD(J)%Tag_ID = 300
               IFD(J)%Data_Type = 0
                        IFD(J)%Counter = 1        !! Y resolution.
      IFD(J)%Offset = 0                           !! unused area.

      N = J                                       !! the total number of records.

      !! Write the Header, the IFD tags, & the trailer, to the graphic file.
      !! Each IFD tag is 12 bytes (2 x 16-bit integers & 2 x 32-bit integers).
      !! Records are 128 bytes.  32-bit integers are copied from structure
      !! IFD until the buffer REC is full, then they are written out.
      Rec_No = 0
      I = 1
      do J = 0, N
         Rec(I) = IFD(J)%Tag_ID + IFD(J)%Data_Type*65536
         call WRITE_BUFFER (Rec, Rec_No, I)

         Rec(I) = IFD(J)%Counter
         call WRITE_BUFFER (Rec, Rec_No, I)

         Rec(I) = IFD(J)%Offset
         call WRITE_BUFFER (Rec, Rec_No, I)
      end  do

      if (I > 1) then                             !! Write the remaining (short) record.
         Rec(I:32) = 0                            !! Zero fill the remainder of the record.
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Rec
      end  if

      !! SECTION TO WRITE OUT THE COLOR PALETTE.
      Colors( 0:15) = Color_Palette(0:45:3)       !! Collect the reds;
      Colors(16:31) = Color_Palette(1:46:3)       !! Collect the greens;
      Colors(32:47) = Color_Palette(2:47:3)       !! Collect the blues.
      Colors(48:63) = Empty

!!      do K = 0,47
!!         CP(K+K+1:K+K+2) = Colors(K)(1:2)
!!      end  do
!!      CP(97:128) = repeat(char(0), 32)
!!      Rec_No = Rec_No + 1
!!      write (unit=9, rec=Rec_No) CP

      Rec_No = Rec_No + 1
      write (unit=9, rec=Rec_No) Colors
!!      write (unit=9, rec=Rec_No) (/ Color_Palette(0:45:3),    &
!!                                    Color_palette(1:46:3),    &
!!                                    Color_Palette(2:47:3),    &
!!                                    (char(0) // char(0), I = 1,16) /)

      !! SECTION TO CREATE THE StripAddressArray, CONTAINING 1024 ADDRESSES,
      !! EACH BEING A POINTER TO ONE PIXEL LINE OF THE IMAGE.
      K = 1
      do Row = 0, 1023
         Rec(K) = 4480 + Row*512
         K = K + 1
         if (K > 32) then
            Rec_No = Rec_No + 1
            write (unit=9, rec=Rec_No) Rec
            K = 1
         end  if
      end  do

      !! SECTION TO WRITE OUT THE PIXEL IMAGES
      do Row = 0, 1023
         !! Prepare one scan line.
         do Col = 0, 1023, 2                      !! Pack two pixels (4 bits each) into a byte.
            Pixel_Line(Col/2+1:Col/2+1) = char (      &
               ichar(Pixel_Image(Row, Col))*16 + ichar(Pixel_Image(Row, Col+1)) )
         end  do
         !! Write out a line.
         do K = 1, 512, 128                       !! The line is written in four chunks.
            Line = Pixel_Line(K:K+127)
            Rec_No = Rec_No + 1
            write (unit=9, rec=Rec_No) Line
         end  do
      end  do

      print *, "TIFF file created."

   end  subroutine  MAKE_TIFF_FILE

end  module  GROUP

program  DISPLAY_SIERPINSKI
   use  COMPASS
   use  GROUP
   use  RASTER
   implicit none
   integer                         :: Order
   character (len=1)               :: Keystroke

   Pixel_Image = char(0)

   print *, "What order Sierpinski curve do you want?"
   read *, Order

!! call VGA@                                      !! Establish graphics mode.

   call SIERPINSKI (North, Order)
                           call MOVE (North, East, h)
   call SIERPINSKI (East, Order)
                           call MOVE (South, East, h)
   call SIERPINSKI (South, Order)
                           call MOVE (South, West, h)
   call SIERPINSKI (West, Order)
                           call MOVE (North, West, h)

   call FILL (15, 10, Color_Tiff_Red)             !! To fill in the Sierpinski curve with a bright color.

!! call GET_KEY@ (Keystroke)                      !! Pause for the user to admire the view.

!! call  TEXT_MODE@                               !! Switch the video monitor back to text mode.

   call MAKE_TIFF_FILE ( )

end  program  DISPLAY_SIERPINSKI

!! Figure 2.16  Program to draw a Sierpiski curve
