module GROUP
   implicit  none
   public :: QUICK_SORT, SWAP

contains

recursive  subroutine  QUICK_SORT (A, Left, Right)
   !! This subroutine implements Quicksort by C. A. R. Hoare.
   !! INCOMING: A     = an array whose elements are to be sorted;
   !!           Left  = a pointer to the left-hand end of the partition to be sorted;
   !!           Right = a pointer to the right-hand end of the partition to be sorted;
   !! OUTGOING: A     = an array whose elements A(Left) through A(Right) are sorted.

   integer, dimension ( : ), intent (in out) :: A
   integer, intent (in)            :: Left, Right

   integer                         :: Mid
   integer                         :: Ref         !! To hold the Reference Element.
   integer                         :: L, R        !! L = a pointer used when searching
                                                  !! from the left.
                                                  !! R = a pointer, used when searching
                                                  !! from the right.

   if (Right <= Left) then                        !! The partition is empty or contains one element.
      return                                      !! There's no work to do.
   end  if

   !! Section to select the median of three elements, and to move it to the left.
   Mid = (Left + Right)/2
   if (A(Mid) > A(Right)) then
      call SWAP (A(Mid), A(Right))
   end if
   if (Left+1 == Right) then                      !! There are 2 elements in the partition,
      return                                      !! and they are in order, so quit.
   end  if
   if (A(Left) > A(Mid)) then
      call SWAP (A(Left), A(Mid))
   end  if
   if (A(Mid) > A(Right)) then
      call SWAP (A(Mid), A(Right))
   end  if
   if (Left+2 == Right) then                      !! There are 3 elements in the partition,
      return                                      !! and they are in order, so quit.
   end  if
   call SWAP (A(Left), A(Mid))                    !! The median is swapped to the left end.
   Ref = A(Left)                                  !! Select the Reference Element.

   L = Left
   R = Right + 1

   !! Partition the elements into three groups.
   do
      if (L >= R) then
         exit
      end  if
      do L = L + 1, R - 1                         !! Scan from the left for an element that
         if (A(L) > Ref) then                     !! is larger than the Reference.
            exit
         end  if
      end  do
      do                                          !! Scan from the right for an element that
         R = R - 1
         if (A(R) <= Ref) then                    !! is less than or equal to the Reference.
            exit
         end  if
      end  do

      if (L < R) then                             !! Swap two elements that are in the
                                                  !! wrong partitions.
         call SWAP (A(L), A(R))
      end  if
   end  do
                                                  !! Partitioning is complete.
   if (Left < R) then                             !! Swap the Reference Element into its
                                                  !! final position R in the array.
      A(Left) = A(R)
      A(R)    = Ref
   end  if
   !! At this point, A(R) is in its correct position in the list.  Elements A(Left) to A(R-1)
   !! are less than or equal to A(R), and elements A(R+1) to A(Right) are greater then A(R).

   do L = R-1, Left+1, -1                         !! Pass over any elements that
      if (A(L) /= Ref) then                       !! are equal to Ref.
         exit
      end  if
   end  do

   call QUICK_SORT (A, Left, L)                   !! Partition the left segment.
                                                  !! The element at R is in its correct position.
   call QUICK_SORT (A, R+1, Right)                !! Partition the right segment.

end  subroutine  QUICK_SORT

!!  insert the procedure SWAP from Figure 2.8 

!! This subroutine swaps the element Left_Element with Right_Element.
subroutine SWAP (Left_Element, Right_Element)
   integer, intent (in out)        :: Left_Element, Right_Element

   integer                         :: Temp

   Temp     = Left_Element
   Left_Element   = Right_Element
   Right_Element  = Temp
end  subroutine  SWAP

!! Figure 2.9  An improved QuickSort algorithm

end  module  GROUP
