module  GROUP
   implicit  none
   public :: REBALANCE_TREE

contains

!! This procedure rebalances and re-colors the nodes of a Red-Black sorted binary tree,
!! following an insertion.  The node pointed at by X has just been inserted into the tree,
!! and is the node where one of the 4 properties of Red-Black trees may have been violated.
!! After a breach of Rule 1 has been rectified, X is adjusted to point at its grandparent,
!! where tests for a breach of the rules is again carried out, and so on for each grandparent
!! in turn.
!! If Rules 2 or 3 are breached, one or two rotations of a branch of the tree are carried out
!! (about the current node X), and the procedure terminates.
!! When this procedure terminates, the root node is black, and the tree is quasi-balanced,
!! and the four properties of Red-Black trees are satisfied:
!!
!! Rule 1. Every node is either red or black;
!! Rule 2. Every null pointer is considered to be pointing to an imaginary black node;
!! Rule 3. If a node is red, then both its children are black; and
!! Rule 4. Every direct path from the root to a leaf contains the same number of black nodes.

subroutine  REBALANCE_TREE (Current_Node)
      !! INCOMING: Current_Node = A pointer to a (red) node that has been inserted in the tree.
      use  Node_Definition
      use  GLOBAL, only: Root
      type (Node), pointer         :: Current_Node

      type (Node), pointer         :: X, Y
      logical                      :: Red_Uncle
      logical                      :: Iterating
      integer                      :: Direction

      X => Current_Node

      do                                          !! We execute this loop, and re-execute it,
                                                  !! when X and its parent are both red.
         Iterating = .not. associated (X, Root)   !! Cannot iterate when we are at the root.
         if (Iterating) then                      !! There must be a parent . . . .
            Iterating = X%Parent%Color .eqv. Red
         end  if
         if (Iterating) then                      !! . . . and there must be a grandparent.
            Iterating = associated (X%Parent%Parent)
         end  if
         if (.not. Iterating) then                !! We enter this loop when X and its parent are both red.
            exit
         end  if

         if ( associated (X%Parent, X%Parent%Parent%D(Left)%Aux ) ) then
                                                  !! The parent is a left node of X's grandparent.
            Direction = Right
         else
            Direction = Left
         end  if
         Y => X%Parent%Parent%D(Direction)%Aux    !! Get the address of X's uncle.
         Red_Uncle = associated (Y)
         if (Red_Uncle) then
            Red_Uncle = Y%Color .eqv. Red
         end  if
         if (Red_Uncle) then                      !! CASE 1.  There is an uncle.  X and its parent and
                                                  !! uncle are all red.  Fix violation of Rule 3.
            X%Parent%Color = Black                !! The parent must be black.
            Y%Color = Black                       !! The uncle must be black.
            X%Parent%Parent%Color = Red
                                                  !! The grandparent must be red.
            X => X%Parent%Parent                  !! Move 2 places up the tree, to the grandparent.
         else                                     !! The uncle is black, or is non-existent.
            if ( associated (X, X%Parent%D(Direction)%Aux ) ) then !! CASE 2.
               X => X%Parent                      !! Move up the tree.
               if (Direction == Right) then
                  call ROTATE_LEFT (X)
               else
                  call ROTATE_RIGHT (X)
               end  if
            end  if
            !! CASE 3.
            X%Parent%Color = Black
            X%Parent%Parent%Color = Red
            if (Direction == Right) then
               call ROTATE_RIGHT (X%Parent%Parent)
            else
               call ROTATE_LEFT (X%Parent%Parent)
            end  if
         end  if
      end  do

      Root%Color = Black                          !! Ensure that the root is black.

end  subroutine  REBALANCE_TREE

!! Figure 3.15  Algorithm for rebalancing a Red-Black binary tree, following insertion

end  module  GROUP
