module  GLOBAL

   use  Node_Definition
   private
   type (Node), pointer, public    :: Ptr
   character (len=1), parameter, public   :: ESC = char(27) !! The ASCII ESCape character.
   character (len=5), parameter, public   :: Rosso = ESC // "[31m"
   character (len=5), parameter, public   :: Verde = ESC // "[36m"
   character (len=5), parameter, public   :: White = ESC // "[37m"
   integer, public                 :: Current_Depth, &
                                      Max_Depth
                                                  !! The maximum depth of the tree (levels of nodes).

end  module  GLOBAL

module GROUP
   implicit none
   public                          :: DISPLAY_TREE, SET_COLOR, VISIT

contains

!! This subroutine displays a tree in graphic form, showing the nodes in their actual
!! positions.  It is intended for a screen display, but the screen parameters can be altered
!! readily enough for a wide screen or printer.  For each node, this subroutine displays the
!! value in its appropriate color (red or black), along with the letters R or B (R=red,
!! B = black), in case you don't have a color screen.  This routine displays 5 levels;
!! at the fifth level, 16 elements are displayed on one screen line of 80 columns.
!!
!! Method: An in-order traversal of the tree is first performed, in order to determine the
!! longest branch (that is, the greatest depth of the tree).  This value is used to
!!  determine the number of nodes as if all branches were of this length.

subroutine  DISPLAY_TREE (Root)
      use  Node_Definition
      use  GLOBAL
      type (Node), pointer :: Root

      integer              :: Level,           &  !! Used to indicate the current level of a node
                                                  !! during an excursion from the root to a node
                                                  !! to be printed.
                              Nodes,           &  !! The maximum number of nodes at level LEVEL.
                              Depth,           &  !! The number of nodes from here to the root.
                              Spaces, Gap, J,  &
                              Node_No, Trip,   &
                              Mask
      character (len=1), dimension (0:1), parameter   :: Palette = (/ "B", "R" /)
      character (len=10)           :: Layout      !! For preparing a dynamic format specification.

      Current_Depth = 0
      Max_Depth = 0
      call VISIT(Root)
      Max_Depth = min(Max_Depth, 5)

      Level = 0
      Nodes = 1
      Ptr => Root
      call SET_COLOR ( )
      write (unit=*, fmt="( / tr35, i3)", advance="no" ) Ptr%Item
      if (Ptr%Color .eqv. Red) then
         J = 1
      else
         J = 0
      end  if
      write (unit=*, fmt="(a)")  Palette(J)
      write (unit=*, fmt="(tr37, a)" )  "/\"
      write (unit=*, fmt="(tr36, a)" )  "/  \"

      Depth = 1
      Spaces = 80

      do                                          !! Visit all the nodes to the specified depth.
         if (Depth >= Max_Depth) then
            exit
         end  if
         Spaces = Spaces/2                        !! Adjust the spacing between nodes.
         Gap = Spaces/2                           !! The initial spacing is half that between nodes.
         Nodes = Nodes * 2
         Depth = Depth + 1                        !! The level of the nodes being displayed.
         print *

         do Node_No = 0, Nodes-1                  !! Visit all the nodes at a given level.
            Ptr => Root                           !! Start from the root.
            Mask = 1
            do J = 1, Depth-2
               Mask = Mask * 2
            end  do

            do Trip = 1, Depth-1
               if (.not. associated (Ptr) ) then
                  exit
               end  if
                                                  !! Travel from the root to a node at the given depth
               if (iand(Node_No, Mask) /= 0) then !! Take the right branch.
                  Ptr => Ptr%Right
               else                               !! Take the left branch.
                  Ptr => Ptr%Left
               end  if
               Mask = Mask/2
            end  do

            if (associated(Ptr) ) then
               call SET_COLOR ( )
               write (unit=Layout, fmt="(a, i3.3, a)" )   "(tr", Gap-4, ",i3)"
               if (Gap < 5) then
                  Layout = "(i3)"
               end  if
               write (unit=*, fmt=Layout, advance="no") Ptr%Item
               !! Comment the next line if your PC or terminal lacks color.
               write (unit=*, fmt="(a)", advance="no" ) " "
               !! Un-comment the next 6 lines if your PC or terminal lacks color.
               !! if (Ptr%Color .eqv. Red) then
               !! J = 1
               !! else
               !! J = 0
               !! end  if
               !! write (unit=*, fmt="(a)", advance="no" ) Palette(J)
            else
               write (unit=Layout, fmt="(a, i3.3, a)" )   "(tr", Gap-4, ",2a)"
               if (Gap < 5) then
                  Layout = "(2a)"
               end  if
               !! Comment the next line if your PC or terminal lacks color.
               write (unit=*, fmt=Layout, advance="no") White, " -  "
               !! Un-comment the next line if your PC or terminal lacks color.
               !! write (unit=*, fmt=Layout, advance="no") " -  "
            end  if
            Gap = Spaces                          !! Assume normal spacing between nodes.
            Mask = Mask/2
         end  do

         !! Comment the next line if your PC or terminal lacks color.
         write (unit=*, fmt="(a)" ) White
         !! Un-comment the next line if your PC or terminal lacks color.
         write (unit=*, fmt=*)
         Gap = Spaces/2
         if (Depth < Max_Depth) then
            do Node_No = 0, Nodes-1
               write (unit=Layout, fmt="(a, i3.3, a)" )   "(tr", Gap-2, ", a)"
               if (Gap < 3) then
                  Layout = "(a)"
               end  if
               write (unit=*, fmt=Layout, advance="no") "/ \"
               Gap = Spaces
            end  do
         end  if

         write (unit=*, fmt="(a)" )
         write (unit=*, fmt="(a)", advance="no" ) " "
         Gap = Spaces/2
         if (Depth < Max_Depth) then
            do Node_No = 0, Nodes-1
               write (unit=Layout, fmt="(a, i3.3, a)" )   "(tr", Gap-4, ", a)"
               if (Gap < 5) then
                  Layout = "(a)"
               end  if
               write (unit=*, fmt=Layout, advance="no") "/  \"
               Gap = Spaces
            end  do
         end  if
      end  do

end  subroutine  DISPLAY_TREE

!! This procedure changes the display color to either red or green.
subroutine  SET_COLOR ( )
      use  Node_Definition
      use  GLOBAL

      if (Ptr%Color .eqv. Red) then
         write (unit=*, fmt="(a)", advance="no" ) Rosso
      else
         write (unit=*, fmt="(a)", advance="no" ) Verde
      end  if
end  subroutine  SET_COLOR

!! This procedure finds the greatest depth in the tree.  It performs an in-order traversal.
recursive  subroutine  VISIT (Current)
      use  Node_Definition
      use  GLOBAL
      type (Node), pointer         :: Current

      Current_Depth = Current_Depth + 1
      if (Current_Depth > Max_Depth) then
         Max_Depth = Current_Depth
      end  if
      if (associated (Current%Left) ) then        !! take the left-hand path . . .
         call VISIT (Current%Left)
         Current_Depth = Current_Depth - 1
      end  if
      if (associated (Current%Right) ) then       !! take the right-hand path.
         call VISIT (Current%Right)
         Current_Depth = Current_Depth - 1
      end  if

end  subroutine  VISIT

end  module  GROUP

!! Figure 3.16  Displaying or printing a tree
