module  GROUP
   implicit  none
   public :: DISPLAY_SIDEWAYS, VISIT

contains

!! This procedure prints/displays a binary tree sideways.
!! Method: An in-order traversal of the tree is first performed, in order to determine the
!! longest branch (that is, the greatest depth of the tree).  This value is used to determine
!! the number of nodes as if all branches were of this length.
!! For each node, an excursion is taken from the root to the node to be printed.  The route is
!! determined by the number of the current node (nodes number from 1 to 2**Max_Depth-1),
!! and is decoded by Mask, initially set to 2**Max_Depth/2.  Each bit in the binary
!! representation of the node's number corresponds to a fork in the tree.  A zero bit causes a
!! right path to be taken, while a one-bit causes a left path to be taken.  When a node is
!! absent from the tree, an empty line is written out so as to maintain the geometry of the
!! tree.

subroutine  DISPLAY_SIDEWAYS (Root)
      use Node_Definition
      type (Node), pointer         :: Root

      type (Node), pointer    :: Ptr
      integer                 :: Max_Nodes,     &
                                 Node_No,       &
                                 Level,         & !! Used to indicate the current level of a node
                                                                                !! during an excursion from the root to a node
                                                                                !! to be printed.
                                 Mask,          & !! is Ariadne's thread (the key to decode the
                                 Dmask,         & !! path to be taken at each fork).
                                 Max_Depth,     & !! The maximum depth of the tree (levels of nodes).
                                 Current_Depth, &
                                 Nlevels
      logical                 :: Printed

      character (len=1), parameter :: ESC = char(27) !! The ASCII ESCape character.
      character (len=3), parameter :: Rosso = ESC // "r1"
      character (len=3), parameter :: Verde = ESC // "r6"
      character (len=10)           :: Layout      !! For preparing a dynamic format specification.

      Current_Depth = 0
      Max_Depth = 0

      call VISIT (Root)                           !! To count the nodes in the tree.

      Max_Nodes = 2**Max_Depth                    !! Determine the depth of the tree.

      !! Section to visit each node of the tree.
      !! Nodes are numbered starting from the right-most leaf, as if doing an in-order traversal.
      do Node_No = 1, Max_Nodes - 1
         Ptr => Root                              !! Begin each journey from the root.
         Mask = Max_Nodes / 2
         Printed = .false.

         !! Section to travel from the root, to the node to be printed.
         do Level = 1, 1000, 1
            if (iand(Node_No, Mask) /= 0) then
                                                                                !! Take the right path.
               Ptr => Ptr%Right
            else                                  !! Take the left path.
               if ((modulo (Node_No, Mask) == 0) .or. (Mask == 1) ) then
                  write (unit=*, fmt=*)
                  Dmask = 2*Mask
                  if (Node_No /= Max_Nodes/2) then
                     write (unit=Layout, fmt="(a, i3, a)" )    &
                           "(tr", 10*(Level-1)+1, ", a)"
                     if (iand(Node_No, Dmask) /= 0) then
                        write (unit=*, fmt=Layout, advance="no" ) "\"
                     else
                        write (unit=*, fmt=Layout, advance="no" ) "/"
                     end  if
                  end  if
                  if (Ptr%Color .eqv. Red) then
                     write (unit=*, fmt="(a)" ) Rosso
                  else
                     write (unit=*, fmt="(a)" ) Verde
                  end  if
                  write (unit=*, fmt="(i3)", advance="no" ) Ptr%Item
                  Printed = .true.
               end  if
               Ptr => Ptr%Left
            end  if
            Mask = Mask / 2
            if ( (Level >= Max_Depth) .or. (.not. associated(Ptr) ) .or. Printed ) then
               exit
            end  if
         end  do

         if (.not. Printed) then                  !! The node did not exist.
            do Nlevels = Level, Max_Depth
               print *                            !! Leave a blank line so as to preserve the
            end  do                               !! structure of the tree.
         end  if
      end  do
      print *, Verde

end  subroutine  DISPLAY_SIDEWAYS

!!  insert subroutine VISIT from Figure 3.16 

!! This procedure finds the greatest depth in the tree.  It performs an in-order traversal.
recursive  subroutine  VISIT (Current)
      use  Node_Definition
      use  GLOBAL
      type (Node), pointer         :: Current

      Current_Depth = Current_Depth + 1
      if (Current_Depth > Max_Depth) then
         Max_Depth = Current_Depth
      end  if
      if (associated (Current%Left) ) then        !! take the left-hand path . . .
         call VISIT (Current%Left)
         Current_Depth = Current_Depth - 1
      end  if
      if (associated (Current%Right) ) then       !! take the right-hand path.
         call VISIT (Current%Right)
         Current_Depth = Current_Depth - 1
      end  if

end  subroutine  VISIT

!! Figure 3.17  Printing a binary tree sideways

end  module  GROUP
