module  M19
   implicit  none
   public :: DELETE

contains

!! This subroutine deletes a node from a Red-Black binary tree.
!! There are four main sections.  In the first section, the tree is checked to see if it
!! is empty, and if it is, the subroutine quits.
!! The second section deals with the case where the node to be deleted has two children.
!! The third section deals with the straight-forward case where the node to be deleted
!! has no children.
!! The fourth section deals with the case where the node to be deleted has one child.
subroutine  DELETE (Vanish)
      !! INCOMING: Vanish points at the node to be deleted.
      use  Node_Definition
      use  GROUP

      type (Node), pointer         :: Vanish

      type (Node), pointer         :: Child, X

      X => Vanish
!! Case 1:  the tree is empty.
      if (.not. associated (X) ) then             !! The tree is empty.
         return
      end  if

!! Case 2:  the node has 2 children.
      !! We find the successor node, move the content of that node to the current node X,
      !! and then delete the successor node.
      if (associated (X%Left) .and. associated (X%Right) ) then !! Find the next-largest node.
                                                  !! The next-largest node is found in the right
                                                  !! subtree.
         Child => X%Right
         do
            if (.not. associated (Child%Left) ) then !! Seek the left-most node of the subtree.
               exit
            end  if
            Child => Child%Left
         end  do
         X%Item = Child%Item                      !! Replace the Item to be deleted by its successor.
         !! Child is a node to be deleted.
         X => Child                               !! The successor is now due for deletion.
                                                  !! This will be handled by Case 3 (the node is a
                                                  !! leaf) or by Case 4 (the node has one child).
      end if

      !! Now delete node X.
!! Case 3:  the node has no children.
      if (.not. associated (X%Right) .and. .not. associated (X%Left) ) then
                                                  !! The node is a leaf.
         if (.not. associated (X%Parent) ) then   !! Node X is the root.
            deallocate (X)
            nullify (Root)
            return
         end  if
         if (X%Color .eqv. Black) then            !! If the color of the node to be deleted is black . . .
            call  RESTRUCTURE_AFTER_DELETION (X)
         end  if
         if (associated (X, X%Parent%Right) ) then!! Chop off the leaf X.
            nullify (X%Parent%Right)              !! Remove X which is a right child..
         else
            nullify (X%Parent%Left)               !! Remove X which is a left child.
         end  if
         deallocate (X)
         return
      end  if

!! Case 4:  the node has one child.  First find out which it is, and then splice it out.
      if (.not. associated (X%Right) ) then
         Child => X%Left                          !! It's a left child of X.
      else
         Child => X%Right                         !! It's a right child of X.
      end  if

      if (.not. associated (X%Parent) ) then      !! Node X is the root.
         Root => Child                            !! X's only child becomes the root.
         deallocate (X)                           !! Get rid of X.
         nullify (Root%Parent)                    !! The root can't have a parent.
         Root%Color = Black                       !! Color it black . . .
         return                                   !! . . . and quit.
      else                                        !! Node X is not the root.       !! Splice out the node.
         !! Make the grandparent point at the child.
         if (associated (X, X%Parent%Right) ) then!! Node X is a right child.
            X%Parent%Right => Child
         else                                     !! Node X is a left child.
            X%Parent%Left => Child
         end  if

         Child%Parent => X%Parent                 !! Make the child point up at the grandparent.

      end  if

      !! The next segment will restructure the tree when the deleted node is black.  However,
      !! there is one particular case when the deleted note is black and its child is red.  In this
      !! situation, its child is merely re-colored black, thus restoring the deficiency of one black node.
      !! The following code indirectly achieves this: RESTRUCTURE_AFTER_DELETION is called,
      !! its main loop is not executed, and the subroutine terminates after re-coloring the child black.
      if (X%Color .eqv. Black) then               !! If the color of the node to be deleted is black . . .
         call  RESTRUCTURE_AFTER_DELETION (Child)
      end  if

      deallocate (X)

end  subroutine  DELETE

!! Figure 3.19  Deleting a node in a Red-Black binary tree

end  module  M19
