module  GROUP
   implicit  none
   public :: RESTRUCTURE_AFTER_DELETION, COLOR_OF

contains

!! This procedure re-colors nodes and/or restructures a Red-Black tree following a deletion.
!! This routine is entered when the deleted node was black.  (Node X is the child of the
!! deleted node).
!! The crux is to balance the two subtrees whose roots are node X and its brother.
!! General strategy:
!! (1) Simple case: If the child of the deleted node is red, change it to black and finish.
!! (2) General case:  Either:
!! (2a) Perform a rotation about the parent, so as to bring a node into the subtree, to
!!      compensate for the deleted node.  In this case, color the new node black and finish.
!! (2b) Change a node of the sibling's subtree from black to red, and move up the tree.
!!      In this event, repeat step (2).
subroutine  RESTRUCTURE_AFTER_DELETION (Current)
      !! INCOMING: Current = points at the child of the deleted node.
      use  Node_Definition
      use  GLOBAL

      type (Node), pointer         :: Current

      type (Node), pointer         :: X
      type (Node), pointer         :: Brother

      X => Current

TREE_CLIMBING_LOOP:                 &
      do
         if ( associated(X, Root) .or. (X%Color .eqv. Red)) then
            exit                                  !! Quit when we are at the root, or if the node is red.
         end  if
         !! The crux balances the two subtrees whose roots are at node X and its brother.
         if (associated (X, X%Parent%Left) ) then !! Node X is a left child.
            Brother => X%Parent%Right
            if (Brother%Color .eqv. Red) then
                                                  !! Case 1: Rotate left about X's parent, and re-colo
               Brother%Color = Black
               X%Parent%Color = Red
               call  ROTATE_LEFT (X%Parent)
               Brother => X%Parent%Right
            end  if                               !!  . . . and move on to apply case 2.

            if ( COLOR_OF(Brother%Left) .eqv. Black .and.   &  !! If both children of the
               COLOR_OF(Brother%Right) .eqv. Black) then       !! brother  are black . . .
               !! Case 2: Color the brother red, and move up the tree.
               Brother%Color = Red
               X => X%Parent                      !! Move up the tree.
            else
               if (COLOR_OF (Brother_Right) .eqv. Black) then
                                                  !! Case 3: Rotate right about the brother, and re-color.
                  if (associated(Brother%Left) ) then
                     Brother%Left%Color = Black
                  end  if
                  Brother%Color = Red
                  call  ROTATE_RIGHT (Brother)
                  Brother => X%Parent%Right
               end  if                            !! . . . and go on to Case 4.

               !! Case 4: Rotate left about X's parent, which brings an extra node to
               !! the subtree through X, and which is then colored black.
               Brother%Color = X%Parent%Color
               X%Parent%Color = Black
               Brother%Right%Color = Black
               call  ROTATE_LEFT (X%Parent)       !! Brings a black node to the left
                                                  !! subtree; Left & Right subtrees are balanced.
               exit TREE_CLIMBING_LOOP            !! Quit, rebalancing is complete.
            end  if
         else                                     !! Node X is a right child.
            !! Same as the THEN clause, with "Right" and "Left" interchanged.
            Brother => X%Parent%Left
            if (Brother%Color .eqv. Red) then
                                                  !! Case 1: Rotate left about X's parent, and re-color.
               Brother%Color = Black
               X%Parent%Color = Red
               call  ROTATE_RIGHT (X%Parent)
               Brother => X%Parent%Left
            end  if                               !!  . . . and move on to apply case 2.

            if ( COLOR_OF(Brother%Left) .eqv. Black .and.   &  !! If both children of the
               COLOR_OF(Brother%Right) .eqv. Black) then !! brother  are black . . .
               !! Case 2: Color the brother red, and move up the tree.
               Brother%Color = Red
               X => X%Parent                      !! Move up the tree.
            else
               if (COLOR_OF(Brother%Left) .eqv. Black) then
                                                  !! Case 3: Rotate left about the brother, and re-color.
                  if (associated(Brother%Right) )  then
                     Brother%Right%Color = Black
                  end  if
                  Brother%Color = Red
                  call  ROTATE_LEFT (Brother)
                  Brother => X%Parent%Left
               end  if                            !! . . . and go on to Case 4.

               !! Case 4: Rotate right about X's parent, which brings an extra node to
               !! the subtree through X, and which is then colored black.
               Brother%Color = X%Parent%Color
               X%Parent%Color = Black
               Brother%Left%Color = Black
               call  ROTATE_RIGHT (X%Parent)      !! Brings a black node to the right
                                                  !! subtree; Left & Right subtrees are balanced.
               exit TREE_CLIMBING_LOOP            !! Quit, rebalancing is complete.
            end  if
         end  if

      end  do  TREE_CLIMBING_LOOP

      X%Color = Black                             !! Having encountered a red node, color it black and quit.

end  subroutine  RESTRUCTURE_AFTER_DELETION

!! This function returns the color of the specified node, or black if the node does not exist.
function  COLOR_OF (Node_Ptr) result (Node_Color)
   !! INCOMING: Node_Ptr = a pointer to the node whose color is to be obtained.
   use  Node_Definition

   type (Node), pointer            :: Node_Ptr

   logical                         :: Node_Color

   if (associated (Node_Ptr)) then
      Node_Color = Node_Ptr%Color
   else
      Node_Color = Black
   end  if

end  function  COLOR_OF

!! Figure 3.20  Restructuring a Red-Black binary tree, after a deletion

end  module  GROUP
