!! This program performs an indirect Shell Sort on a one-way linked list.
module DEFINE_NODE
   implicit  none
   type, public                    :: Node
      integer                      :: Value
      type (Node), pointer         :: Link
   end  type  Node

   type, public                    :: DTpointer
      type (Node), pointer         :: Ptr
   end type DTpointer

end  module  DEFINE_NODE

module  GLOBAL

   use  DEFINE_NODE
   private
   type (Node), pointer, public    :: Head

end  module  GLOBAL

module GROUP
   implicit none
   public :: CREATE_LIST, TRAVERSE_LIST, COUNT_NODES, ORGANIZE, MAKE_ARRAY, &
             SHELL_SORT, REMAKE_LIST

contains

!! Create a one-way linked list by attaching to the head.
subroutine  CREATE_LIST (Head)
      use  DEFINE_NODE
      type (Node), pointer         :: Head

      type (Node), pointer         :: New_Head
      integer                      :: I, N

      print *, "How many values do you want to sort?"
      read *, N
      print *, "Please type ", N, " data, one per line:"
      allocate (Head)
      read *, Head%Value
      nullify (Head%Link)

      do I = 2, N
         allocate (New_Head)
         read *, New_Head%Value
         New_Head%Link => Head
         Head => New_Head
      end  do

end  subroutine  CREATE_LIST

subroutine  TRAVERSE_LIST (Head)
      !! INCOMING: Head points at the head of a one-way linked list.
      !! The procedure returns the number of nodes in the one-way linked list.
      use  DEFINE_NODE
      type (Node), pointer         :: Head

      type (Node), pointer         :: Current

      Current => Head

      !! In this loop (which traverses the linked list), the pointer Current points at the
      !! present node.
      do
         if (.not. associated (Current) ) then
            return
         end  if
         print *, Current%Value
         Current => Current%Link
      end  do

end  subroutine  TRAVERSE_LIST

!! This procedure counts the nodes in the one-way linked list whose head is Head.
function  COUNT_NODES (Head) result (Number_of_Nodes)
      !! INCOMING: Head points at the head of a one-way linked list.
      !! The procedure returns the number of nodes in the one-way linked list.
      use  DEFINE_NODE
      type (Node), pointer         :: Head
      integer                      :: Number_of_Nodes

      type (Node), pointer         :: Current
      integer                      :: N

      Current => Head

      !! In this loop (which traverses the linked list), the pointer Current points at the
      !! present node.  N holds a count of the nodes.
      N = 0
      do
         if (.not. associated (Current) ) then
            exit
         end  if
         Current => Current%Link
         N = N + 1
      end  do
      Number_of_Nodes = N                         !! The number of nodes in the linked list.
end  function  COUNT_NODES

subroutine  ORGANIZE (N)
   use  DEFINE_NODE
   use  GLOBAL
   integer, intent (in)              :: N
   type (DTpointer), dimension ( N ) :: H         !! Define an array of N pointers.

   call MAKE_ARRAY (Head, H)
   call SHELL_SORT (H)
   call REMAKE_LIST (Head, H)

end  subroutine  ORGANIZE

!! This procedure creates an array H containing the pointers to each node of a
!! one-way linked list.  The procedure traverses the linked list, and as it visits each
!! node, it copies the pointer contained in it to H.  The head pointer is copied to H(1).
subroutine  MAKE_ARRAY (Head, H)
      !! INCOMING: Head points to the head of a one-way linked list.
      !! OUTGOING: H    holds pointers to all the nodes, in the same order as the linked list.
      use  DEFINE_NODE
      type (Node), pointer         :: Head
      type (DTpointer), dimension ( : ), intent (out) :: H

      type (Node), pointer         :: Current
      integer                      :: J

      Current => Head                             !! Point at the head of the list.

      !! In this loop, the pointer Current points at the present node.  That pointer is copied into
      !! array H at position J.
      do J = 1, ubound(H, 1)
         if (.not. associated (Current) ) then    !! End of list.
            return                                !! On exit, H contains pointers to each node.
         end  if
         H(J)%Ptr => Current                      !! Store a pointer in H.
         Current => Current%Link                  !! Move on to the next node in the list.
      end  do

end  subroutine  MAKE_ARRAY

!! This procedure performs an indirect sort on the nodes.
!! This subroutine sorts an array by sorting continually-changing sub-groups.
subroutine  SHELL_SORT (H)
      !! INCOMING: H = an array of pointers pointing at nodes.
      !! OUTGOING: H = the pointers, sorted in accordance with the node values.  The nodes are
      !!               in ascending order.
      use  DEFINE_NODE
      type (DTpointer), dimension ( : ), intent (in out)    :: H

      type (Node), pointer         :: Temp
      integer                      :: J, K, N,     &
                                      Gap

      N = ubound(H, 1)                            !! The number of elements in the array P.

      Gap = N/2                                   !! The initial gap is half the number of elements.
      do
         if (Gap < 1) then                        !! We are done after the gap is 1, when all
            exit                                  !! elements are examined in the next loop:
         end  if
         !! Perform a ranking sort on selected elements.
         do  K = 1, N-Gap
            if (H(K)%Ptr%Value > H(K+Gap)%Ptr%Value) then
                                                  !! Select elements every Gap positions apart . .
                  Temp => H(K+Gap)%Ptr
                  do  J = K, 1, -Gap
                     if (H(J)%Ptr%Value <= Temp%Value) then
                        exit
                     end  if
                     H(J+Gap)%Ptr => H(J)%Ptr     !! Leapfrog by Gap locations.
                  end  do
                  H(J+Gap)%Ptr => Temp
            end  if
         end  do
         Gap = Gap/2                              !! Halve the gap for the next ranking sort.
      end  do

   end  subroutine  SHELL_SORT

!! This procedure copies the pointers from the indirectly-sorted array H, into the nodes,
!! thereby re-creating the one-way linked list, the nodes of which are now in the new order.
subroutine  REMAKE_LIST (Head, H)
      use  DEFINE_NODE
      type (Node), pointer         :: Head
      type (DTpointer), dimension ( : ), intent (in out) :: H

      integer                      :: J

      Head => H(1)%Ptr

      do J = 1, ubound (H, 1) - 1
         H(J)%Ptr%Link => H(J+1)%Ptr
      end  do

      nullify (H(J)%Ptr%Link )                    !! The last node has a null pointer.

   end  subroutine  REMAKE_LIST

end  module  GROUP

!! This master procedure organizes the procedures for counting the nodes, creating
!! an array of pointers, performing an indirect sort, and remaking the one-way linked
!! list of sorted nodes.
program  SORT_LIST
   use  DEFINE_NODE
   use  GLOBAL
   use  GROUP
   implicit  none
   integer                         :: N

   call CREATE_LIST (Head)
   N = COUNT_NODES (Head)
   call ORGANIZE (N)
   print *, "The sorted values are:"
   call TRAVERSE_LIST (Head)

end  program  SORT_LIST

!! Figure 3.25  Algorithm for an indirect sort of a linked list
