module DEFINE_NODE
   type, public                    :: Node        !! A template for the node & the pointer to it.
      integer                      :: Value       !! For the value.
      type (Node), pointer         :: Ptr         !! For a pointer to the next node.
   end  type  Node
end  module  DEFINE_NODE

module  GLOBAL

   use  DEFINE_NODE
   private

   type (Node), pointer, public    :: Current     !! An auxiliary pointer.
   type (Node), pointer, public    :: Head, &
                                      Tail, &     !! A pointer to the tail of the list.
                                      Old_Tail
   type (Node), pointer, public    :: Previous_Ptr
   type (Node), pointer, public    :: Test_Prime_Ptr,   &
                                      Prime_Ptr

end  module  GLOBAL

module GROUP
   implicit none
   public                          :: CREATE_HEAD, APPEND_NODE, DELETE_NON_PRIME, PRINT_PRIMES

contains

   subroutine  CREATE_HEAD (Number)
      use  DEFINE_NODE
      use  GLOBAL
      integer, intent (in)         :: Number

      allocate (Tail)                             !! Create the node.
      Tail%Value = Number                         !! Place a value in the node.
      Head => Tail                                !! Point at the node.
      nullify (Tail%Ptr)                          !! A null link for the node

   end  subroutine  CREATE_HEAD

   subroutine  APPEND_NODE (Number)
      use  DEFINE_NODE
      use  GLOBAL
      integer, intent (in)         :: Number

      Old_Tail => Tail                            !! Remember where the tail is.
      allocate (Tail)                             !! Create a new node, to become the tail.
      Tail%Value = Number                         !! Place a value in the new node.
      nullify (Tail%Ptr)                          !! The new tail will have a null pointer.
      Old_Tail%Ptr => Tail                        !! Make the old tail point at the new tail.

   end  subroutine  APPEND_NODE

   subroutine  DELETE_NON_PRIME (Previous_Ptr)
      !! This procedure deletes one node from a linked list.
      !! The dummy argument Previous_Ptr points at the node before the node to be deleted.  This
      !! node is made to point to the node that follows the node being deleted.
      use DEFINE_NODE
      type (Node), pointer         :: Previous_Ptr
      type (Node), pointer         :: Next_Ptr

      Next_Ptr => Previous_Ptr%Ptr                !! Advance to the node to be deleted.
      Previous_Ptr%Ptr => Next_Ptr%Ptr            !! Copy the pointer in the node being deleted
                                                  !! into the previous node.
      deallocate (Next_Ptr)                       !! Give back storage occupied by the node.

   end  subroutine  DELETE_NON_PRIME

   subroutine  PRINT_PRIMES ( )
      !! Traverse the linked list, printing each element as we go.
      use  DEFINE_NODE
      use  GLOBAL

      Current => Head                             !! Start at the Head node.
      do
         if (.not. associated (Current)) then     !! Until we reach the tail.
            exit
         end  if
         write (unit=*, fmt="(i6)", advance="no" ) Current%Value
         Current => Current%Ptr                   !! Point at the next node.
      end  do

   end  subroutine  PRINT_PRIMES

end  module  GROUP

!! This program determines prime numbers using the method of the Sieve of Eratosthenes.
program  PRIMES
   use  DEFINE_NODE
   use  GLOBAL
   use  GROUP
   implicit  none
   integer                         :: K, N

   print *, "This program determines and prints all the integers from 2 to N that are prime."
   print *, "Please type a value for N:"
   read *, N
   print *, "The primes from 2 to", N, " will be formed."

   !! SECTION TO CREATE A LINKED LIST OF INTEGERS.
   call CREATE_HEAD (2)
   Current => Tail                                !! Points at the prime number 2.

   do K = 3, N, 2                                 !! Attaches all the odd integers.
      call APPEND_NODE (K)
   end  do

   !! SECTION TO SEARCH FOR AND DELETE NON-PRIMES.
   Prime_Ptr => Current%Ptr                       !! Points at the integer 3.

   if (N > 2) then
      do                                          !! Work through all the divisors from 3 to sqrt(N) . . .
         if (Prime_Ptr%Value > sqrt (real(N)) ) then
            exit
         end  if

         !! In this inner loop, we "cross off" all multiples of the divisor Prime_Ptr%Value . . .
         !! (We work through all nodes beginning with Test_Prime_Ptr%Value to the end of the list.)
         Test_Prime_Ptr => Prime_Ptr%Ptr
         Previous_Ptr => Test_Prime_Ptr
         do
            if (.not. associated(Test_Prime_Ptr)) then
               exit
            end  if
            if (modulo (Test_Prime_Ptr%Value, Prime_Ptr%Value) == 0) then
               !! "Cross off" Test_Prime_Ptr%Value from the list.
               call DELETE_NON_PRIME (Previous_Ptr)
            end  if
            Previous_Ptr => Test_Prime_Ptr
            Test_Prime_Ptr => Test_Prime_Ptr%Ptr  !! Move to the next node.
         end  do

         Prime_Ptr => Prime_Ptr%Ptr               !! Move to the next prime.
      end  do
   end  if

   call PRINT_PRIMES ( )

end  program  PRIMES

!! Figure 3.3  Primes and the Sieve of Eratosthenes
