module DEFINE_NODE
   type, public                    :: Node        !! A template for the node.
      character (len=20)           :: Value       !! For the person's name.
      type (Node), pointer         :: Ptr         !! For a pointer to the next node.
   end  type  Node
end  module  DEFINE_NODE

module  GLOBAL

   use  DEFINE_NODE
   private

   type (Node), pointer, public    :: Tail
   type (Node), pointer, public    :: Current     !! An auxiliary pointer.
   type (Node), pointer, public    :: Head, Old_Tail
   type (Node), pointer, public    :: Previous_Ptr

end  module  GLOBAL

module GROUP
   implicit none
   public                          :: CREATE_HEAD, APPEND_NODE, DELETE_NAME

contains

   !! This subroutine creates the head of a linked list.
   subroutine  CREATE_HEAD (Name)
      !! INCOMING: Name = the name to be stored in the node to be created.
      use  DEFINE_NODE
      use  GLOBAL

      character (len=*), intent (in)   :: Name

      allocate (Tail)                             !! Create the node.
      Tail%Value = Name                           !! Place a value in the node.
      Head => Tail                                !! Point at the node.
      nullify (Tail%Ptr)                          !! A null link for the node.

   end  subroutine  CREATE_HEAD

   !! This subroutine appends a node to the tail of a linked list.
   subroutine  APPEND_NODE (Name)
      !! INCOMING: Name = the name to be stored in the node to be created.
      use  DEFINE_NODE
      use  GLOBAL
      character (len=*), intent (in)   :: Name

      Old_Tail => Tail                            !! Remember where the tail is.
      allocate (Tail)                             !! Create a new node, to become the tail.
      Tail%Value = Name                           !! Place a value in the new node.
      nullify (Tail%Ptr)                          !! The new tail will have a null pointer.
      Old_Tail%Ptr => Tail                        !! Make the old tail point at the new tail.

   end  subroutine  APPEND_NODE

   !! This procedure deletes one node from a linked list.
   !! The dummy argument Previous_Ptr points at the node before the node to be deleted.  This
   !! node is made to point to the node that follows the node being deleted.
   subroutine  DELETE_NAME (Previous_Ptr)
      !! Previous_Ptr = a pointer to the node prior to the one to be deleted.
      use  DEFINE_NODE
      type (Node), pointer          :: Previous_Ptr
      type (Node), pointer          :: Next_Ptr

      Next_Ptr => Previous_Ptr%Ptr                !! Advance to the node to be deleted.
      Previous_Ptr%Ptr => Next_Ptr%Ptr            !! Copy the pointer in the node being deleted
                                                  !! into the previous node.
      print *, trim (Next_Ptr%Value) // " has been deleted."
      deallocate (Next_Ptr)                       !! Give back storage occupied by the node.

   end  subroutine  DELETE_NAME

end  module  GROUP

!! This program implements Josephus' Problem.
program  JOSEPHUS
   use  DEFINE_NODE
   use  GLOBAL
   use  GROUP
   implicit  none
   character (len=20)              :: Name
   integer                         :: K, N, Gap

   print *, "This program runs the Josephus' problem."
   print *, "How many persons will participate?"
   read *, N
   print *, "Please type their names, one to a line:"

   read (unit=*, fmt="(a20)" ) Name
   call CREATE_HEAD (Name)

   Current => Tail                                !! Points at the first person.

   do K = 2, N                                    !! Creates a node for each person.
      read (unit=*, fmt="(a20)" ) Name
      call APPEND_NODE (Name)
   end  do

   Tail%Ptr => Head                               !! Close the circle -- make the tail point at the head.

   print *, "All the names have been entered.  Tell me how large the gap is."
   print *, "For example, to delete every fourth person, type 4.  How large is the gap?"
   read *, Gap

   do
      if ( associated (Current, Current%Ptr) ) then   !! Stop when there is only one node
         exit                                     !! pointing at itself.
      end  if

      do  K = 1, Gap - 1
         Current => Current%Ptr                   !! Move to the next node.
      end  do
      !! We are now pointing at the person prior to the one to be removed.

      call DELETE_NAME (Current)                  !! To delete the (next) node.
   end  do

   print *, "The winner is: ", Current%Value

end  program  JOSEPHUS

!! Figure 3.4  Josephus' problem solved with a circular linked list
