module  DEFINE_NODE
   type, public                    :: Node        !! A template for the node.
      integer                      :: Value       !! For the value.
      type (Node), pointer         :: Left        !! For a pointer to the left-hand part of the tree.
      type (Node), pointer         :: Right       !! For a pointer to the right-hand part of the tree.
   end  type  Node
end  module  DEFINE_NODE

module  GLOBAL

   use  DEFINE_NODE
   private
   type (Node), pointer, public    :: Current     !! An auxiliary pointer.
   type (Node), pointer, public    :: Head
   type (Node), pointer, public    :: Tail,    &
                                      Root
end  module  GLOBAL

module GROUP
   implicit none
   public                          :: CREATE_NODE, CLIMB, TRAVERSE

contains

   subroutine  CREATE_NODE (Number)
      !! This procedure creates a node.
      use  DEFINE_NODE
      use  GLOBAL
      integer, intent (in)         :: Number

      allocate (Head)                             !! Create the node.
      Head%Value = Number                         !! Place a value in the node.
      Tail => Head                                !! Point at the node.
      nullify (Tail%Left)                         !! A null link for the left-hand pointer.
      nullify (Tail%Right)                        !! A null link for the right-hand pointer.
   end  subroutine  CREATE_NODE

   !! This subroutine climbs from the root to a node containing a null pointer.
   !! It then creates a (leaf) node & joins it to the tree, & inserts NUMBER in the new node.
   subroutine  CLIMB (Dummy_Root, Number)
      use  DEFINE_NODE
      use  GLOBAL
      type (Node), pointer         :: Dummy_Root
      integer, intent (in)         :: Number

      Current => Dummy_Root
      do
         if (Number > Current%Value) then         !! Take the right branch.
            if (.not. associated (Current%Right) ) then
               call CREATE_NODE (Number)          !! Create a leaf node...
               Current%Right => Tail              !! ... and point to it.
               return
            end  if
            Current => Current%Right              !! Move forward.
         else                                     !! Take the left branch.
            if (.not. associated (Current%Left) ) then
               call CREATE_NODE (Number)          !! Create a leaf node...
               Current%Left => Tail               !! ... and point to it.
               return
            end  if
            Current => Current%Left               !! Move forward.
         end  if
      end  do

   end  subroutine  CLIMB

   !! This recursive subroutine retrieves and prints the values from the tree in ascending order.
   recursive  subroutine  TRAVERSE (Current)
      use  DEFINE_NODE
      type (Node), pointer         :: Current

      if (associated (Current%Left) ) then        !! Take the left subtree.
         call TRAVERSE (Current%Left)
      end  if

      print *, Current%Value                      !! Retrieve value from tree and print it.

      if (associated (Current%Right) ) then       !! Take the right subtree.
         call TRAVERSE (Current%Right)
      end  if

   end  subroutine  TRAVERSE

end  module  GROUP

!! This program creates a binary search tree.  It uses the subroutine CREATE_NODE to create
!! nodes for the tree, and the subroutine CLIMB to determine where to append the node.
!! Subroutine TRAVERSE retrieves nodes from the tree in ascending order.
program  SORT
   use  DEFINE_NODE
   use  GLOBAL
   use  GROUP
   implicit  none

   integer                         :: Number, Status

   print *, "Please type values, one to a line (to end, enter 0, or give the end-of-file key sequence):"
   read *, Number                                 !! The first value to be inserted at the root.
   call CREATE_NODE (Number)                      !! Create the root node.
   Root => Head

   do
      read (unit=*, fmt=*, iostat=Status) Number  !! The next value to be inserted.
      if (Status < 0) then                        !! We encountered the end of the file.
         exit
      end  if
      if (Number == 0) then                       !! A zero will terminate input.
         exit
      end  if
      call CLIMB (Root, Number)                   !! Climb the tree, looking for a place to insert.
   end  do

   print *, "The sorted values are:"

   call TRAVERSE (Root)

end  program  SORT

!! Figure 3.7  Sorting using a binary tree
