module  DEFINE_NODE
   type, public                    :: Node        !! A template for the node.
      integer                      :: Value       !! For the value.
      type (Node), pointer         :: Left        !! For a pointer to the left-hand part of the tree.
      type (Node), pointer         :: Right       !! For a pointer to the right-hand part of the tree.
   end  type  Node
end  module  DEFINE_NODE

module  GLOBAL

   use  DEFINE_NODE
   private

   type (Node), pointer, public    :: Current     !! An auxiliary pointer.
   type (Node), pointer, public    :: Head, Tail

   type (Node), pointer, public    :: Root

end  module  GLOBAL

module  GLOBAL_ARRAY

   integer, dimension ( : ), allocatable, public   :: A
   integer, public                     :: K

end  module  GLOBAL_ARRAY

module GROUP
   implicit none
   public :: RETRIEVE, INITIAL_INSERT, INSERT_SUBGROUP, CREATE_NODE, CLIMB

contains

!! This subroutine retrieves all the elements from a binary tree and writes them into the
!! array A in positions 1 through K in ascending order.  It also deletes the binary tree.
recursive  subroutine  RETRIEVE (Current)
      use  DEFINE_NODE
      use  GLOBAL_ARRAY
      type (Node), pointer         :: Current

      if (.not. associated(Current) ) then
         return
      end  if

      call RETRIEVE (Current%Left)                !! Visit the left subtree.
      A(K) = Current%Value                        !! Retrieve value from the tree and store it in A(K).
      K = K + 1
      call RETRIEVE (Current%Right)               !! Visit the right subtree.
      deallocate (Current)                        !! Release the storage for the node.
end  subroutine  RETRIEVE

!! This subroutine creates a balanced and sorted binary tree from N pre-sorted elements.
!! The root and sub-roots at the same level are stored in order, thus ensuring a perfectly
!! symmetrical and balanced tree.
subroutine  INITIAL_INSERT (N)
      use  DEFINE_NODE
      use  GLOBAL
      use  GLOBAL_ARRAY
      integer, intent (in)         :: N

      integer                      :: J, Step

      Step = (N+1)/2
      call CREATE_NODE (A(Step))
      Root => Head

      do
         if (Step <= 1) then
            exit
         end  if
         do J = Step/2, N-1, Step
            call CLIMB (Root, A(J))
         end  do
         Step = Step/2
      end  do

end  subroutine  INITIAL_INSERT

!! This subroutine inserts elements A(start) through A(finish) into the binary tree.
subroutine  INSERT_SUBGROUP (Start, Finish)
      use  DEFINE_NODE
      use  GLOBAL
      use  GLOBAL_ARRAY
      integer, intent (in)         :: Start, Finish

      integer                      :: J

      do J = Start, Finish
         call CLIMB (Root, A(J))
      end  do

end  subroutine  INSERT_SUBGROUP

   subroutine  CREATE_NODE (Number)
      !! This procedure creates a node.
      use  DEFINE_NODE
      use  GLOBAL
      integer, intent (in)         :: Number

      allocate (Head)                             !! Create the node.
      Head%Value = Number                         !! Place a value in the node.
      Tail => Head                                !! Point at the node.
      nullify (Tail%Left)                         !! A null link for the left-hand pointer.
      nullify (Tail%Right)                        !! A null link for the right-hand pointer.
   end  subroutine  CREATE_NODE

   !! This subroutine climbs from the root to a node containing a null pointer.
   !! It then creates a (leaf) node & joins it to the tree, & inserts NUMBER in the new node.
   subroutine  CLIMB (Dummy_Root, Number)
      use  DEFINE_NODE
      use  GLOBAL
      type (Node), pointer         :: Dummy_Root
      integer, intent (in)         :: Number

      Current => Dummy_Root
      do
         if (Number > Current%Value) then         !! Take the right branch.
            if (.not. associated (Current%Right) ) then
               call CREATE_NODE (Number)          !! Create a leaf node...
               Current%Right => Tail              !! ... and point to it.
               return
            end  if
            Current => Current%Right              !! Move forward.
         else                                     !! Take the left branch.
            if (.not. associated (Current%Left) ) then
               call CREATE_NODE (Number)          !! Create a leaf node...
               Current%Left => Tail               !! ... and point to it.
               return
            end  if
            Current => Current%Left               !! Move forward.
         end  if
      end  do

   end  subroutine  CLIMB

end  module  GROUP

!! This program produces a sorted binary tree.  Elements from a given array are inserted
!! in the tree.  After 511 elements have been inserted, those sorted elements are retrieved
!! from the tree, and are placed in an array.  The 511 array elements are then written back
!! into the tree in such a way as to create a perfectly-balanced tree.  The remaining
!! elements (if any) are then inserted in the tree, via a normal insertion.
program  BALANCED_SORT
   use  DEFINE_NODE
   use  GLOBAL
   use  GLOBAL_ARRAY
   use  GROUP
   implicit  none
   integer                         :: Number

   integer                         :: N

   print *, "This program sorts values using a binary tree."
   print *, "How many values do you wish to sort?"
   read *, N

   allocate (A(N))

   print *, "Please type ", N, " values:"
   read *, A
   print *, "The values are: ", A

   call CREATE_NODE (A(1))                        !! Insert the first element.
   Root => Head
   if (N > 511) then
      call INSERT_SUBGROUP (2, 511)               !! To complete insertion of the first 511 elements.
      K = 1                                       !! Where to insert in array A.
      call RETRIEVE (Root)                        !! Fetch the sorted elements into array A.
      call INITIAL_INSERT (511)                   !! Remake a balanced tree.
      call INSERT_SUBGROUP (512, N)               !! Insert the remaining elements.
   else
      call INSERT_SUBGROUP (2, N)                 !! Complete the insertion of N elements
                                                  !! in the tree (N < 512)
   end  if

   !! The values are all in the binary search tree.  Print them.
   print *, "The sorted values are:"
   K = 1
   call RETRIEVE (Root)
   print *, A

end  program  BALANCED_SORT

!! Figure 3.8  Sorting using a binary tree (improved)
