!! This program generates and prints a color image on a Fujitsu DL1150 series printer.
module  GLOBAL
   character (len=1), dimension (0:1023, 0:1023), public :: Pixel_Image
end  module  GLOBAL

module GROUP
   implicit none
   public                          :: FRACTALS, PRINT_GRAPHICS, WRITE_BYTES

contains

!!  insert subroutine FRACTALS from Figure 4.11 

!! This subroutine generates Julia set fractals.
subroutine  FRACTALS ( )
   use  GLOBAL
   complex                         :: Z, C
   integer, parameter              :: Screen_Width = 1024, &
                                      Screen_Height =768        !! Measured in pixels.
   integer                         :: Max_Iterations
   integer                         :: Color
   integer                         :: Row, Col
   real                            :: Dx, Dy,           &
                                      Xmin, Ymin,       &
                                      Xmax, Ymax
   real                            :: R, M, M_Squared

   print *, "This program creates a Julia Set fractal, using the equation z = z**2 + c."
   do
      print *, "Please type the fractal constant  c  as a complex number, e.g. (0.25, -0.25) :"

      read *, C
      if (aimag(C) == 0) then
         print *, "The imaginary component must be nonzero.  Please try again."
      else
         exit
      end  if
   end  do

   print *, "Please type the break-out limit (suggest 100):"
   read *, M
   M_Squared = M*M

   print *, "Please type the maximum number of iterations (suggest 30):"
   read *, Max_Iterations

   Xmin = -1.5                                    !! These four lines specify a rectangle in the X-Y
   Xmax = 1.5                                     !! plane.  The fractal will be generated for this
   Ymin = -1.5                                    !! rectangle.  To magnify an area, make the rectangle
   Ymax = 1.5                                     !! small.  The rectangle can be moved around the
                                                  !! X-Y plane by changing any or all of the values.

   Dx = (Xmax - Xmin)/(Screen_Width-1)
   Dy = (Ymax - Ymin)/(Screen_Height-1)

   !! For all the pixels (points) on the screen:
   do Row = 0, Screen_Height-1                    !! . . . in the vertical . . .
      do Col = 0, Screen_Width-1                  !! . . . in the horizontal . . . .
         Z = cmplx (Xmin + Col*Dx, Ymin + Row*Dy) !! Take as the initial value the
                                                  !! co-ordinates of the current point (Row, Col).
         Color = 0                                !! Select Black.

         !! In this loop, the value Z will either escape to infinity, or will remain less than 1.
break_out_loop: do
            Z = Z*Z + C                           !! Transform.
            Color = Color + 1                     !! Select the next color.
            if (Color > Max_Iterations) then      !! The point Z collapsed.
               Color = 0                          !! Select Black.
               exit  break_out_loop
            end  if
            R = real(Z)*real(Z) + aimag(Z)*aimag(Z)
            if (R > M_Squared) then               !! The point Z "escaped" to infinity.
               exit  break_out_loop
            end  if
         end  do  break_out_loop
         if (Color > 7) then
            Color = modulo(Color-1, 7) + 1
         end  if
         Pixel_Image(Row, Col) = char(Color)      !! Color the pixel.
      end  do
   end  do

end  subroutine  FRACTALS

   !! This subroutine prints a graphics file on a Fujitsu 1150 (DPL24C) dot-matrix color printer
   !! at 360 dpi
   subroutine  PRINT_GRAPHICS ( )
      use  GLOBAL
      character (len=3), dimension (1024) :: BC
      integer                      :: Row, Col, RowJ
      integer                      :: Color
      character (len=2)            :: No_Bytes
      character (len=1), parameter :: ESC = char(27), FS = char(28), CR = char(13), LF=char(10)
      integer                      :: Bits
      character (len=3), parameter :: Empty = char(0) // char(0) // char(0)
      integer                      :: J

      open (unit=9, file="GRAPHIC.PRN", access="direct", action="write", &
         recl=1, status="new")

      call WRITE_BYTES (CR // LF)                 !! Reset printer to the left-hand end of the line.

      do RowJ = 0, 999, 24
         do Color = 1, 6                          !! For each of the colors Magenta, Cyan, Violet,
                                                  !! Yellow, Orange, and Green.
                                                  !! Color 0 (black) is not printed, best left as white.
            BC = Empty
            do Col = 0, 1023                      !! For each of the bits in one column.
               Bits = 0
               do Row = 0, 23                     !! For each of the bits in 24 rows of one column.
                  if (Pixel_Image(RowJ+Row, Col) == char(Color) ) then
                     Bits = ishft(Bits, 1) + 1
                  else
                     Bits = ishft(Bits, 1)
                  end  if
               end  do
               !! "Bits" holds 24 bits for the first 24 rows in column "Col" in a given color.
               BC(Col+1) = char(ishft(Bits, -16) ) // char (iand(Bits/256, 255) ) //   &
                       char (iand(Bits, 255))
            end  do

            !! Check whether there are any pixels to be colored in.
            if (any(BC /= Empty)) then
               call WRITE_BYTES (ESC // "r" // char(Color) )   !! Select the color.
               No_Bytes = char(0) // char(4)
               call WRITE_BYTES (FS // "Z" // No_Bytes)     !! 360 dpi 24-pin graphics mode,
               do J = 1, ubound(BC, 1)            !! with 3x1024 data bytes.
                  call WRITE_BYTES (BC(J))
               end  do
               !! After doing a color, start at the left-hand end of the line.
               call WRITE_BYTES (CR)
            end  if
         end  do
         !! At end of printing (up to) 6 colors for a given line,
         !! advance to the next line by 48/360 of an inch.
         call WRITE_BYTES (ESC // "eJ048")
      end  do

      call WRITE_BYTES (CR // LF)                 !! Carriage return to left-hand side.

   end  subroutine  PRINT_GRAPHICS

   !! This subroutine writes to the graphics file, a byte at a time.
   subroutine  WRITE_BYTES (String)
      character (len=*), intent (in) :: String

      integer, save                  :: Rec_No = 0
      integer                        :: J

      do J = 1, len(String)
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) String(J:J)
      end  do

   end  subroutine  WRITE_BYTES

end  module  GROUP

program  FRACTAL
   use  GLOBAL
   use  GROUP
   implicit  none

   Pixel_Image = char(0)
   call FRACTALS ( )
   call PRINT_GRAPHICS ( )

end  program  FRACTAL

!! Figure 4.13  Capturing a fractal on a color printer
