!! character (len=1), dimension (0:1023, 0:1023), public :: Pixel_Image
!! Pixel_Image = char(0)            !! Set the image to the background color (white or black).

module  PCX_Header
   !! The next set of declarations constitutes the PCX_Header (which is 128 bytes in size).
   type, public                    :: Header
      character (len=1)            :: Identifier  !! File Identification.
      character (len=1)            :: Version     !! Version Number.
      character (len=1)            :: Encoding_format !! 0 = no encoding (uncompressed)
                                                      !! 1 = run-length encoding (compressed).
      character (len=1)            :: Bits_per_pixel  !! 4 bits per pixel.
                                                  !! Co-ordinates for the start of the image:
      character (len=2)            :: X_Start     !! X co-ordinate of start.
      character (len=2)            :: Y_Start     !! Y co-ordinate of start.
                                                  !! Co-ordinates for the end of the image:
      character (len=2)            :: X_end       !! X co-ordinate of end.
                                                  !! 0:1023 = length of image in pixels.
      character (len=2)            :: Y_end       !! Y co-ordinate of end.
                                                  !! 0:1023 = height of image in pixels.
      character (len=2)            :: Horizontal_Resolution !! in dots per inch.
      character (len=2)            :: Vertical_Resolution   !! in dots per inch.
      character (len=1), dimension (0:47) :: Palette        !! The color palette.
      character (len=1)            :: Reserved
      character (len=1)            :: Num_Bit_Planes
      character (len=2)            :: Bytes_per_line    !! # of bytes per line unencoded.
      character (len=2)            :: Palette_type
      character (len=2)            :: Horizontal_Screen_Size
      character (len=2)            :: Vertical_Screen_Size
      character (len=1), dimension (54)   :: Reserved2
   end  type  Header                              !! End of declarations for the PCX header.

end  module  PCX_Header

module GROUP
   implicit none
   public                          :: MAKE_PCX_FILE, PACK_CHAR

contains

!! This subroutine creates a picture file in PCX format.
!! The record size of the output file is 128 bytes.
subroutine  MAKE_PCX_FILE ( )
   !! A picture file commences with a 128-byte header, which gives video screen
   !! parameters, colors, and size of image.
   !! For details of PCX file formats, see J.D. Murray & W. vanRyper, "Encyclopedia of Graphics File
   !! Formats", 2nd Ed., O'Reilly, Sebastopol, CA, 1996.

   use  PCX_Header

   type (Header)                   :: H

   character (len=128)             :: Pixel_line

   integer                         :: Rec_No
   integer                         :: I, K, L
   character (len=512)             :: LINE

   character (len=1)               :: Nil, Qtr, Half, Tqtr, Full

   !! INITIALIZATION SECTION: Initialize parameters in the PCX file header.
               !! The next array holds the color palette, 16 colors in groups of 3.
               !! Each line represents the components of one color: red, green, & blue.
               !! 0 represents none of that color, 255 represents the
               !! full intensity of that color.
               !! The order of the following colors is unimportant, and has been
               !! selected for providing best contrast on paper.
               !! White is color #0, best background for printing on white paper.  For
               !! displaying on screen, black is usually the best background, so exchange
               !! black and white (#0 with #13).
   H%Palette = (/ &  !!  Red     Green      Blue
                     char(255), char(255), char(255), & !!  0 = White.
                     char(255), char(0),   char(0),   & !!  1 = Red.
                     char(255), char(255), char(0),   & !!  2 = Yellow.
                     char(255), char(0),   char(255), & !!  3 = Purple.
                     char(102), char(255), char(102), & !!  4 = Light Green
                     char(0),   char(255), char(255), & !!  5 = Light blue.
                     char(0),   char(0),   char(255), & !!  6 = Dark blue.
                     char(255), char(102), char(0),   & !!  7 = Orange.
                     char(0),   char(255), char(0),   & !!  8 = Green.
                     char(255), char(127), char(255), & !!  9 = Pale purple.
                     char(255), char(255), char(127), & !! 10 = Lemon.
                     char(0),   char(127), char(255), & !! 11 = Medium blue.
                     char(255), char(192), char(0),   & !! 12 = Light brown.
                     char(0),   char(0),   char(0),   & !! 13 = Black.
                     char(0),   char(0),   char(0),   &
                     char(0),   char(0),   char(0)  /)

   H%Identifier = char(10)
   H%Version   = char(5)
   H%Encoding_format = char(0)                    !! No encoding (uncompressed).
   H%Bits_per_pixel = char(4)
   H%X_Start = PACK_CHAR (0)
   H%Y_Start = H%X_Start
   H%X_End = PACK_CHAR (1023)
   H%Y_End = H%X_End
   H%Horizontal_Resolution = PACK_CHAR (300)
   H%Vertical_Resolution = PACK_CHAR (300)

   H%Reserved = char (0)
   H%Num_Bit_Planes = char (1)
   H%Bytes_Per_Line = PACK_CHAR (512)
   H%Palette_Type = PACK_CHAR (4)
   H%Horizontal_Screen_Size = H%X_End
   H%Vertical_Screen_Size = H%Y_End
   H%Reserved2 = char(0)
   !! End of initialization section.

   open (unit=9, file="FRACTAL.PCX", action="write",  &
      access="direct", form="unformatted", status="new", recl=128)

   Rec_No = 1
   write (unit=9, rec=Rec_No) H                   !! Write out the parameters of the picture file.

   do I = 0, 1023                                 !! For each of the scan lines . . . .
      !! Each element of Pixel_Image holds only 4 bits.  Pack two elements into each byte.
      L = 1
      do K = 0, 1023, 2
         LINE(L:L) = char ( ichar(Pixel_Image (I,K))*16 + ichar(Pixel_Image(I,K+1) ) )
         L = L + 1
      end  do

      do K = 0, 3                                 !! Write out the line in four segments.
         Pixel_line = LINE (K*128+1:K*128+128)
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Pixel_line
      end  do
   end  do

end  subroutine  MAKE_PCX_FILE

!! This function converts a binary integer -- containing a positive integer of
!! 16 significant bits -- to a two-byte integer in reverse format, that is,
!! least-significant byte first.  The reverse-byte integer is returned as CHARACTER.
function  PACK_CHAR (Whole_Number) result (Pair)
      integer, intent (in)         :: Whole_Number

      character (len=2)            :: Pair

      Pair(2:2) = char (ishft (Whole_Number, -8) )!! The high-order bits come second.
      Pair(1:1) = char (iand (Whole_Number, 255) )!! The low-order bits come first.

   end  function  PACK_CHAR

end  module  GROUP

!! Figure 4.16  Procedure to write out a picture file in PCX format
