module  GLOBAL
   character (len=1), dimension (0:1023, 0:1023), public :: Pixel_Image
end  module  GLOBAL

module  PCX_Header
   !! The next set of declarations constitutes the PCX_Header (which is 128 bytes in size).
   type, public                    :: Header
      character (len=1)            :: Identifier  !! File Identification.
      character (len=1)            :: Version     !! Version Number.
      character (len=1)            :: Encoding_format !! 0 = no encoding (uncompressed);
                                                      !! 1 = run-length encoding (compressed).
      character (len=1)            :: Bits_per_pixel  !! 4 bits per pixel.
                                                  !! Co-ordinates for the start of the image:
      character (len=2)            :: X_Start     !! X co-ordinate of start.
      character (len=2)            :: Y_Start     !! Y co-ordinate of start.
                                                  !! Co-ordinates for the end of the image:
      character (len=2)            :: X_end       !! X co-ordinate of end.
                                                  !! 0:1023 = length of image in pixels.
      character (len=2)            :: Y_end       !! Y co-ordinate of end.
                                                  !! 0:1023 = height of image in pixels.
      character (len=2)            :: Horizontal_Resolution   !! in dots per inch.
      character (len=2)            :: Vertical_Resolution  !! in dots per inch.
      character (len=1), dimension (0:47) :: Palette  !! The color palette.
      character (len=1)            :: Reserved
      character (len=1)            :: Num_Bit_Planes
      character (len=2)            :: Bytes_per_line    !! # of bytes per line unencoded.
      character (len=2)            :: Palette_type
      character (len=2)            :: Horizontal_Screen_Size
      character (len=2)            :: Vertical_Screen_Size
      character (len=54)           :: Reserved2
   end  type  Header                              !! End of declarations for the PCX header.

end  module  PCX_Header

module GROUP
   use  GLOBAL
   implicit none
   private
   public                          :: WRITE_BYTE_INTO_BUFFER, COMPRESS, MAKE_PCX_FILE, PACK_CHAR
   integer, public                 :: Rec_No
   integer, public                 :: WBP
   character (len=128), public     :: Buffer

contains

!! These subroutines create a compressed picture file in PCX format.
!! The record size of the output file is 128 bytes.

!! This subroutine accepts a single byte and stores it in a buffer,
!! writing out the buffer whenever it becomes full.
subroutine  WRITE_BYTE_INTO_BUFFER (Byte_to_Write)
      character (len=*), intent (in)   :: Byte_to_Write

      WBP = WBP + 1
      if (WBP > 128) then
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Buffer
         WBP= 1
      end  if

      Buffer(WBP:WBP) = Byte_to_Write
   end  subroutine  WRITE_BYTE_INTO_BUFFER

!! This subroutine uses run-length encoding to compress the picture image in Pixel_Image.
!! The area of the image to be compressed is 0:X_end, 0:Y_end.
!! The image is written out via subroutine WRITE_BYTE_INTO_BUFFER.
subroutine  COMPRESS (X_end, Y_end)
   integer, intent (in)            :: X_end       !! The address of the RHS column of the image.
   integer, intent (in)            :: Y_end       !! The address of the bottom row of the image.

   character (len=1)               :: Byte, CHR
   integer                         :: I, CP, Start, Ending, Run_Length
   integer                         :: UB4, LB4
   character (len=1)               :: B8

   do I = 0, Y_end                                !! For each of the scan lines.
      CP = 0
      do
         UB4 = ichar(Pixel_Image(I,CP) )          !! Note the pair of pixels at the start of a potential
         LB4 = ichar(Pixel_Image(I,CP+1) )        !! run.
         B8 = char(UB4*16 + LB4)
         Start = CP                               !! The location of the byte at the start of the run.
         do                                       !! Look for the end of the run.
            CP = CP + 2
            if (CP >= X_end) then
               exit
            end  if
            UB4 = ichar(Pixel_Image(I, CP) )
            LB4 = ichar(Pixel_Image(I, CP+1) )
            if (B8 /= char(UB4*16 + LB4) ) then
               exit
            end  if
         end  do
         !! When we get here, CP points at the byte that terminates the run, or is > X_end.
         Ending = CP
         Run_Length = (Ending - Start)/2          !! The length of the run. [Note: not
                                                  !! Ending-Start+1.].
         Byte = B8                                !! The byte at the start of the run.

         !! This loop breaks very long runs to chunks of 63 bytes.
         do
            if (Run_Length <= 63) then
               exit
            end  if
            call WRITE_BYTE_INTO_BUFFER (char(255) )  !! 'C0'X + 63.
            call WRITE_BYTE_INTO_BUFFER (Byte)
            Run_Length = Run_Length - 63
         end  do

         !! This section handles runs of length 1 (special case) to length 63.
         if (Run_Length == 1) then                !! Just one byte in this run.
            if (Byte <= char(63) ) then           !! '00111111'B
               !! Special case: upper 2 bits are zero.
               call WRITE_BYTE_INTO_BUFFER(Byte)
            else                                  !! Bad luck: one data byte expands to 2 bytes.
               !! The count byte is '11000001'B = 193, for a one-byte run.
               CHR = char(193)
               call WRITE_BYTE_INTO_BUFFER (CHR)
               call WRITE_BYTE_INTO_BUFFER (Byte)
            end  if
         else if (Run_Length > 1) then
            CHR = char (192 + Run_Length)         !! 'C0'X = '11000000'B = 192
                                                  !! 'C0'X + length of run.
            call WRITE_BYTE_INTO_BUFFER (CHR)
            call WRITE_BYTE_INTO_BUFFER (Byte)
         end  if

         if (CP >= X_end) then                    !! We're past the end of a scan line.
            exit
         end  if
      end  do
   end  do

   do I = WBP+1, len(Buffer)+1                    !! Fill out buffer with zero bytes, and one more,
      call WRITE_BYTE_INTO_BUFFER (char(0) )      !! and write out the final record.
   end  do

end  subroutine  COMPRESS

!! This subroutine writes the header for a picture file in PCX format.
!! It then calls COMPRESS to prepare the graphic image.
!! The record size of the output file is 128 bytes.
subroutine  MAKE_PCX_FILE ( )
   !! A picture file commences with a 128-byte header, which gives video screen
   !! parameters, colors, and size of image.
   !! For details of PCX file formats, see J.D. Murray & W. vanRyper, "Encyclopedia of Graphics
   !! File Formats", 2nd Ed., O'Reilly, Sebastopol, CA, 1996.

   use  PCX_Header

   type (Header)                   :: H

   integer                         :: Bytes_per_line, IX_end, IY_end

   integer                         :: Height, Width

   !! INITIALIZATION SECTION: Initialize parameters in the PCX file header.
               !! The next array holds the color palette, 16 colors in groups of 3.
               !! Each line represents the components of one color: red, green, & blue.
               !! 0 represents none of that color, 255 represents the
               !! full intensity of that color.
               !! The order of the following colors is unimportant, and has been
               !! selected for providing best contrast on paper.
               !! White is color #0, best background for printing on white paper.  For
               !! displaying on screen, black is usually the best background, so exchange
               !! black and white (#0 with #13).
   H%Palette = (/ &  !!  Red     Green      Blue
                     char(255), char(255), char(255), & !!  0 = White.
                     char(255), char(0),   char(0),   & !!  1 = Red.
                     char(255), char(255), char(0),   & !!  2 = Yellow.
                     char(255), char(0),   char(255), & !!  3 = Purple.
                     char(102), char(255), char(102), & !!  4 = Light Green
                     char(0),   char(255), char(255), & !!  5 = Light blue.
                     char(0),   char(0),   char(255), & !!  6 = Dark blue.
                     char(255), char(102), char(0),   & !!  7 = Orange.
                     char(0),   char(255), char(0),   & !!  8 = Green.
                     char(255), char(127), char(255), & !!  9 = Pale purple.
                     char(255), char(255), char(127), & !! 10 = Lemon.
                     char(0),   char(127), char(255), & !! 11 = Medium blue.
                     char(255), char(192), char(0),   & !! 12 = Light brown.
                     char(0),   char(0),   char(0),   & !! 13 = Black.
                     char(0),   char(0),   char(0),   &
                     char(0),   char(0),   char(0)  /)

   H%Identifier = char(10)
   H%Version   = char(5)
   H%Encoding_format = char(1)                    !! Run length encoding (compressed file).
   H%Bits_per_pixel = char(4)
   H%X_Start = PACK_CHAR (0)
   H%Y_Start = H%X_Start
   H%X_End = PACK_CHAR (1023)
   H%Y_End = H%X_End
   H%Horizontal_Resolution = PACK_CHAR (300)
   H%Vertical_Resolution = PACK_CHAR (300)

   H%Reserved = char (0)
   H%Num_Bit_Planes = char (1)
   H%Bytes_Per_Line = PACK_CHAR (512)
   H%Palette_Type = PACK_CHAR (4)
   H%Horizontal_Screen_Size = H%Horizontal_Resolution
   H%Vertical_Screen_Size = H%Vertical_Resolution
   H%Reserved2 = repeat (char(0), len(H%Reserved2) )
   !! End of initialization section.

   !! This section crops the image -- it removes empty borders at the right-hand side and
   !! the bottom.
   !! Determine the height of the actual image.
   do Height = ubound(Pixel_Image, 1), 2, -1
      if ( any(Pixel_Image(Height, : ) /= char(0) ) ) then
         exit
      end  if
   end  do
   !! When we get here, Height is the row number of the bottom of the image.
   IY_end = Height
   H%Y_end = PACK_CHAR (IY_end)

   !! Determine the width of the actual image.
   do Width = ubound(Pixel_Image, 2), 2, -1
      if ( any(Pixel_Image( : , Width) /= char(0) ) ) then
         exit
      end  if
   end  do
print *, "Width=", Width, "Height=", Height
   !! When we get here, Width is the column number of the RHS of the image.
   IX_end = Width
   Bytes_per_line = (IX_end+8)/8*ichar(H%Bits_per_pixel)
                                                  !! +8, not +7, because pixels number from 0.

   if (modulo(Bytes_per_line, 2) /= 0) then
      Bytes_per_line = Bytes_per_line + 1         !! Make sure that the line has an even # of bytes.
   end  if
   H%Bytes_per_line = PACK_CHAR (Bytes_per_line)
   IX_end = Bytes_per_line*8/ichar(H%Bits_per_pixel)-1   !! Recompute the image area.
   H%X_end = PACK_CHAR (IX_end)

   open (unit=9, file="FRACTAL.PCX", action="write",  &
      access="direct", form="unformatted", status="new", recl=128)

   Rec_No = 1
   write (unit=9, rec=Rec_No) H                   !! Write out the parameters of the picture file.

   call COMPRESS (IX_end, IY_end)                 !! Write out the compressed image.

end  subroutine  MAKE_PCX_FILE

!!  insert function PACK_CHAR from Figure 4.16 

!! This function converts a binary integer -- containing a positive integer of
!! 16 significant bits -- to a two-byte integer in reverse format, that is,
!! least-significant byte first.  The reverse-byte integer is returned as CHARACTER.
function  PACK_CHAR (Whole_Number) result (Pair)
      integer, intent (in)         :: Whole_Number

      character (len=2)            :: Pair

      Pair(2:2) = char (ishft (Whole_Number, -8) )!! The high-order bits come second.
      Pair(1:1) = char (iand (Whole_Number, 255) )!! The low-order bits come first.

   end  function  PACK_CHAR

end  module  GROUP

!! Figure 4.17  Procedures to write out a compressed picture file in PCX format

module M15
   implicit  none
   public :: MANDELBROT_FRACTAL

contains

!! This subroutine generates the Mandelbrot Set.
subroutine  MANDELBROT_FRACTAL ( )
   use  GLOBAL
   complex                         :: Z, C
   integer, parameter              :: Screen_Width = 640,  &
                                      Screen_Height = 350          !! Measured in pixels.
   integer                         :: Max_Iterations
   integer                         :: Color
   integer                         :: Row, Col
   real                            :: Dx, Dy,           &
                                      Xmin, Ymin,       &
                                      Xmax, Ymax
   real                            :: M, M_squared, R

   print *, "This program creates a Mandelbrot Set fractal, using the equation z = z**2 + c."

   print *, "Now input the dimensions of the Mandelbrot image that you want to view"
   print *, "(for the entire Mandelbrot Set fractal, please type -2, 2, -2, 2):"
   print *, "Please type the left end:"
   read *, Xmin
   print *, "Please type the right-hand end:"
   read *, Xmax
   print *, "Please type the lower end:"
   read *, Ymin
   print *, "Please type the upper end:"
   read *, Ymax

   print *, "Please type the break-out limit (suggest 2):"
   read *, M

   M_squared = M*M

   print *, "Please type the maximum number of iterations:"
   print *, "(suggest 15 for the full Mandelbrot set; 150 for a small part of it):"
   read *, Max_Iterations

   Dx = (Xmax - Xmin)/(Screen_Width-1)
   Dy = (Ymax - Ymin)/(Screen_Height-1)

!! call VGA@                     !! For Salford graphics

   !! For all the pixels (points) on the screen:
   do Row = 0, Screen_Height-1                    !! . . . in the vertical . . .
print *, Row
      do Col = 0, Screen_Width-1                  !! . . . in the horizontal . . . .

         C = cmplx (Xmin + Col*Dx, Ymin + Row*Dy) !! Derive the fractal constant C from the
                                                  !! co-ordinates (Row, Col).
         Z = 0                                    !! Initialize the first Z.

         !! In this loop, the value Z will either escape to infinity, or will remain less than 1.
         Color = 0                                !! Begin with black (Color = 0) . . .
         do
            Z = Z*Z + C                           !! Transform
            R = real(Z)*real(Z) + aimag(Z)*aimag(Z)
            if (R > M_squared) then               !! The point Z "escaped" to infinity.
               exit
            end  if
            Color = Color + 1
            if (Color > Max_Iterations) then      !! The point collapsed to zero . . .
               Color = 0                          !! . . . so color the point black . . .
               exit                               !! . . . and quit.
            end  if
         end  do

         if (Color > 7) then
            Color = modulo(Color-1, 7) + 1
         end  if
         Pixel_Image(Row, Col) = char(Color)      !! Store the color.
!!       call SET_PIXEL@ (Col, Screen_Height-1-Row, Color)  !! For Salford graphics.
      end  do
   end  do

end  subroutine  MANDELBROT_FRACTAL

!! Figure 4.15  Subroutine for generating the Mandelbrot Set

end  module  M15

program test
   use  M15
   use  GROUP
   use  GLOBAL

   Pixel_Image = char (0)

   call MANDELBROT_FRACTAL ( )
   call MAKE_PCX_FILE ()
end  program  test
