module GROUP
   implicit none
   public                          :: SIMEQ, PRINT_COEFFICIENTS
   integer, parameter, public      :: Double = selected_real_kind (14)

contains

   !! This subroutine solves N complex linear algebraic simultaneous equations using
   !! GAUSS elimination and backward-substitution.
   subroutine  SIMEQ (A, R, X)
      !! INCOMING: A = a set of N coefficients for each of N equations, stored in matrix form.
      !!               The coefficients are changed.
      !!               The first equation is stored in the first row; the second in row 2, & so on.
      !!           R = a vector of N right-hand sides for the above equations.  The
      !!               right-hand side of the first equation is stored in R(1), etc.
      !!               These also are changed;
      !! OUTGOING: X = the solutions of the above equations.

      complex (kind = Double), dimension ( : , : ), intent (in out) :: A
      complex (kind = Double), dimension ( : ), intent (in out)   :: R
      complex (kind = Double), dimension ( : ), intent (out)   :: X

      complex (kind = Double)      :: Partial_Sum, Multiplier
      integer                      :: N
      integer                      :: I, J, K

      !! First check that the arrays are compatible.
      N = ubound(A, 1)                            !! The number of equations.
      if (N /= ubound(A, 2) .or. N /= ubound (R, 1) .or. N /= ubound (X, 1) ) then
         print *, "**ERROR, the dimensions of the arrays for the " //   &
            " Gauss equation solver are incompatible."
         print *, "The coefficient matrix is of size ", N, " by ", ubound(A, 2)
         print *, "The vector of right-hand sides is of size ", ubound (R, 1)
         print *, "The solution vector X is of size ", ubound (X, 1)
         print *, "The solution is abandoned."
         return
      end  if

      !! The first section performs Gauss elimination,
      !! to reduce the equations to upper triangular form.

      do I = 1, N-1                               !! For each pivot equation.
         do J = I+1, N                            !! For each row below the pivot equation.
            Multiplier = A(J,I) / A(I,I)
            do K = I+1, N                         !! For each coefficient in the equation . . .
               A(J,K) = A(J,K)- A(I,K)*Multiplier
            end  do
            R(J) = R(J) - R(I)*Multiplier         !! . . . and the right-hand side.
         end  do
      end  do

      !! The second section performs back substitution.

      X(N) = R(N) / A(N,N)                        !! The last solution comes out first.
      do I = N-1, 1, -1
         Partial_Sum = 0
         do J = N, I+1, -1                        !! Substitute the solutions into Equation i.
            Partial_Sum = Partial_Sum + A(I,J)*X(J)
         end  do
         X(I) = (R(I) - Partial_Sum) / A(I,I)     !! . . . thus giving the unknown X(i).
      end  do

   end  subroutine  SIMEQ

!! This subroutine prints the coefficients and right-hand sides of a set of simultaneous
!! equations.
subroutine  PRINT_COEFFICIENTS (A, R)
      !! INCOMING: A = the coefficients;
      !!           R = the right-hand sides.
      complex (kind = Double), dimension ( : , : ), intent (in) :: A
      complex (kind = Double), dimension ( : ), intent (in)     :: R

      integer                             :: J

      print *, "The equations are:"
      do J = 1, ubound(A, 1)
         write (unit=*, fmt="(6(tr2, f5.1, f5.1))" ) A(J, : ), R(J)
      end  do

   end  subroutine  PRINT_COEFFICIENTS

end  module  GROUP

!! This program reads in data for the complex linear simultaneous equation solver
!! subroutine SIMEQ, which it calls.
!! It uses allocatable arrays for holding the coefficients, right-hand sides, and solutions.
!! After reading in N - the number of equations - the arrays are allocated, and then the
!! coefficients and right-hand sides are read in.
!! These are printed by subroutine PRINT_COEFFICIENTS.  SIMEQ is then called.
!! Finally, the complex solutions are printed.
program  SOLVER
   use  GROUP
   implicit  none
   integer                             :: N
   complex (kind = Double), allocatable, dimension ( : , : ) :: A
   complex (kind = Double), allocatable, dimension ( : )     :: R, X
   integer                             :: J, K

   print *, "How many equations do you want to solve?"
   read *, N

   allocate (A(N,N), R(N), X(N) )

   do K = 1, N
      print *, "Please type the coefficients and the right-hand side of equation", K
      read *, A(K, : )  ,  &                      !! Read in the coefficients of equation K.
            R(K)                                  !! Read in the right-hand side for equation K.
   end  do

   call PRINT_COEFFICIENTS (A, R)
   call SIMEQ (A, R, X)

   print *, "The solutions are:"
   do J = 1, N
      print *, "I(", J, ")=", X(J)
   end  do

end  program  SOLVER

!! Figure 4.18  Solving complex linear equations
