!! These modules implement complex integers.
!! Addition, subtraction, multiplication and division are supported.
!! Type conversions supported on assignment are between single and
!! extended precision COMPLEX.

module COMPLEX_INTEGERS
   implicit  none

   type, public                    :: Complex_Integer
      integer                      :: Real_Part
      integer                      :: Imaginary_Part
   end  type  Complex_Integer
   integer, parameter, public      :: Double = selected_real_kind (14)

end  module  COMPLEX_INTEGERS

module  COMPLEX_ARITHMETIC
   implicit  none
   public :: COMPLEX_ADD, COMPLEX_SUBTRACT, COMPLEX_MULTIPLY, COMPLEX_DIVIDE
   public :: COMPLEX_ASSIGNMENT
   public :: COMPLEX_ASSIGNMENT_2, COMPLEX_ASSIGNMENT_3
   public                          :: COMPLEX_ASSIGNMENT_4, COMPLEX_ASSIGNMENT_5
   public                          :: operator (+), operator (-), operator (*), operator (/)
   public                          :: assignment (=)

   !! Interface blocks for supporting addition, subtraction, multiplication,
   !! & division.
   interface operator (+)
      module procedure COMPLEX_ADD
   end  interface

   interface operator (-)
      module procedure COMPLEX_SUBTRACT
   end  interface

   interface operator (*)
      module procedure COMPLEX_MULTIPLY
   end  interface

   interface operator (/)
      module procedure COMPLEX_DIVIDE
   end  interface

   !! An interface block for supporting ordinary assignment, and type conversions
   !! on assignment to/from COMPLEX and extended precision complex.
   interface assignment (=)
      module procedure COMPLEX_ASSIGNMENT
      module procedure COMPLEX_ASSIGNMENT_2
      module procedure COMPLEX_ASSIGNMENT_3
      module procedure COMPLEX_ASSIGNMENT_4
      module procedure COMPLEX_ASSIGNMENT_5
   end  interface

contains

   !! Define ordinary addition between complex integer variables.
   function  COMPLEX_ADD (A, B) result (Summation)
      use COMPLEX_INTEGERS

      type (Complex_Integer), intent (in) :: A, B

      type (Complex_Integer)        :: Summation

      Summation%Real_Part = A%Real_Part + B%Real_Part
      Summation%Imaginary_Part = A%Imaginary_Part + B%Imaginary_Part

   end  function  COMPLEX_ADD

   !! Define ordinary subtraction between complex integer variables.
   function  COMPLEX_SUBTRACT (A, B) result (Difference)
      use COMPLEX_INTEGERS

      type (Complex_Integer), intent (in) :: A, B

      type (Complex_Integer)        :: Difference

      Difference%Real_Part = A%Real_Part - B%Real_Part
      Difference%Imaginary_Part = A%Imaginary_Part - B%Imaginary_Part

   end  function  COMPLEX_SUBTRACT

   !! Define ordinary multiplication between complex integer variables.
   function COMPLEX_MULTIPLY (A, B) result (Prod)
      use COMPLEX_INTEGERS

      type (Complex_Integer), intent (in) :: A, B

      type (Complex_Integer)        :: Prod

      !! Use extended precision so as to avoid intermediate overflow.
      Prod%Real_Part = real(A%Real_Part, Double) * B%Real_Part - &
                real(A%Imaginary_Part, Double) * B%Imaginary_Part
      Prod%Imaginary_Part = real(A%Imaginary_Part, Double) * B%Real_Part + &
                real(A%Real_Part, Double) * B%Imaginary_Part

   end  function  COMPLEX_MULTIPLY

   !! Define ordinary division between complex integer variables.
   function COMPLEX_DIVIDE (A, B) result (Quotient)
      use COMPLEX_INTEGERS

      type (Complex_Integer), intent (in) :: A, B
      type (Complex_Integer)        :: Quotient

      complex (kind=Double)         :: CMPLX_A, CMPLX_B

      CMPLX_A = cmplx ( A%Real_Part, A%Imaginary_Part)
      CMPLX_B = cmplx ( B%Real_Part, B%Imaginary_Part)

      CMPLX_A = CMPLX_A/CMPLX_B

      Quotient%Real_Part = int(CMPLX_A)
      Quotient%imaginary_Part = int(CMPLX_B)

   end  function  COMPLEX_DIVIDE

   !! This subroutine implements assignment statements for complex integers.
   subroutine COMPLEX_ASSIGNMENT (A, B)
      use COMPLEX_INTEGERS

      type (Complex_Integer), intent (out)   :: A
      type (Complex_Integer), intent (in) :: B

      A%Real_Part = B%Real_Part
      A%Imaginary_Part = B%Imaginary_Part

   end  subroutine  COMPLEX_ASSIGNMENT

   !! This subroutine implements assignments to complex integer variables
   !! from COMPLEX.
   subroutine COMPLEX_ASSIGNMENT_2 (A, B)
      use COMPLEX_INTEGERS

      type (Complex_Integer), intent (out)   :: A
      complex, intent (in)          :: B

      A%Real_Part = real(B)
      A%Imaginary_Part = aimag(B)

   end  subroutine  COMPLEX_ASSIGNMENT_2

   !! This subroutine implements assignments to complex integer variables
   !! from extended precision COMPLEX.
   subroutine COMPLEX_ASSIGNMENT_3 (A, B)
      use COMPLEX_INTEGERS

      type (Complex_Integer), intent (out)   :: A
      complex (kind=Double), intent (in)  :: B

      A%Real_Part = real(B)
      A%Imaginary_Part = aimag(B)

   end  subroutine  COMPLEX_ASSIGNMENT_3

   !! This subroutine implements assignments to COMPLEX variables from
   !! complex integer variables.
   subroutine COMPLEX_ASSIGNMENT_4 (A, B)
      use COMPLEX_INTEGERS

      complex, intent (out)         :: A
      type (Complex_Integer), intent (in) :: B

      A = cmplx(B%Real_Part, B%Imaginary_Part)

   end  subroutine  COMPLEX_ASSIGNMENT_4

   !! This subroutine implements assignments to extended precision COMPLEX variables from
   !! complex integer variables.
   subroutine COMPLEX_ASSIGNMENT_5 (A, B)
      use COMPLEX_INTEGERS

      complex (kind=Double), intent (out) :: A
      type (Complex_Integer), intent (in) :: B

      A = cmplx (B%Real_Part, B%Imaginary_Part, Double)

   end  subroutine  COMPLEX_ASSIGNMENT_5

end  module  COMPLEX_ARITHMETIC

!! Figure 4.21  Modules for implementing complex integer arithmetic.
