module GROUP
   implicit none
   public                          :: POLY
   integer, parameter, public      :: Double = selected_real_kind (14)
   integer, parameter, public      :: Extended =                                       &
      (abs(min(max(selected_real_kind(18), -4), 0) + 2)/2 ) * selected_real_kind(18) + &
      (abs(min(max(selected_real_kind(14), -4), 0) + 2)/2 -                            &
       abs(min(max(selected_real_kind(18), -4), 0) + 2)/2 ) * selected_real_kind(14)

contains

   !! This procedure finds the N complex roots of a complex polynomial of degree N, using the
   !! Newton-Raphson iteration method to locate a root, and Horner's method to deflate the
   !! polynomial.
   subroutine  POLY (A, Roots)
      !! INCOMING: A     = an array of coefficients A(0) to A(n) of the polynomial
      !!                   A(n)*Z**n + A(n-1)*Z**(n-1) + . . . . + A(0)
      !! OUTGOING: Roots = an array of complex roots of the polynomial, in
      !!                   Roots(1) to Roots(n).

      complex (kind = Double), dimension (0: ), intent (in out) :: A
      complex (kind = Double), dimension ( : ), intent (out)    :: Roots

      complex (kind = Double), dimension (0:ubound(A, 1))       :: B !! For the deflated polynomial.
      integer                             :: N
                                                        !! Initially the degree of the polynomial, but
                                                        !! subsequently the degree of the reduced polynomial.
      complex (kind = Extended)    :: FPz,  &           !! To hold the value of the polynomial f(z).
                                      DPz               !! To hold the value of the derivative f'(z).
      complex (kind = Double)      :: Z,    &
                                      Delta_Z           !! To hold FPz/DFz.
      real, parameter              :: Eps = 0.00000001
                                                        !! Threshold for convergence.
      integer                      :: Num_Steps
      integer                      :: J

      N = ubound(A, 1)                            !! The degree of the polynomial.

      do N = N, 1, -1                             !! For each root of the polynomial . . . .

         Z = (0.1, 0.1)                           !! An initial guess for a solution.

         do Num_Steps = 1, 50                     !! Limit the iterations to 50.
            FPz = 0
            DPz = 0

            !! We evaluate the polynomial f(z) and its derivative f '(z).
            do J = N, 1, -1
               FPz = FPz*Z + A(J)                 !! Adds the term in A(j), with Z**J, to FPz,
                                                  !! en route to evaluating F(Z).
               DPz = DPz*Z + J*A(J)               !! Adds the term in A(j), with Z**(J-1),
                                                  !! en route to evaluating the derivative f'(z).
            end  do
            FPz = FPz*Z + A(0)
            !! At this point, FPz holds the value of the function at the point Z,
            !! and DPz holds the value of the derivative at the point Z.

            if (abs(real(DPz)) + abs(aimag(DPz)) == 0) then
               print *, "The differential evaluates to zero at the point ", Z
               print *, "The Newton-Raphson iteration has been abandoned."
               return
            end  if
            Delta_Z = FPz/DPz                     !! Calculate f(z)/f '(z), for the Newton-Raphson
                                                  !! formula.
            Z = Z - Delta_Z                       !! To complete the Newton-Raphson formula,
                                                  !! z = z - f(z)/f'(z).

            if (abs (real(Delta_Z)) + abs (aimag(Delta_Z)) < Eps) then
               exit                               !! We met the convergence criteria.
            end  if
         end  do

         Roots(N) = Z                             !! Save a root of the polynomial.

         !! DEFLATE THE POLYNOMIAL.
         B(N) = A(N)
         do J = N-1, 1, -1                        !! Prepares B(N-1) to B(1).
            B(J) = A(J) + B(J+1)*Z
         end  do

         A(0:N-1) = B(1:N)                        !! Move the coefficients to A(0) to A(N-1).

      end  do

   end  subroutine  POLY

end  module  GROUP

program  POLY_ROOTS
   use  GROUP
   implicit  none
   complex (kind = Double), dimension ( : ), allocatable :: A     !! For coefficients of the polynomial.
   complex (kind = Double), dimension ( : ), allocatable :: Roots !! For the roots.
   integer                          :: J, N

   print *, "What is the degree of the polynomial?"
   read *, N

   allocate (A(0:N), Roots (N))

   print *, "Please type the ", N+1,            &
         " coefficients, starting with the coefficient for the greatest power: "

   do J = N, 0, -1
      print *, "Coefficient of z**", J, "="
      read *, A(J)
   end  do

   print *, "The coefficients are: "
   print *, A

   call POLY (A, Roots)
   print *, "The roots are: "
   print *, Roots

end  program  POLY_ROOTS

!! Figure 4.6  Locating the roots of a polynomial
