module GROUP
   implicit none
   public                          :: B, NEWC
   integer, parameter, public       :: Double = selected_real_kind (14)

contains

   !! This subroutine evaluates the given polynomial of degree N at the point Z, using Horner's
   !! method.
   subroutine  B (FABS, Z, FZ, N, A)
      !! INCOMING: Z    = the point at which the polynomial is to be evaluated.
      !!           N    = the degree of the (reduced) polynomial.
      !!           A    = the N+1 coefficients of the (reduced) polynomial.
      !! OUTGOING: FZ   = the value of the polynomial at the point Z.
      !!           FABS = the square of the absolute value of the complex number FZ.
      real (kind = Double),         intent (out) :: FABS
      complex (kind = Double),      intent (in)  :: Z
      integer, intent (in)                       :: N
      complex (kind = Double), dimension (0: ), intent (in) :: A
      complex (kind = Double), intent (out)      :: FZ

      integer                          :: I

      FZ = A(0)
      do I = 1, N                                 !! Evaluate the polynomial, using Horner's method.
         FZ = FZ*Z + A(I)
      end  do

      FABS = real(FZ)**2 + aimag(FZ)**2           !! The square of the absolute value of FZ.
   end  subroutine  B

   !! This procedure computes the N complex roots of the complex polynomial of degree N,
   !! namely, A(0)*z**N + A(1)*z**(N-1) + . . . + A(N-1)*z + A(N).
   subroutine  NEWC(A)
      !! INCOMING: A(0:N)  = the complex coefficients of the polynomial.
      !! OUTGOING: A(1:N)  = the complex roots of the polynomial.
      complex (kind = Double), dimension (0: ), intent (in out)   :: A

      integer                      :: N           !! Initially the degree of the polynomial, but
                                                  !! subsequently the degree of the reduced polynomial
      real (kind=Double)           :: R, U, Minimum, F, FA, F0, R0, Eps, F2, FF
      complex (kind = Double)      :: Z0, F0Z, Z, DZ, F1Z, FZ, V, W, FW
      complex (kind = Double), dimension (0:ubound(A,1)-1)  :: A1
      logical                       :: Stage1, Div2
      integer                       :: J, K
      real (kind=Double), parameter :: Unity = 1.0_Double

      N  = ubound(A,1)                            !! The degree of the polynomial.

find_root:                       &                !! Each time around this loop locates one root . . .
      do
         if (N <= 2) then
            exit
         end  if

         do K = 0, N-1                            !! Differentiate the polynomial.
            A1(K) = A(K)*(N-K)
         end  do

         !! Form an initial estimate of the root. Find the minimum of a(n)/a(k),
         !! & take the n-th root as the initial approximation.
         R = abs(A(N))
         Minimum = (R/abs(A(0))) ** (Unity/N)

         do K = 1, N-1
            if (A(K) /= 0) then
               U = (R/abs(A(K))) ** (Unity/(N-K))
               if (U < Minimum) then
                  Minimum = U
               end  if
            end  if
         end  do

         Z0 = 0
         F0 = real(A(N))**2 + aimag(A(N))**2
         FF = F0
         F0Z = A(N-1)
         if (A(N-1) == 0) then
            Z = 1
         else
            Z = -A(N)/A(N-1)
         end  if
         Z = Z/abs(Z)*Minimum/2
         DZ = Z
         call B(F, Z, FZ, N, A)
         R0 = 2.5*Minimum
         Eps = 4*N**2 / 16.0**26 * F0

         do
            if ((Z + DZ == Z) .or. (F <= Eps) ) then
               exit
            end  if
            call B(U, Z, F1Z, N-1, A1)            !! Evaluate the differentiated polynomial at Z, F'(Z)
            if (U == 0) then                      !! The derivative at Z is zero.
               DZ = DZ*5*(0.6, 0.8)               !! Make a different Delta Z, and try again . . .
            else
               DZ = FZ/F1Z                        !! Delta Z = F(Z)/F'(Z), an increment for Newton's
                                                  !! formula.
               V = (F0Z - F1Z)/(Z0 - Z)
               F2 = real(V)**2 + aimag(V)**2
               Stage1 = (F2/U > U/F/4) .or. (F /= FF)
               R = abs(DZ)
               if (R > R0) then
                  DZ = DZ*R0/R*(0.6, 0.8)
               end  if
               R0 = 5*R
            end  if
            Z0 = Z
            F0 = F
            F0Z = F1Z

            do
               W = Z0 - DZ
               Z = W
               call B(F, Z, FZ, N, A)             !! Evaluate the polynomial at Z
               FF = F
               if (Stage1) then
                  J = 1
                  Div2 = F > F0
                  do
                     if (J > N) then
                        exit
                     end  if
                     if (Div2) then
                        DZ = DZ/2
                        W = Z0 - DZ               !! Completing Newton's formula.
                     else
                        W = W - DZ                !! Completing Newton's formula.
                     end  if
                     call B(FA, W, FW, N, A)      !! Evaluate the polynomial at W.
                     if (FA < F) then
                        F = FA
                        FZ = FW
                        Z = W
                        J = J + 1
                        if (Div2 .and. (J == 3)) then
                           DZ = DZ*(0.6, 0.8)
                           Z = Z0 - DZ
                           call B(F, Z, FZ, N, A) !! Form F(Z).
                           J = N + 1
                        end  if
                     else
                        J = N + 1
                     end  if
                  end  do
               end  if
               if ((R < abs(Z) / 16**6) .and. (F >= F0)) then
                  Z = Z0
                  DZ = DZ*(0.3, 0.4)
                  if (Z + DZ /= Z) then
                     cycle
                  end  if
               end  if
               exit
            end  do
         end  do

         !! Deflate the polynomial, by application of the formula a(k) = a(k) + a(k-1)*Z.
         V = 0
         do K = 0, N-1
            V = V*Z + A(K)
            A(K) = V
         end  do
         A(N) = Z                                 !! Store the present root Z . . .
         N = N - 1                                !! & move on to find the next one.
      end  do  find_root

      if (N == 1) then                            !! The original polynomial was of degree 1.
         A(1) = -A(1)/A(0)
      else                                        !! The polynomial has been reduced to degree 2.
         if (A(1) == 0) then                      !! The polynomial is of the form a(0)*z**2 + a(2).
            A(1) = sqrt(A(2)/A(0))
            A(2) = -A(1)
         else                                     !! The polynomial is of the form of a quadratic.
                                                  !! Use the traditional formula for roots of a quadratic.
            V = sqrt(1 - 4*A(0)*A(2)/A(1)**2)
            if (real(V) < 0) then
               A(1) = (-1-V)*A(1)/(2*A(0))
            else
               A(1) = (-1+V)*A(1)/(2*A(0))
            end  if
            A(2) = A(2)/(A(0)*A(1))
         end  if
      end  if

   end  subroutine  NEWC

end  module  GROUP

program  SOLVE
   use  GROUP
   implicit  none
   integer                         :: N
   complex (kind = Double), allocatable,  &
      dimension ( : )              :: A           !! For coefficients of the polynomial.
   integer                         :: J

   print *, "What degree is the polynomial?"
   read *, N

   allocate (A(0:N))

   print *, "Please type the ", N+1,      &
      " coefficients, starting with the coefficient for the greatest power: "
   do J = 0, N
      print *, "Coefficient of z**", N-J, "="
      read *, A(J)
   end  do

   print *, "The coefficients are:"
   print *, A
   call NEWC (A)
   print *, "The roots are:"
   do J = N, 1, -1
      print *, "A(", J, ")=", A(J)
   end  do

end  program  SOLVE

!! Figure 4.7  Locating the roots of a polynomial equation by Newton's method
