module  GROUP
   implicit  none

   public :: FOUR1
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine forms the discrete Fourier transform of the given data in the array F,
!! or it forms N times the inverse discrete Fourier transform, depending on Mode.
subroutine  FOUR1 (F, N, Mode)
      !! INCOMING: F    =  a one-dimensional array of N data readings of type COMPLEX.
      !!           N    =  the number of data readings in array F.  N must be a power of 2.
      !!           Mode =  1 causes F to be replaced by its discrete Fourier transform;
      !!                   -1 causes F to be replaced by N times its inverse discrete Fourier
      !!                   transform.
      !! OUTGOING: F    =  discrete or inverse discrete Fourier transform, times N, according
      !!                   to the value of Mode.
      complex, dimension (0: ), intent(in out) :: F
      integer, intent (in)                     :: N, Mode

      integer                       :: Mmax
      integer                       :: I, J, K, M
      integer                       :: Inc, Step
      complex                       :: Temp, TF
      real (kind=Double)            :: Theta
      complex (kind = Double)       :: W, WP
      integer                       :: NM1
      integer                       :: Carry
      real (kind=Double), parameter :: Two_PI = 6.283185307179593_Double

      !! Section to verify that N is an integral power of 2.
      NM1 = N - 1
      if ((iand(N, NM1) /= 0) .or. (N <= 1)) then !! N is not a power of 2.
         print *, "The number of data points must be an exact power of 2."
         print *, "The number of data points is ", N
         return
      end  if

      !! The bit-reversal section, which re-orders the data, based on the reversed bit pattern of
      !! the position of the original data.  E.g., data at positions J = 0, 1, 2, 3, 4, 5, 6, 7
      !! are re-ordered to positions 0, 4, 2, 6, 1, 5, 3, 7.
      !! This loop actually generates in K all the integers from 1 to N, by adding 1 to K each
      !! time around the loop.  Each K is the mirror image of a normal binary integer, and each
      !! addition is the mirror image of normal binary addition.
      K = ishft(N, -1)                            !! K = N/2.
      do J = 1, N-2                               !! Starts at 1, finishes at N-2,
                                                  !! because positions 0 and N-1 never have a swap.
         if (K > J) then                          !! Exchange the complex numbers at positions J
                                                  !! and K (but only if we haven't already
            Temp = F(K)                           !! swapped them).
            F(K) = F(J)
            F(J) = Temp
         end  if
         !! Section to increment in binary in mirror-image form, to obtain the next value of J.
         Carry = ishft(N, -1)
         do
            if (iand(Carry, K) == 0) then
               K = iand(K, Carry-1)               !! Mask in the "upper" bits (actually the lower bits).
               K = ior(K, Carry)                  !! Complete the sum, by Or-ing in the Carry bit.
               exit
            end  if
            Carry = ishft (Carry, -1)             !! Propagate the carry to the right.
         end  do
      end  do

      !! The Danielson-Lanczos algorithm.
      Mmax = 1
      Inc = 2
      Step = 1
      do
         if (Mmax >= N) then                      !! Calculates transforms of length 2, 4, 8, 16, ... N.
            exit
         end  if
         Theta = Mode*Two_PI / (Mmax*2)
         !! Initialization for the trigonometric recurrence.
         WP = cmplx( -2 * sin(Theta/2)**2, sin(Theta), kind=Double)
         W = 1.0

         do M = 0, Mmax-1
            do I = M, N-1, Inc
               !! The Danielson-Lanczos formulas.
               TF = W * F(I+Step)
               F(I+Step) = F(I) - TF
               F(I) = F(I) + TF
            end  do
            W = W*WP + W                          !! A simple trigonometric recurrence relation for
                                                  !! multiple angles avoids the need to recompute
                                                  !! sines.  Calculates exp(i*(k+1)*Theta) from
                                                  !! exp(i*k*Theta), using the short-cut formula:
                                                  !! exp(i*k*Theta)*(C+i*S)+exp(i*k*Theta),
                                                  !! where C+i*S = -2*sin(Theta/2)**2+i*sin(Theta) is
                                                  !! WP, and exp(i*k*Theta) is W.
         end  do

         Inc = Inc*2
         Step = 2*Step                            !! To double the length of the next transform.
         Mmax = Mmax * 2
      end  do

end  subroutine  FOUR1

!! Figure 5.1  Algorithm to perform a Fast Fourier Transform

end  module  GROUP
