module  GROUP
   implicit  none

   public :: FFT_REAL
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine calculates the Fourier transform of a set of N floating-point data points,
!! held in the data array F.
!! The Fourier transform of those data points is stored in the array CD of N/2 floating-
!! point complex elements.
subroutine  FFT_REAL (F, CD)
      !! INCOMING: F  = An array of N floating-point data points.
      !!                The size of this array must be a power of 2.
      !! OUTGOING: CD = the positive frequency half of the complex Fourier transform of
      !!                the incoming array F.  Contains N/2 elements.

      real, dimension ( : ), intent (in out) :: F
      complex, dimension ( : ), intent (out) :: CD

      integer                       :: I, J, K, N
      real (kind=Double)            :: Theta
      real                          :: C1, C2
      complex (kind = Double)       :: W, WP
      complex                       :: H1, H2
      integer                       :: Half_N
      integer                       :: NM1
      integer, parameter            :: One = 1
      real (kind=Double), parameter :: PI = 3.141592653589793_Double

      N = ubound(F, 1)
      if (N /= 2*ubound(CD, 1) ) then
         print *, "**ERROR, The arrays in FFT_REAL are not compatible."
         print *, "The incoming data array has",            &
            N, " elements, while the outgoing array has",   &
            ubound(CD, 1), " elements."
         return
      end  if

      !! Section to check that N is an integral power of 2.
      NM1 = N - 1
      if ( (iand(N, NM1) /= 0) .or. (N <= 1) ) then !! N is not a power of 2.
         print *, "The number of data points must be an exact power of 2."
         print *, "The number of data points is ", N
         return
      end  if

      Half_N = N/2
      Theta = PI / Half_N
      C1 = 0.5
      C2 = -0.5
      CD = cmplx ( F( : :2), F(2: :2) )           !! Pairs of elements of F become one complex element of CD.
      call FOUR1 (CD, Half_N, One)
      F(1: :2) = real(CD)                         !! and back again into F.
      F(2: :2) = aimag(CD)

      !! Initialization for the trigonometric recurrence.
      WP = cmplx (-2 * sin(Theta/2)**2, sin(Theta), kind=Double)
      W  = cmplx(1.0e0 + real(WP), aimag(WP), kind=Double)
      K  = 1

      do I = 3, N/2-1, 2                          !! A pointer to the left-hand end of the data array F.
         J = N - I + 2                            !! A pointer to the opposite end of the data array F.
         K = K + 1
         !! Gather the values from opposite ends of the data array F.
         H1 = C1 * cmplx ( F(I) + F(J), F(I+1) - F(J+1) )
         H2 = C2 * cmplx (-F(I+1) - F(J+1), F(I) - F(J) )
         !! Recombine the separate transforms.
         CD(K) = H1 + W*H2
         CD(Half_N-K+2) = conjg(H1) - conjg(W*H2)
         W = W*WP + W                             !! A simple trigonometric recurrence relation for
                                                  !! multiple angles avoids the need to recompute
                                                  !! sines.  Calculates exp(i*(k+1)*Theta) from
                                                  !! exp(i*k*Theta), using the short-cut formula:
                                                  !! exp(i*k*Theta)*(C+i*S)+exp(i*k*Theta),
                                                  !! where C+i*S = -2*sin(Theta/2)**2+i*sin(Theta) is
                                                  !! WP, and exp(i*k*Theta) is W.
      end  do

      CD(1) = cmplx( F(1) + F(2), F(1) - F(2) )
                                                  !! Make into 2 statements when F & CD are the same array.

end  subroutine  FFT_REAL

!! Figure 5.3  Subroutine to perform a Fast Fourier Transform of real data

end  module  GROUP
