module  GROUP
   implicit  none

   public :: FFT_SIN
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine performs a Fast Fourier sine transformation of a set of N floating-point
!! data points that are held in the incoming array Y.
subroutine  FFT_SIN(Y)
      !! INCOMING: Y = an array of N floating-point data points.  N must be a power of 2.
      !! OUTGOING: Y = the Fourier sine transform of the data in the incoming array Y
      !! If this procedure is used to perform the inverse sine transform of data, the outgoing array
      !! Y must be multiplied by 2/N.

      real, dimension ( : ), intent (in out) :: Y

      integer                       :: N
      integer                       :: J
      real                          :: Partial_Sum, Y1, Y2
      real (kind=Double)            :: Theta
      complex (kind = Double)       :: W, WP
      integer                       :: NM1
      real (kind=Double), parameter :: PI = 3.141592653589793_Double

      !! Section to check that N is an integral power of 2.
      N = ubound(Y, 1)
      NM1 = N - 1
      if ( (iand(N, NM1) /= 0) .or. (N <= 1) ) then !! N is not a power of 2.
         print *, "The number of data points must be an exact power of 2."
         print *, "The number of data points is ", N
         return
      end  if

      Theta = PI/N
      !! Initialization for the trigonometric recurrence.
      W = 1
      WP = cmplx(-2*sin(Theta/2)**2, sin(Theta), kind=Double)
      Y(1) = 0
      do J = 1, N/2
         W = W * WP + W                           !! The trigonometric recurrence formula.
         Y1 = aimag(W)*(Y(J+1) + Y(N-J+1))
         Y2 = (Y(J+1) - Y(N-J+1)) / 2
         Y(J+1) = Y1 + Y2
         Y(N-J+1) = Y1 - Y2
      end  do

      call REALFT(Y, N, 1)

      Partial_Sum = 0
      Y(1) = Y(1)/2
      Y(2) = 0
      do J = 1, N-1, 2
         Partial_Sum = Partial_Sum + Y(J)
         Y(J) = Y(J+1)
         Y(J+1) = Partial_Sum
      end  do

end  subroutine  FFT_SIN

!! Figure 5.4  Subroutine to perform a Fast Fourier sine Transform

end  module  GROUP
