module  GROUP
   implicit  none

   public :: MMULT
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine forms matrix multiplication of the two matrices A and B, and deposits the
!! product in matrix C.
subroutine  MMULT (A, B, C, Error)
      !! INCOMING: A     = a matrix of M rows and N columns;
      !!           B     = a matrix of N rows and K columns;
      !! OUTGOING: C     = a matrix of M rows and K columns, being the
      !!                   matrix product of A with B.
      !!           Error = an error code:  0 = normal case;
      !!                                   1 = incompatible incoming matrices;
      !!                                   2 = incompatible outgoing matrix.

      real, dimension ( : , : ), intent (in)  :: A, B
      real, dimension ( : , : ), intent (out) :: C
      character (len=*), intent (out)         :: Error

      real (kind=Double)           :: Partial_Sum !! Extended precision for the partial sums.
      integer                      :: I, J, K, L, M, N

      M = ubound (A, 1)                           !! Determine the number of rows in A.
      N = ubound (A, 2)                           !! Determine the number of columns in A.
      J = ubound (B, 1)                           !! Determine the number of rows in B.
      K = ubound (B, 2)                           !! Determine the number of columns in B.

      if (N /= J) then                            !! The incoming matrices are not compatible.
         Error = "1"
         return
      end  if

      J = ubound (C, 1)                           !! Determine the number of rows in C.
      L = ubound (C, 2)                           !! Determine the number of columns in C.
      if ( (J /= M) .or. (L /= K) ) then          !! The outgoing matrix is incompatible with
                                                  !! the matrices A and B.
            Error = "2"
            return
      end  if

      do I = 1, K                                 !! For each of the columns of matrix B . . .
         do L = 1, M                              !! For each of the rows of matrix A . . .
            Partial_Sum = 0
            do J = 1, N                           !! To work along a row of A and a column of B.
               Partial_Sum = Partial_Sum + real (A(L, J), kind=Double) * B(J, I)
            end  do
            C(L, I) = Partial_Sum
         end  do
      end  do

      Error = "0"
   end  subroutine  MMULT

!! Figure 5.6  Matrix Multiplication

end  module  GROUP
