module  GROUP
   implicit  none

   public :: PROD

contains

!! Special matrix multiplication for Transitive Closure.
!! The three arrays are one-dimensional arrays of integers, that are treated like matrices
!! of single-bit elements.  Each element holds BIT_SIZE one-bit elements.
subroutine  PROD (A, B, C, Error)
   !! INCOMING: A     = a matrix of M rows and N "columns";
   !!                   (actually a row vector of M rows of bit strings of length N.)
   !!           B     = a matrix of N "rows" and K columns;
   !!                   (actually a column vector of K columns of bit strings of length N.)
   !! OUTGOING: C     = a matrix of M rows and K "columns", being the
   !!                   matrix product of A with B.
   !!                   Column 1 is at the high end of a word, that is, column 1 is bit 31
   !!                   in a 32-bit word machine.
   !!           Error = an error code:  0 = normal case;
   !!                                   1 = incompatible incoming matrices;
   !!                                   2 = outgoing matrix cannot have more than BIT_SIZE
   !!                           columns.

   integer, dimension ( : ), intent (in)  :: A, B
   integer, dimension ( : ), intent (out) :: C
   character (len=*), intent (out)        :: Error

   integer                         :: I, J, K, M

   M = ubound (A, 1)                              !! Determine the number of rows in A.
   K = ubound (B, 1)                              !! Determine the number of columns in B.

   J = ubound (C, 1)                              !! Determine the number of rows in C.
   if (J /= M) then                               !! The outgoing matrix is incompatible with A.
      Error = "1"
      return
   end  if
!! {XE "BIT_SIZE built-in function"}
   if (K > bit_size(C(1)) ) then                  !! The number of columns in the result cannot
                                                  !! exceed the number of bits in an integer.
      Error = "2"
      return
   end  if

   C = 0
   do I = 1, M
      do J = 1, K
         if ( iand(A(I), B(J) ) /= 0 ) then
            C(I) = ibset (C(I), bit_size(C(1)) - J)
         end  if
      end  do
   end  do

   Error = "0"
end  subroutine  PROD

!! Figure 5.8  Matrix multiplication for transitive closure

end  module  GROUP
