module  M9
   implicit  none

   public :: TRANSITIVE_CLOSURE

contains

!! This procedure computes the Transitive Closure of the Adjacency Matrix (Adj)
!! (improved Warshall's algorithm).
!! The three arrays are one-dimensional arrays of bit strings, that are treated like matrices
!! of single-bit elements.
subroutine  TRANSITIVE_CLOSURE (Adj, Path, Error)
   !! INCOMING: Adj   = the Adjacency matrix of M rows and N "columns";
   !!                   (actually a row vector of M rows of bits of length N.)
   !! OUTGOING: Path  = the Path matrix of N "rows" and K columns;
   !!                   (actually a column vector of K columns of bits of length N.)
   !!                   Column 1 is at the high end of a word, that is, column 1 is bit 31
   !!                   in a 32-bit word machine.
   !!           Error = an error code:  0 = normal case;
   !!                                   1 = incompatible incoming matrices;
   !!                                   2 = outgoing matrix cannot have more than BIT_SIZE
   !!                           columns.

   integer, dimension ( : ), intent (in)  :: Adj
   integer, dimension ( : ), intent (out) :: Path
   character (len=*), intent (out)        :: Error

   integer                                :: I, K, M

   M = ubound (Adj, 1)                            !! Determine the number of rows in A.
   K = ubound (Path, 1)                           !! Determine the number of columns in B.

   if ( (M /= K) ) then                           !! The incoming matrices are not compatible.
      Error = "1"
      return
   end  if

   if (M > bit_size(Adj(1)) ) then                !! As Adj must be square, the number of rows
                                                  !! cannot exceed the number of bits in an integer.
      Error = "2"
      return
   end  if

   Path = Adj
   do K = 1, M
      do I = 1, M
         if (btest (Path(I), bit_size(Path(1)) - K) ) then !! Tests whether Path(I, K) is 1.
            Path(I) = ior (Path(I), Path(K) )
         end  if
      end  do
   end  do

   Error = "0"
end  subroutine  TRANSITIVE_CLOSURE

!! Figure 5.9  Transitive closure algorithm

end  module M9
