!! This program reads in a document from the file DOCUMENT.DAT, reworks each paragraph so as
!! to remove redundant "space" characters, takes words down to the next line when a line is
!! too long, or brings words up from the next line when a line is too short.  At the end of a
!! sentence, two spaces are inserted.  The finished document is written to the file OUTFILE.OUT.
!! The program requests the user to provide parameters for the document:  the line width and
!! the amount of the indent for the first line of the paragraph.

module  VARIABLES

   character (len=200), public     :: Out_Line    !! Where the output lines are built.
   integer, public                 :: Indent, Line_Length
   integer, public                 :: Len_Out_Line!! Holds the working length of Out_Line.
   integer, public                 :: CP          !! Points at the next character at which to
                                                  !! begin a search for a word in the input line.

end  module  VARIABLES

module GROUP
   implicit none
   public                          :: EXTRACT_WORD, BUILD_LINE

contains

!!  include subroutine EXTRACT_WORD from Figure 6.1b 

!! This subroutine extracts the next word from Line, commencing with the CP-th character.
!! The end of the word is indicated by a blank character.
!! When this subroutine terminates, CP is pointing at first blank character at the end of the
!! word (or the end of the line, whichever comes first).

subroutine  EXTRACT_WORD (Line, Word)
   !! INCOMING:   Line  = the content of a line, as read in.
   !! OUTGOING:   Word = the next word extracted from Line.
   use  VARIABLES, only: CP                       !! CP is a pointer to somewhere in Line.
                                                  !! Initially it points to the beginning of Line.
                                                  !! During this procedure, CP points to the initial
                                                  !! letter of a word in Line, then to successive
                                                  !! characters in that word, and finally, to the blank
                                                  !! following that word.  Punctuation characters are
                                                  !! treated as being part of the word.
   character (len=*), intent (in)   :: Line
   character (len=*), intent (out)  :: Word

   integer                         :: J

   !! This section looks for the initial character of the next word.
   do
      if (CP > len (Line) ) then
         exit
      end  if
      if (Line(CP: CP) /= " " ) then
         exit
      end  if
      CP = CP + 1
   end  do

   if (CP > len (Line) ) then                     !! There are no words found during the search, . . .
      Word = " "
      CP = 1
      return                                      !! . . . so return to caller.
   end  if

   !! This section searches for the blank at the end of the word.
   J = index (Line(CP: ), " " )                   !! Look for a blank at the end of the word.
   if (J == 0) then                               !! The word terminates at the end of the line.
      Word = Line(CP: )
      CP = len (Line) + 1                         !! Make CP point past the end of the line.
   else                                           !! The word terminates in a space.
      Word = Line(CP: CP+J-1)
      CP = CP + J                                 !! Adjust CP to point to the blank at the end of the
   end  if                                        !! word.
end  subroutine  EXTRACT_WORD

!!  include subroutine BUILD_LINE from Figure 6.1c 

!! This subroutine builds lines for the output file, by appending words one-at-a-time to the
!! end of the current output line, until the line becomes full.  Then the line is written
!! out.  One space separates each word, except for the end of a sentence, in which case two
!! blanks are included.
subroutine  BUILD_LINE (Word)
   !! INCOMING:   Word = a word, extracted from Line.
   use  VARIABLES

   character (len=*), intent (in)   :: Word

   character (len=1)               :: Ch
   integer                         :: Spaces
   character (len=2), parameter    :: Blanks = "  "

   !! Section to deal with the first word of a paragraph.
   if (Out_Line(1:Len_Out_Line) == " ") then
      Out_Line(Len_Out_Line+1 : ) = Word
      Len_Out_Line = Len_Out_Line + len_trim (Word)
      return
   end  if

   !! Section to deal with the end of a sentence, indicated by a period (.), ? and !.
   Ch = Out_Line (Len_Out_Line: Len_Out_Line)     !! The last character written out.

   if ( (Ch == "." ) .or. (Ch == "?" ) .or. (Ch == "!" ) ) then !! We are at the end of a sentence.
      Spaces = 2                                  !! Allow two spaces at the end of the sentence.
   else
      Spaces = 1                                  !! Allow one space at the end of the previous word.
   end  if

   !! Section to append the word, and to compute the length of the line.
   if (Len_Out_Line + Spaces + len_trim (Word) > Line_Length) then
                                                  !! The line will overflow.
      write (unit=9, fmt="(a)" ) Out_Line(1:Len_Out_Line)   !! Write out what we have . . .
      Out_Line = Word                             !! and start the word on a fresh line.
      Len_Out_Line = len_trim (Word)
   else
      Out_Line(Len_Out_Line+1 : ) = Blanks(1: Spaces) // Word
                                                  !! Append the word to the line.
      Len_Out_Line = Len_Out_Line + Spaces + len_trim(Word)
   end  if
end  subroutine  BUILD_LINE

end  module  GROUP

program  DOC
   use  VARIABLES
   use  GROUP
   implicit  none
   character (len=30)              :: Word
   character (len=200)             :: Line        !! Holds a copy of the line read in.
   integer                         :: Status

   CP = 1

   !! Section to open the input and output files.
   open (unit=8, file="DOCUMENT.DAT", action="read", iostat=Status, status="old" )
   if (Status /= 0) then
      print *, "Error attempting to open the input file"
      stop
   end  if

   open (unit=9, file="OUTFILE.OUT", action="write", iostat=Status, status="new" )
   if (Status /= 0) then
      print *, "Error attempting to open the output file"
      stop
   end  if

   !! Section to obtain from the user the characteristics of the required document
   !! (length of a line, and indent for the first line of a paragraph).
   print *, "This document re-formatting program will make text conform to a standard line width."
   do
      print *, "What length of line do you require? (number of columns):"
      read *, Line_Length
      if (Line_Length >= 20 .and. Line_Length <= 132) then
         exit
      end  if
      print *, "The line length must be from 20 to 132 columns.  Please try again."
   end  do
   print *, "The line length is ", Line_Length, " columns."

   do
      print *, "What indent do you want for the start of each paragraph?"
      read *, Indent
      if (Indent < 0) then
         print *, "The indent cannot be negative."
      else if (Indent + 10 > Line_Length) then
         print *, "The indent is too large for the line width you specified."
      else
         exit
      end  if
   end  do

   print *, "The indent for the first line of each paragraph is ", Indent, " columns."

   Out_Line = " "
   Len_Out_Line = Indent

   !! Section to read in a line of the original document, to determine whether a new
   !! paragraph is required, to call subroutine EXTRACT_WORD to extract a word
   !! from that line, and to call BUILD_LINE to attach that word to the end of the
   !! output line.  BUILD_LINE also outputs the line when it becomes full.
   do
      read (unit=8, fmt="(a200)", iostat=Status) Line
      if (Status < 0) then                        !! End of file 8, so
         exit                                     !! exit main loop.
      end  if

      !! Section to handle a new paragraph.
      if (Line == " " ) then                      !! An empty line signifies the end of a paragraph.
         write (unit=9, fmt="(a)" ) Out_Line (1: Len_Out_Line)
         if (Out_Line /= " " ) then               !! Maintain an empty line between paragraphs.
            write (unit=9, fmt="(a)" ) " "
         end  if
         Out_Line = " "
         Len_Out_Line = Indent
      end  if

      !! Section to handle all the words in an input line, and to attach them to the
      !! output line (outputting them when the line becomes full).
      do
         call EXTRACT_WORD(Line, Word)            !! To obtain one word from the input line.

         if (Word /= " " ) then                   !! We have a word from the input line.
            call BUILD_LINE (Word)                !! To attach one word to the output line
                                                  !! (and to output that line if it becomes full).
         else                                     !! There are no more words in the input line,
            exit                                  !! so quit.
         end  if
      end  do
   end  do

   !! Section to write out the last line.
   if (Out_Line /= " " ) then
      write (unit=9, fmt="(a)" ) Out_Line(1: Len_Out_Line)
   end  if

end  program  DOC

!! Figure 6.1a  The main program for a simple word processor.
