module  VARIABLES

   integer, parameter, public       :: Maxsym = 255 !! 1 position for each character.
   integer, public                  :: Class        !! 0 for a match, 1 for mismatches.
   integer, parameter, public       :: Match = 0,     &
                                       Mismatch = 1
   integer, dimension (0:Maxsym), public  :: T
   integer, public                  :: B
   integer, parameter, public       :: Word_size = bit_size(B)
   integer, public                  :: Mask, Lim, OVMask

end  module  VARIABLES

module GROUP
   implicit none
   public                           :: PATMAT, PREPROCESSING, SEARCH, LOG2

contains

   !! The three procedures PATMAT, PREPROCESSING, and SEARCH together comprise the
   !! Baeza-Yates-Gonnet approximate string search algorithm.
   subroutine  PATMAT (N, Pattern, String, Matches)
      !! INCOMING: N       = number of mismatches permitted between the pattern and the
      !!                     String;
      !!           Pattern = the string for which a match is sought;
      !!           String  = the string being searched.
      !! OUTGOING: Matches = the number of occurrences of Pattern in String,
      !!                     with N mismatches.
      use  VARIABLES

      integer, intent (in)           :: N
      character (len=*), intent (in) :: Pattern
      character (len=*), intent (in) :: String
      integer, intent (out)          :: Matches

      integer                        :: K, M

      K = N
      M = len(Pattern)
      Class = Mismatch
      if (2*K > M) then
         Class = Match
         K = M-K
         if (K < 0) then
            K = 0
         end  if
      end  if

      B = LOG2(K+1) + 1

      if (M*B > Word_size) then
         print *, "The combination of pattern and the number of mismatches" //         &
            " causes the capacity of the program to be exceeded."
         stop
      end  if
      call PREPROCESSING (Pattern, K)

      call SEARCH (String, Pattern, Matches)
   end  subroutine  PATMAT

   !! This procedure performs initialization of tables and processes the pattern.
   subroutine  PREPROCESSING (Pattern, K)
      !! INCOMING: Pattern  = the string for which a match is sought;
      !!       K      = number of mismatches permitted between Pattern and String;
      use  VARIABLES

      character (len=*), intent (in)   :: Pattern
      integer, intent (in)         :: K

      integer                      :: Pat
      integer                      :: J, M
      integer, parameter           :: Ones = huge(Mask),   &
                           Zeros = 0
      integer                      :: TwoBM1
      integer                      :: UP
      integer                      :: Temp

      M = len(Pattern)
      TwoBM1 = ishft(1, B-1)                      !! Form 2**(B-1).
      Lim = ishft(K, (M-1)*B)                     !! Left shift K by (M-1)*B places.
      OVMask = 0
      do J = 1, M                                 !! For each character in the pattern . . . .
         OVMask = ior ( ishft(OVMask, B), TwoBM1)
      end  do

      if (Class == Match) then
         T = 0
      else
         Temp = ishft(1, (M-1)*B)                 !! Form 2**(M-1)*B.
         Lim = Lim + Temp
         T = ishft (OVMask, -B+1)
      end  if

      J = 1
      Pat = 1
      do                                          !! For each character in the pattern . . . .
         if (Pat > len (Pattern)) then
            exit
         end  if
         UP = ichar (Pattern(Pat:Pat))
         if (Class == Match) then
            T(UP) = T(UP) + J
         else
            T(UP) = iand (T(UP), not(J) )
         end  if
         J = ishft(J, B)                          !! Left shift J by B places.
         Pat = Pat + 1
      end  do

      if (M*B == Word_size) then
         Mask = Ones
      else
         Mask = J-1
      end  if

   end  subroutine  PREPROCESSING

   !! The search function, allowing for K mismatches.
   subroutine  SEARCH (String, Pattern, Matches)
      !! INCOMING: String  = the string being searched.
      !!           Pattern = the string for which a match is sought;
      !! OUTGOING: Matches = the number of times that Pattern occurs in String, allowing for
      !!                     K mismatches.
      use  VARIABLES

      character (len=*), intent (in) :: String, Pattern
      integer, intent (out)        :: Matches

      integer                      :: State, Overflow
      integer                      :: PString
      integer                      :: OS

      Matches = 0
      if (Class == Match) then
         State = 0
         Overflow = 0
      else
         State = iand (Mask, not(OVMask) )
         Overflow = OVMask
      end  if

      PString = 1
      do                                          !! For each character in the string . . .
         if (PString > len (String)) then
            exit
         end  if
         OS = ishft(State, B)                     !! Left shift by B places.
         OS = iand(OS, huge(OS) )                 !! Remove the most significant bit.
         State = iand (OS + T(ichar ( String(PString:PString))), Mask)
         Overflow = iand (ior (ishft(Overflow, B), iand(State, OVMask)), Mask)
         State = iand (State, not(OVMask) )
         if (Class == Match) then
            if (ior(State, Overflow) >= Lim) then
               Matches = Matches + 1
            end  if
         else
            if (ior(State, Overflow) < Lim) then
               Matches = Matches + 1
            end  if
         end  if
         PString = PString + 1
      end  do

   end  subroutine  SEARCH

   !! This function computes the logarithm of Value, to base 2.
   function  LOG2 (Value) result (Logarithm)

      integer, intent (in)         :: Value
      integer                      :: Logarithm

      integer                      :: K

      K = 1
      do
         if (K >= Value) then
            Logarithm = K
            return
         end  if
         K = K * 2
         if (K <= 0) then
            Logarithm = 0
            return                                !! Error
         end  if
      end  do

   end  function  LOG2

end  module  GROUP

!! The Baeza-Yates & Gonnet approximate string search algorithm.
program  BAEZA
   use  VARIABLES
   use  GROUP
   implicit  none

   integer                         :: K, Matches

   character (len=34)              :: String
   character (len=4)               :: Pattern

   String = "Time time Rime rime mime rame same"
   Pattern = "rime"
   print *, "The pattern is " // Pattern
   print *, "How many mismatches do you want?"
   read *, K
   call PATMAT (K, Pattern, String, Matches)
   print *, "K=", K
   print *, "No. of matches=", Matches

end  program  BAEZA

!! Figure 6.11  The Baeza-Yates & Gonnet approximate string search
