module  VARIABLES

   integer, public                 :: CP
   character (len=160), public     :: Text

end  module  VARIABLES

module GROUP
   implicit none
   public :: SEARCH_DICTIONARY, ENTER_WORD, PRINT_DICTIONARY, GET_NEXT_WORD

contains

   !! This function performs a search of a list of words, held in array A, for the
   !! given word, held in Key.  A standard binary search is used.

   function  SEARCH_DICTIONARY (A, N, Key) result (Position)
      !! INCOMING: A   = an array of elements;
      !!           N   = the number of words in the dictionary A.
      !!           Key = the word for which the search is conducted.
      !! RETURNS:  the position of the word in the dictionary, if it exists;
      !!           or a negative value, indicating that the word does not exist in the dictionary.
      !!           The magnitude of the returned value is the position in the dictionary where
      !!           the word should be inserted.

      character (len=*), dimension ( : ), intent (in) :: A
      integer, intent (in)           :: N
      character (len=*), intent (in) :: Key
      integer                        :: Position

      integer                        :: L, R, Mid

      L = 0                                       !! Set the left pointer.
      R = N + 1                                   !! Set the right pointer.

      do
         if (L+1 >= R) then                       !! While the pointers haven't met . . .
            exit                                  !! The pointers have met.
         end  if
         Mid = (L+R)/2                            !! Find the mid-point of the region.
         if (Key < A(Mid) ) then                  !! Bring the right pointer to the mid-point.
            R = Mid
         else if (Key > A(Mid) ) then             !! Bring the left pointer to the mid-point.
            L = Mid
         else                                     !! We located the element - Hurrah!
            Position = Mid                        !! Say where it is!
            return
         end  if
      end  do

      Position = -L                               !! The item was not found in the array, but
                                                  !! indicate where it should be placed (negative).
   end  function  SEARCH_DICTIONARY

   !! This procedure enters a new word in the dictionary, at POSITION.
   subroutine  ENTER_WORD (Dictionary, Word, N, Position)
      !! INCOMING: Dictionary = a list of words in alphabetic order;
      !!           Word       = a word to be inserted in the dictionary;
      !!           N          = the number of words already in the dictionary;
      !!           Position   = the position in the dictionary where the new word must be
      !!                        inserted.
      !! OUTGOING: Dictionary = the updated list of words (extended by one word);
      !!           N          = the (updated) number of words in the dictionary.

      character (len=*), dimension ( : ), intent (in out) :: Dictionary
      character (len=*), intent (in) :: Word
      integer, intent (in out)       :: N
      integer, intent (in)           :: Position

      integer                        :: Local_Position

      if (N == 0) then
         Local_Position = 1
      else
         Local_Position = Position
      end  if

      if (N+1 > ubound(Dictionary, 1) ) then
         print *, "The dictionary is full."
         stop
      end  if

      Dictionary(N+1:Local_Position+1:-1) = Dictionary(N:Local_Position:-1)
                                                  !! Make room to insert the word.
      Dictionary(Local_Position) = Word           !! Insert the word.
      N = N + 1                                   !! The number of words in the dictionary increased.
   end  subroutine  ENTER_WORD

   !! This subroutine obtains the next word from the input file ARTICLE.
   subroutine  GET_NEXT_WORD (Word)
      !! OUTGOING: Word = the next word from the input file ARTICLE, or
      !!              blank, indicating the end of the file ARTICLE.
      use  VARIABLES
      character (len=*), intent (out)  :: Word

      logical                      :: Word_Found
      integer                      :: Start_Word, &
                           End_Word
      integer                      :: Status
      character (len=52), parameter :: Alphabet =  &
                     "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

      !! Look for the start of a word.  If necessary, read another line from the input file ARTICLE.
      do
         if (CP > len (Text) ) then
            read (unit=8, fmt="(a160)", iostat=Status) Text
            if (Status < 0) then
               Word = " "                         !! A signal that the end of file has been reached.
               return
            end  if
            CP = 1
         end  if
         Start_Word = scan (Text(CP: ), Alphabet)
         Word_Found = Start_Word > 0
         if (Word_Found) then
            exit
         else                                     !! No word found; therefore
            CP = len(Text) + 1                    !! force a new line to be read in.
         end  if
      end  do
      Start_Word = Start_Word + CP - 1            !! Allow for the start of the search being at a column
                                                  !! other than the first.
      CP = Start_Word                             !! . . . so that both point at the start of the word.

      !! Look for the end of a word.
      End_Word = verify (Text (Start_Word: ), Alphabet)

      if (End_Word == 0) then                     !! The word finished at the very end of the line.
         CP = len (Text)+1
         Word = Text (Start_Word : )              !! Extract the word.
      else                                        !! The word finished before the end of the line.
         End_Word = CP+End_Word-2
         CP = End_Word + 1
         Word = Text (Start_Word : End_Word)
      end  if                                     !! Extract the word.

   end  subroutine  GET_NEXT_WORD

   !! This procedure prints all the words in the dictionary, one to a line.
   subroutine  PRINT_DICTIONARY (Dictionary, N)
      !! INCOMING: Dictionary = a list of words in alphabetic order;
      !!           N          = the number of words already in the dictionary.

      character (len=*), dimension ( : ), intent (in) :: Dictionary
      integer, intent (in)         :: N

      integer                      :: J

      do J = 1, N
         print *, Dictionary(J)
      end  do

   end  subroutine  PRINT_DICTIONARY

end  module  GROUP

!! This procedure contains a loop that is executed once for each word in the file ARTICLE.
program  WORDS
      use  VARIABLES
      use  GROUP
      implicit  none
      character (len=30), dimension (5000) :: Dictionary
      integer                      :: N           !! The number of words currently in the dictionary.
      character (len=30)           :: Word
      integer                      :: Position

      CP = len (Text) + 1
      Text = " "

      open (unit=8, file="ARTICLE.DAT", access="sequential", action="read", status="old" )

      N = 0

      do
         call GET_NEXT_WORD(Word)                 !! Obtains the next word.  A blank word is

         if (Word == " ") then
            exit                                  !! returned when the end of file is reached.
         end  if
         Position = SEARCH_DICTIONARY (Dictionary, N, Word)
         if (Position <= 0) then                  !! It's a new word, so enter it.
            call ENTER_WORD (Dictionary, Word, N, 1-Position)
         end  if

      end  do

      call PRINT_DICTIONARY (Dictionary, N)

end  program  WORDS

!! Figure 6.2  Program to form a list of words
