module  Alpha
   character (len=52), parameter, public :: Alphabet = &
                  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
end  module  Alpha

module GROUP
   implicit none
   public :: DEBLANK, CENTER, EXTRACT_WORD, TRANSLATE, EMPHASIZE_WORD

contains

   !! This function returns the column number of the next non-blank character in Text.  The search
   !! commences from position Pos.
   !! If the remaining characters are blank, the function returns one less than <Pos>.
   function  DEBLANK (Text, Pos) result (Current_Column)
      !! INCOMING: Text = the text to be searched.
      !!           Pos  = the position from which to begin the search.

      character (len=*), intent (in) :: Text
      integer, intent (in)           :: Pos
      integer                        :: Current_Column

      Current_Column = verify (Text(Pos: ), " " ) + Pos - 1

   end  function  DEBLANK

   !! This function places the given characters in Text in the center of a string of width Width
   !! characters.  Blank characters are placed fore and aft.
   !! If the length of the string Text is greater than Width, then Text is returned.
   function  CENTER (Text, Width) result (Centered_Text)
      !! INCOMING:   Text   = the text to be centered;
      !!       Width = the width of text within which <Text> is to be centered.

      character (len=*), intent (in)         :: Text
      integer, intent (in)                   :: Width
      character (len=max(len(Text), Width) ) :: Centered_Text

      integer                                :: DL

      if (len(Text) >= Width) then
         Centered_Text = Text
         return
      end  if
      Centered_Text = " "
      DL = Width - len(Text) + 1
      Centered_Text(DL/2: ) = Text

   end  function  CENTER

   !! Looks for the end of the current word, and returns the word.
   function  EXTRACT_WORD (Text, Pos) result (Word)
      !! INCOMING: Text = the text to be searched.
      !!           Pos  = the position from which to begin the search.
      !! RETURNS:       = the next word.
      use  Alpha
      character (len=*), intent (in) :: Text
      integer, intent (in)           :: Pos

      character (len=verify (Text(Pos : ) // " ", Alphabet)-1) :: Word

      integer                        :: K

      K = len (Word)
      Word = Text(Pos : Pos+K-1)                  !! Extract the word.

   end  function  EXTRACT_WORD

   !! This function translates certain characters in Text.  A list of the characters to be translated is held
   !! in Chfrom.  A list of the corresponding replacement characters is held in Chto.
   !! This function is normally used to translate upper case letters to lower case, & vice versa.
   function  TRANSLATE (Text, Chto, Chfrom) result (Translated_Text)
      !! INCOMING: Text   = the string whose upper-case characters are to be converted to lower
      !!                    case;
      !!           Chto   = characters to replace any matching those in Chfrom;
      !!           Chfrom = any characters in Text that match any in Chfrom will be replaced
      !!                    by corresponding characters in Chto.
      !! RETURNS: a copy of string Text in which certain characters matching those in Chfrom
      !!          have been replaced.
      character (len=*), intent (in) :: Text, Chto, Chfrom
      character (len=len(Text))      :: Translated_Text

      integer                        :: J, K

      Translated_Text = Text
      do J = 1, len(Text)
         do K = 1, len(Chfrom)
            if (Text(J:J) == Chfrom (K:K))  then
               Translated_Text(J:J) = Chto(K:K)
               exit
            end if
         end  do
      end  do

   end  function  TRANSLATE

   !! This subroutine italicizes any word prefixed with an asterisk,
   !! or boldfaces any word prefixed by a hash.
   !! For the former, it prefixes the word with <i> and suffixes it with </i>.
   !! For the latter, it prefixes the word with <b> and suffixes it with </b>.
   subroutine  EMPHASIZE_WORD (Text, Special, Code)
      !! INCOMING: Text    = A line of text to be translated;
      !!           Special = either an asterisk (*) or a hash (#);
      !!           Code    = the corresponding emphasis code -- either "i" or "b".
      !! OUTGOING:`Text    = the input argument Text is modified when highlighting is carried
      !!                     out.

      character (len=*), intent (in out) :: Text
      character (len=*), intent (in)     :: Special, Code

      character (len=len(Text))          :: TText, Word
      character (len=3)                  :: Prefix
      character (len=4)                  :: Suffix
      integer                            :: L, LW, K

      Prefix = "<" // Code // ">"
      Suffix = "</" // Code // ">"

      K = index (Text, Special)
      do
         if (K <= 0 .or. K == len(Text)) then
            exit
         end  if
         TText = Text(1:K)                        !! Take all the text before the word.
         TText(K:K) = "<"                         !! Replace the code with the start of the HTML code.
         Word = EXTRACT_WORD (Text, K+1)          !! Extract the word.
         LW = len_trim (Word)
         TText(K+1: ) = Prefix(2:3) // Word(1:LW) // Suffix !! Append the highlighted word.
         Text = TText(1:K+LW+6) // Text(K+1+LW: )
                                                  !! Join the suffix text to the line.
         L = index (Text(K+LW+7 : ), Special)     !! Look for another code from where we left off.
         if (L > 0) then
            K = L + K + LW + 6                    !! Calculate the position from the start of Text.
         else
            K = 0
         end  if
      end  do

   end  subroutine  EMPHASIZE_WORD

end  module  GROUP

!! This program converts a plain text file to an HTML file.
program  HTML
   use  GROUP
   implicit  none
   character (len=1000)            :: Text
   character (len=60)              :: IFN, OFN
   character (len=1)               :: First_character
   character (len=len(Text))       :: Title
                                                  !! An array for storing indent levels and indent type (UL
   integer, dimension (0:100)      :: Indent      !! The number of columns of indent.
   character (len=1), dimension (0:100) :: Indent_Type !! The corresponding type is either 'U'
                                                  !! (unnumbered list) or 'O' (ordered list).
   integer                         :: Indent_Level!! A pointer to arrays Indent & Indent_Type.
                                                  !! Indent_Level is incremented for each nested
                                                  !! numbered or unnumbered list.  As each such
                                                  !! list is terminated, Indent_Level is decremented.
   integer                         :: Status

   integer                         :: J, K
   character (len=1), parameter    :: TAB = char(9)
   logical                         :: in_numbered_list

   Indent_Level = 0
   Indent(0) = 0
   Indent_Type(0) = "U"

   print *, "This program converts a plain text file to an HTML document."
   print *, "Please type the name of the plain text input file:"
   read (unit=*, fmt="(a)" ) IFN
   IFN = adjustl (IFN)
   print *, "The input file name is " // trim (IFN)

   print *, "Please type the name of the output file where the HTML text is to be written:"
   read (unit=*, fmt="(a)" ) OFN
   OFN = adjustl (OFN)
   print *, "The output file name is " // trim (OFN)

   open (unit=8, file = trim(IFN), access="sequential", action="read", status="old" )
   open (unit=9, file = trim(OFN), access="sequential", action="write", status="new" )

   write (unit=9, fmt="(a)" ) "<html>"            !! Write out the HTML prolog.
   write (unit=9, fmt="(a)" ) "<head>"
   write (unit=9, fmt="(a)" ) "<title>"
   read  (unit=8, fmt="(a160)", iostat = Status) Text
   if (Status < 0) then
      print *, "The input file is empty."
      stop
   end  if
   write (unit=9, fmt="(a)" ) trim (adjustl(Text) )   !! The first line becomes the title.
   write (unit=9, fmt="(a)" ) "</title>"
   write (unit=9, fmt="(3a)" ) "<H2> ", trim (adjustl(Text) ), " </H2>"

   !! Each time around this loop, we read in & process one line of the file to be converted.

MAIN_LOOP:                 &
   do
      read (unit=8, fmt="(a160)", iostat=Status) Text
      if (Status < 0) then                        !! The end of the input file.
         exit
      end  if

      !! Replace any TAB characters with blanks.
      do K = 1, len(Text)
         if (Text(K:K) == TAB) then
            Text(K:K) = " "
         end  if
      end  do

      if (Text == " " ) then
         write (unit=9, fmt="(a)" ) "<p>"
         cycle  MAIN_LOOP
      end  if

      in_numbered_list = .false.
      K = verify (Text, " " )                     !! Look for the first non-blank character.
      if (verify (Text(K:K), "123456789") == 0) then
         !! We might have a numbered list.
         J = verify (Text(K: ), "1234567890")
         if (J+K-1 < len (Text) ) then
            if ( (Text(J+K-1:J+K-1) == "." ) .or.  &
               (Text(J+K-1: J+K-1) == " " ) ) then
                  in_numbered_list = .true.
            end  if
         end  if
      end  if

      if ((K == Indent(Indent_Level) .and. (Text(K:K) /= "*")) .and.    &
         (.not. in_numbered_list) ) then          !! End of an unnumbered list structure.
         write (unit=9, fmt="(a)" ) "</" // Indent_Type(Indent_Level) // "L> "
         if (Indent_Level > 0) then
            Indent_Level = Indent_Level - 1       !! Cancel one level.
         end  if
      else if ( K < Indent(Indent_Level) ) then
         write (unit=9, fmt="(a)" ) "</" // Indent_Type(Indent_Level) // "L> "
         if (Indent_Level > 0) then
            Indent_Level = Indent_Level - 1       !! Cancel one level.
         end  if
      end  if

      !! SECTION TO DEAL WITH HEADINGS.
      if (K > 20) then                            !! We have a sub-heading.
         Title = TRANSLATE (Text, " ", TAB)
         Title = adjustl (Text)
         Text = "<H2> " // trim(Title) // " </H2>"
         Text = CENTER (Text, 80)
         write (unit=9, fmt="(a)" ) trim (Text)
         cycle  MAIN_LOOP
      end  if

      !! SECTION TO DEAL WITH UNNUMBERED LISTS (BULLETED LISTS).
      First_character = Text(K:K)
      if (First_character == "*" ) then
         if (K < len(Text) ) then                 !! K+1 is dangerous if we are at the end of a line.
            if (Text(K+1: K+1) == " " ) then
               !! Make a bullet.
               if (K > Indent(Indent_Level) .or.            &
                  (Indent_Type(Indent_Level) /= "U" ) ) then
                  !! INSERT <UL>
                  Text(K:K) = "<"

                  if (K < len(Text) ) then        !! We're before the end of the line.
                     Text(K+1: ) = "UL> " // Text(K+1: )
                  else                            !! We're at the end of the line.
                     Text = Text // "UL>"
                  end  if

                  Indent_Level = Indent_Level + 1
                  Indent(Indent_Level) = K        !! Record the amount of indent.
                  Indent_Type(Indent_Level) = "U" !! and the type of the list.
                  K = K + 4                       !! We inserted 3 extra characters--extra 1 to
                                                  !! advance.
               end  if
               !! INSERT <LI>
               Text(K:K) = "<"
               if (K < len(Text) ) then           !! We're before the end of the line.
                  Text(K+1: ) = "LI>" // Text(K+1: )
               else                               !! We're at the end of the line.
                  Text = Text // "LI>"
               end  if
            end  if
         end  if
      end  if

      !! SECTION TO DEAL WITH NUMBERED LISTS.
      if (verify (First_character, "123456789") == 0) then
         if (K < len(Text) ) then                 !! K+1 is dangerous if we are at the end of a line.
            if (Text(K+1: K+1) == " " ) then
               !! Make a number for an ordered list.
               if (K > Indent(Indent_Level) .or. Indent_Type(Indent_Level) /= "O" ) then
                  !! INSERT <OL>
                  Text(K:K) = "<"

                  if (K < len(Text) ) then        !! We're before the end of the line.
                     Text(K+1: ) = "OL> " // Text(K+1: )
                  else                            !! We're at the end of the line.
                     Text = Text // "OL> "
                  end  if

                  Indent_Level = Indent_Level + 1
                  Indent(Indent_Level) = K        !! Record the amount of indent.
                  Indent_Type(Indent_Level) = "O"
                  K = K + 4                       !! We inserted 3 extra characters--extra 1 to advance.
               end  if
               !! INSERT <LI>
               Text(K:K) = "<"
               if (K < len(Text) ) then           !! We're before the end of the line.
                  Text(K+1: ) = "LI>" // Text(K+1: )
               else                               !! We're at the end of the line.
                  Text = Text // "LI>"
               end  if
            end  if
         end  if
      end  if

      if (First_character == "-") then            !! It's a new paragraph.
         K = DEBLANK (Text, 1)                    !! Find the hyphen.
         Text (K:K) = " "                         !! Delete the hyphen (New Paragraph mark).
         K = DEBLANK (Text, 1)                    !! Find the first non-blank character.
         if (K < 3) then
            Text = "<p>" // Text(K: )             !! Prefixes a New Paragraph mark to the line.
         else
            Text(1:3) = "<p>"                     !! Inserts a New Paragraph mark in the line.
         end  if
      end  if

      !! This section italicizes any word prefixed with an asterisk.
      call EMPHASIZE_WORD (Text, "*", "i")

      !! This section boldfaces any word prefixed with a hash.
      call EMPHASIZE_WORD (Text, "#", "b")

      write (unit=9, fmt="(a)" ) trim (Text)

   end  do  MAIN_LOOP

   !!**************************************************!!
   !! SECTION FOR END-OF-FILE PROCESSING.   !!
   !!**************************************************!!

   !! Close off any unnumbered lists or numbered lists.
   do K = Indent_Level, 1, -1
      write (unit=9, fmt="(a)" ) "</" // Indent_Type(Indent_Level) // "L>"
   end  do

   write (unit=9, fmt="(a)" ) "</body>", "</html>"!! Write out the HTML epilog.

   print *, "HTML document written to the file " // trim (OFN)

end  program  HTML

!! Figure 6.3  Program to convert plain text to an HTML document.
