module M8
   implicit  none
   public :: SEARCH_RABIN

contains

!! This function implements the Rabin-Karp string search algorithm.
function  SEARCH_RABIN (String, Pattern) result (Position)
      !! INCOMING: String  = the string to be searched;
      !!           Pattern = the pattern to be used for searching the string.
      !! RETURNS:  The position of Pattern in the String, if successful; or
      !!           zero, indicating an unsuccessful search.

      character (len=*), intent (in) :: String, Pattern
      integer                        :: Position

      integer, dimension (0:255)     :: Code
      integer                        :: Hash_Pattern, &
                                        Hash_String,  &
                                        Power
      integer                        :: LP, LS,       &
                                        Counter,      &
                                        K, J,         &
                                        UK, US
      integer, parameter             :: Prime_Number = 1234567

      if (len (Pattern) == 0) then                !! Nothing with which to search, so finish.
         Position = 0
         return
      end  if
      K = index (String, Pattern(1:1))            !! Find a character that matches the first
                                                  !! character of the pattern.
      if (K == 0) then                            !! None, so finish.
         Position = 0                             !! The search was unsuccessful.
         return
      end  if
      if (len (Pattern) >len (String) ) then      !! The string is not long enough.
         Position = 0                             !! The search was unsuccessful.
         return
      end  if

      !! SETUP SECTION
      !! Associate an integer with each (different) character in the pattern.
      LP = len (Pattern)
      LS = len (String)
      Code = LP
      Counter = 0

      do J = 1, LP
         UK = ichar (Pattern(J:J))
         if (Code(UK) == LP) then                 !! We have an un-assigned character.
            Code(UK) = Counter
            Counter = Counter + 1
         end  if
      end  do

      !! HASH THE PATTERN; and calculate (LP+1)**(LP-1).
      Power = 1
      Hash_Pattern = Code(ichar (Pattern(1:1)) )
      do J = 2, LP
         Hash_Pattern = Hash_Pattern * (LP+1) + Code(ichar (Pattern(J:J)) )
         Hash_Pattern = modulo (Hash_Pattern, Prime_Number)
         Power = modulo (Power*(LP+1), Prime_Number)
      end  do

      !! HASH ONE FRAME OF THE STRING, starting at position K.
      Hash_String = 0
      do J = K, K + LP - 1
         Hash_String = Hash_String * (LP+1) + Code(ichar (String(J:J)) )
         Hash_String = modulo (Hash_String, Prime_Number)
      end  do

      !! SEARCH THE STRING
      !! We drag the Pattern past characters of STRING.
      !! The subset of characters being considered in STRING is called a frame.
      !! We derive the hash of the string at position K+1 from the hash at position K.
      do
         if (Hash_Pattern == Hash_String) then    !! The hash codes are equal, so compare strings.
            if (Pattern == String(K:K+LP-1) ) then!! We have a match.
               Position = K                       !! The search was successful.
               return
            end  if
         end  if
         if (K+LP > LS) then                      !! We're at the end of the string, without a match.
            Position = 0                          !! The search was unsuccessful.
            return
         end  if
         !! "Move" the string frame one position to the right.
         UK = ichar (String(K:K))                 !! The binary character at position K (the first of
                                                  !! the frame).
         US = ichar (String(K+LP:K+LP))           !! The binary character to the right of the frame.
         Hash_String = (Hash_String - Code(UK)*Power)*(LP+1) + Code (US)
                                                  !! The hash at position K+1 is derived from the
                                                  !! hash at position K.
         Hash_String = modulo (Hash_String, Prime_Number)
         K = K + 1                                !! Move along one position.
         if (K + LP - 1 > LS) then
            exit
         end  if
      end  do

      Position = 0                                !! The search was unsuccessful.
   end  function  SEARCH_RABIN

!! Figure 6.8  The Rabin-Karp string search

end  module  M8
