!! Payroll Program.
!! This payroll program reads in data from an unformatted file containing information
!! about employees.  For each employee, it calculates the weekly gross pay,
!! deductions, tax payable, and net pay.  It also calculates the total gross salary
!! for the year, and total net salary for the year.  The employee data including the
!! updated latter two values are written to an updated employee file.
!! The program prints the employee data obtained from the file, as well as the
!! totals for the year so far.
!! When all the employees have been processed, the program prints the week's
!! totals of gross salary and net salary.

module DT

   type, public                    :: Address
         character (len=20)        :: Street
         character (len=20)        :: Suburb
         integer                   :: Postcode
   end  type  Address

   type, public                    :: Employee
      character (len=20)           :: Name
      type (Address)               :: Addr
      character (len=1)            :: Classification
      integer                      :: Deductions
      integer                      :: Pay_To_Date,   &
                                      Tax_To_Date
   end  type  Employee

end  module  DT

module  GROUP
   implicit  none
   public                          :: DOLLAR

contains

   !! This function returns a copy of the dummy argument, with a dollar sign ($)
   !! before the most significant digit.
   function DOLLAR (Buffer) result (Dollar_Amount)
      !! INCOMING: Buffer = a decimal number, in character string form.
      !! RETURNS:  The same string Buffer, with leading and trailing blanks removed, and
      !!           with a dollar sign in front.
      character (len=*), intent (in)         :: Buffer

      character (len=len_trim(adjustl(Buffer))+1)  :: Dollar_Amount

      Dollar_Amount = "$" // trim( adjustl(Buffer) )
   end  function DOLLAR

end  module  GROUP

program  PAYROLL
   use  DT
   use  GROUP
   implicit  none

   type (Employee)                 :: Sales_Person

   integer                         :: Gross_Total,   &
                                      Net_Total

   integer                         :: Net_Salary, Taxable_Salary, Gross_Salary, Tax
   character (len=10)              :: Buffer
   integer                         :: Status

   open (unit=8, file="EMPLOY.DAT", form="unformatted", action="read", status="old" )
   open (unit=9, file="UPDATED.OUT", form = "unformatted", action="write", status="new" )

   Gross_Total = 0
   Net_Total = 0
   do
      read (unit=8, iostat = Status) Sales_Person
      if (Status < 0) then                        !! The end of the input file.
         exit
      end  if
      !! Compute the gross salary.
      select  case (Sales_Person%Classification)
         case ( "A")
            Gross_Salary = 500
         case ( "B")
            Gross_Salary = 450
         case ( "C")
            Gross_Salary = 350
         case  default
            print *, "***ERROR in classification for " // Sales_Person%Name
            Gross_Salary = 0
      end  select

      Taxable_Salary = Gross_Salary - Sales_Person%Deductions

      !! Calculate the tax.
      select  case (Taxable_Salary)
         case (500: )
            Tax = Taxable_Salary * 0.35
         case (400:499)
            Tax = Taxable_Salary * 0.30
         case (300:399)
            Tax = Taxable_Salary * 0.25
         case  default
            Tax = 0
      end  select

      Net_Salary = Taxable_Salary - Tax

      !! Calculate running totals for the employee for the year.
      Sales_Person%Pay_To_Date = Sales_Person%Pay_To_Date + Gross_Salary
      Sales_Person%Tax_To_Date = Sales_Person%Tax_To_Date + Tax

      write (unit=9) Sales_Person                 !! Write out the updated data.

      write (unit=Buffer, fmt="(i10)" ) Net_Salary
      write (unit=*, fmt="(a, tr1, a, i5, tr1, i5, tr1, i10, tr1, a, tr1, i10, tr1, i10)" )  &
         Sales_Person%Name,                                  &
         Sales_Person%Classification, Gross_Salary, Tax,     &
         Sales_Person%Deductions, DOLLAR(Buffer),            &
         Sales_Person%Pay_To_Date, Sales_Person%Tax_To_Date

      !! Calculate weekly totals for all employees.
      Gross_Total = Gross_Total + Gross_Salary
      Net_Total = Net_Total + Net_Salary

   end  do

   !! Come here for end-of-file processing.

   print *, "Payroll completed normally."
   write (unit=Buffer, fmt="(i10)" ) Gross_Total
   write (unit=*, fmt="(2a)" ) "The gross salary payout is $", adjustl (Buffer)
   write (unit=Buffer, fmt="(i10)" ) Net_Total
   write (unit=*, fmt="(2a)" ) "The total net salary payout is $", adjustl (Buffer)

end  program  PAYROLL

!! Figure 7.2  Reading a sequential unformatted file
