module  Defined_Types

   type, public                    :: Element
      integer                      :: Column
      real                         :: Value
      type (Element),  pointer     :: Link
   end  type  Element

   type, public                    :: XX
      type (Element), pointer      :: PP
   end  type  XX

end  module  Defined_Types

module  GLOBAL
   use  Defined_Types
   private
   type (XX), dimension ( : ), allocatable, public :: P
   type (Element), pointer, public                 :: Tail

end  module  GLOBAL

module GROUP
   implicit none
   public :: DISPLAY_ROW, APPEND_ELEMENT, OUTPUT, GAUSS_ELIMINATION, &
             BACKWARD_SUBSTITUTION, COPY_SUB_LIST

contains

   subroutine  DISPLAY_ROW (Temp)
      use  Defined_Types

      type (Element), pointer      :: Temp

      type (Element), pointer      :: Current

      Current => Temp
      print *
      do
         if ( .not. associated(Current) ) then
            exit
         end  if
         write (unit=*, fmt="(a, i3)" ) " Column", Current%Column
         write (unit=*, fmt="(f6.2)" ) Current%Value
         Current => Current%Link
      end  do
   end  subroutine  DISPLAY_ROW

   !! This procedure appends a node to the linked list whose head pointer is in P(Row).
   subroutine  APPEND_ELEMENT (Row, Col, New_Value)
      use  Defined_Types
      use  GLOBAL

      integer, intent (in)         :: Row, Col
      real, intent (in)            :: New_Value

      type (Element), pointer      :: Current

      allocate (Current)

      Current%Column = Col
      Current%Value = New_Value
      nullify (Current%Link)

      if (.not. associated (P(Row)%PP ) ) then    !! There are no elements in this row yet.
         P(Row)%PP => Current
         Tail => Current
         return
      end  if

      Tail%Link => Current                        !! Make the tail point at the new node.
      Tail => Current                             !! "Tail" points at the new tail.

   end  subroutine  APPEND_ELEMENT

   !! This procedure prints the content of a sparse matrix whose row heads are in P.
   subroutine  OUTPUT (P, RHS)
      use  Defined_Types

      type (XX), dimension ( : ), intent (in)   :: P
      real, dimension ( : ), intent (in)        :: RHS

      integer                      :: Row
      type (Element), pointer      :: Current

      do Row = 1, ubound(P, 1)
         Current => P(Row)%PP
         if ( associated (Current) ) then
            print *, "Row ", Row
            print *
         end  if
         do
            if ( .not. associated (Current) ) then
               exit
            end  if
            write (unit=*, fmt="(a, i3, a)" ) " Column ", Current%Column, ","
            write (unit=*, fmt="(f6.2)" ) Current%Value
            Current => Current%Link
         end  do
         write (unit=*, fmt="(a, f6.2)" ) " RHS=", RHS(Row)
      end  do

   end  subroutine  OUTPUT

   !! This subroutine solves a set of sparse simultaneous equations using Gauss
   !! Elimination.
   subroutine  GAUSS_ELIMINATION (P, RHS)
      use  Defined_Types

      type (XX), dimension ( : ), intent (in out) :: P
      real, dimension ( : ), intent (in out)      :: RHS

      type (Element), pointer      :: Current, Pivot_Col_Ptr, RowK_Col_Ptr
      real                         :: Multiplier
      real                         :: Temp_Val
      integer                      :: I, J, K, N

      N = ubound (P, 1)                           !! The number of equations.

      do I = 1, N-1                               !! For each pivotal equation . . .

         !! For each equation (row) below the pivotal equation . . .
         do J = I+1, N
            if (P(J)%PP%Column > I) then
               cycle
            end  if

            Current => P(J)%PP
            Temp_Val = P(J)%PP%Value
            do
               if ( .not. associated (Current) ) then
                  exit
               end  if
               if (Current%Column >= I+1) then
                  exit
               end  if
               Current  => Current%Link           !! Eliminate any element before column I+1.
               P(J)%PP => Current
            end  do
            !! At this stage, Current is pointing at column I in row J.

            !! Subtract a multiple of row i from row j, thereby setting to zero the coefficients
            !! below the pivot.

            Multiplier = Temp_Val / P(I)%PP%Value

            Pivot_Col_Ptr => P(I)%PP% Link        !! A pointer to the next nonzero column after
                                                  !! column I in the pivotal equation.
            RowK_Col_Ptr => P(J)%PP               !! A pointer to a column of equation J.

   FOREVER_LOOP:  &
            do K = I+1, N
               if (.not. associated (Pivot_Col_Ptr) ) then
                  exit                            !! There's no further work in row J.
               end  if
               if (.not. associated (RowK_Col_Ptr) ) then
                  !! Append the remainder of the pivot row to row J, & scale it.
                  !! Find the tail of row J.
                  Current => P(J)%PP
                  do
                     Current => Current%Link
                     if (.not. associated (Current%Link) ) then
                        exit
                     end  if
                  end  do
                  Current%Link => COPY_SUB_LIST (Pivot_Col_Ptr)
                  Current => Current%Link         !! Moves to the beginning of the sub-list.
                  do
                     Current%Value = -Current%Value * Multiplier
                     Current => Current%Link
                     if (.not. associated (Current) ) then
                        exit
                     end  if
                  end  do
                  exit  FOREVER_LOOP
               else
                  if (RowK_Col_Ptr%Column == Pivot_Col_Ptr%Column) then
                     RowK_Col_Ptr%Value = RowK_Col_Ptr%Value - &
                              Pivot_Col_Ptr%Value * Multiplier
                  else if (RowK_Col_Ptr%Column > Pivot_Col_Ptr%Column) then
                                                  !! Insert a new element in row K, before the current element.
                     allocate (Current)
                     Current = RowK_Col_Ptr       !! Copy the whole record.
                     RowK_Col_Ptr = Pivot_Col_Ptr !! Ditto.

                     RowK_Col_Ptr%Link => Current !! row K.

                     RowK_Col_Ptr%Value = -Pivot_Col_Ptr%Value * Multiplier
                                                  !! Subtract elements.
                     RowK_Col_Ptr%Column = Pivot_Col_Ptr%Column
                  end  if
                  !! ELSE we are finished with row K.
               end  if
               !! Move to the next element in the pivot row and row K.
               if (Pivot_Col_Ptr%Column < RowK_Col_Ptr%Column) then
                  Pivot_Col_Ptr => Pivot_Col_Ptr%Link
               else if (RowK_Col_Ptr%Column < Pivot_Col_Ptr%Column) then
                  RowK_Col_Ptr => RowK_Col_Ptr%Link
               else
                  Pivot_Col_Ptr => Pivot_Col_Ptr%Link
                  RowK_Col_Ptr => RowK_Col_Ptr%Link
               end  if
            end  do  FOREVER_LOOP
            RHS(J) = RHS(J) - RHS(I) * Multiplier !! . . . & the corresponding right-hand side.

         end  do
      end  do

   end  subroutine  GAUSS_ELIMINATION

   !! This procedure performs backward substitution of the solutions, following Gauss Elimination.
   subroutine  BACKWARD_SUBSTITUTION (P, RHS)
      use  Defined_Types

      type (XX), dimension ( : ), intent (in out) :: P
      real, dimension ( : ), intent (in out)      :: RHS

      type (Element), pointer      :: Next
      real                         :: Partial_Sum
      integer                      :: I, N

      N = ubound(P, 1)                            !! The number of equations.

      RHS(N) = RHS(N) / P(N)%PP%Value             !! The last solution comes out first.

      do I = N-1, 1, -1
         Partial_Sum = 0
         Next => P(I)%PP%Link                     !! Moves to the second node in the list for row I.
         !! Form the sum of the products of known solutions and coefficients.
         do
            if ( .not. associated (Next) ) then
               exit
            end  if
            Partial_Sum = Partial_Sum + Next%Value * RHS(Next%Column)
            Next => Next%Link
         end  do
         RHS(I) = (RHS(I) - Partial_Sum) / P(I)%PP%Value !! Form solution X(I).
      end  do

   end  subroutine  BACKWARD_SUBSTITUTION

   !! This function makes a copy of the sub-list pointed at by the argument.
   !! It returns the head of the list created.
   function  COPY_SUB_LIST (Temp_Ptr) result (Head_of_New_List)
      use  Defined_Types

      type (Element), pointer      :: Temp_Ptr
      type (Element), pointer      :: Head_of_New_List

      type (Element), pointer      :: Original, & !! A pointer to the original list.
                        Current                   !! A pointer to elements in the new list,
                                                  !! as they are created.
      type (Element), pointer      :: Head
      type (Element), pointer      :: Tail

      nullify (Head)                              !! Each time we enter, initialize to NULL.
      Original => Temp_Ptr

      do
         if ( .not. associated (Original) ) then
            exit
         end  if
         allocate (Current)
         if ( associated (Head) ) then            !! First time, note the head of the new list.
            Head => Current
            Tail => Current
         else
            Tail%Link => Current                  !! Make the previous node point at the current one.
         end  if
         Current = Original                       !! Copy whole record.
         Original => Original%Link                !! Move to the next element in the original list.
      end  do

      Head_of_New_List => Head
   end  function  COPY_SUB_LIST

end  module  GROUP

!! This program solves a set of sparse simultaneous equations using Gauss elimination
!! & backward substitution.  Linked lists are used for holding the nonzero
!! coefficients.
!!       **** ORGANIZATION ****
!! The matrix of coefficients is stored as a set of linked lists.
!! There is one linked list for each row.  The head of each row is
!! stored in the pointer array P.
!! Each nonzero element of the matrix of coefficients is stored
!! as a node of a linked list.  Each node contains the coefficient,
!! its column number, and a link to the next node.  Nodes in the list are
!! stored in order of increasing column number.
!! The right-hand sides are stored in a vector RHS of the same length
!! as P.
!! During the Gauss elimination process, the lower triangle is set to
!! zero.  As each coefficients beneath the pivot element is set to
!! zero, it is eliminated from the linked list for that row.
program  SPARSE
   use  Defined_Types
   use  GLOBAL
   use  GROUP
   implicit  none
   real, dimension ( : ), allocatable  :: RHS
   real                            :: New_Value
   integer                         :: Current_Row,      &
                                      Current_Col
   integer                         :: Row, Col

   integer                         :: J, N

   open (unit=8, file="Equation.dat", access="sequential", action="read", status="old" )

   print *, "This program solves a set of simultaneous equations using"
   print *, "the Gauss elimination method."
   print *
   print *, "Sparse matrix input."

   print *, "How many equations are there?"
   read (unit=8, fmt=*) N
   print *, "There are ", N, " equations."

   allocate ( RHS(N), P(N) )

   print *, "Please type each element as a triple, that is, ",       &
            "with row number, column number, and value."
   print *, "Please type the elements a row at a time."

   do J = 1, N
      nullify ( P(J)%PP )
   end  do

   Current_Row = 0
   Current_Col = 0

   do

      print *, "Please type a triple:"
      read (unit=8, fmt=*) Row, Col

      if ((Row == 0) .or. (Col == 0)) then
         exit
      end  if

      read (unit=8, fmt=*) New_Value

      if (Row > Current_Row) then                 !! A new row; reset Current_Col to zero.
         Current_Col = 0
      end  if
      if (Row < Current_Row) then
         print *, "Error, row ", Row, " is out of order."
      end  if
      if (Col == Current_Col) then
         print *, "Error, repeated element for column ", Col,  &
            " in row ", Row, "."
      end  if

      if (Col < Current_Col) then
         print *, "Error, column ", Col,              &
            " in row ", Row, " is out of order."
      end  if
      if (Col > N) then
         print *, "Error, the column number ", Col, " exceeds ", N
      end  if
      if (Row > N) then
         print *, "Error, the row number ", Row, " exceeds ", N
      end  if

      call APPEND_ELEMENT (Row, Col, New_Value)

      Current_Row = max (Row, Current_Row)
      Current_Col = max (Col, Current_Col)

      if (Row == 0) then
         exit
      end  if
   end  do

   print *
   print *, "Completed the input of the coefficients."

   print *
   print *, "Please type the right-hand sides:"
   read (unit=8, fmt=*) RHS

   call OUTPUT (P, RHS)

   call GAUSS_ELIMINATION (P, RHS)

   print *, "The coefficients & RHS after elimination:"
   call OUTPUT (P, RHS)

   call BACKWARD_SUBSTITUTION (P, RHS)

   print *, "The solutions are:", RHS

end  program  SPARSE

!! Figure 8.10  Solving a set of equations having sparse coefficients
