module  GROUP
   implicit  none

   public :: MATINV
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine uses the traditional Gauss-Jordan method to invert a matrix.
!! It uses row and column interchanges at each stage so as to bring the element
!! of largest magnitude to the pivot row.
subroutine  MATINV (A, Error)

      real (kind=Double), dimension ( : , : ), intent (in out)    :: A
      character (len=*), intent (out)                             :: Error

      integer, dimension (ubound(A,1)) :: R       !! A temporary for holding row numbers.
      integer, dimension (ubound(A,1)) :: C       !! A temporary for holding column numbers.
      real (kind=Double), dimension (ubound(A,1))  :: TempVec  !! A temporary for holding a row
                                                               !! or column.

      real (kind=Double)           :: Largest, Temp, &
                                      Akk
      integer                      :: N
      integer                      :: I, J, K

      Error = "0"                                 !! To anticipate success.
      N = ubound (A, 1)                           !! The number of equations.
      if (N /= ubound(A,2) ) then                 !! The coefficient matrix is not square.
         Error = "1"
         return
      end  if

      if (N == 1) then                            !! A 1 x 1 matrix is a simple case.
         A(1,1) = 1/A(1,1)
         return
      end  if

      do K = 1, N
         !! Search for the element of largest magnitude.
         R(K) = K
         C(K) = K
         Largest = abs (A(K,K))
         do I = K, N
            do J = K, N
               if (Largest < abs(A(I,J)) ) then
                  Largest = abs (A(I,J))
                  R(K) = I                        !! Note the position of the element of largest
                  C(K) = J                        !! magnitude.
               end  if
            end  do
         end  do

         !! Interchange rows.
         J = R(K)
         if (J > K) then                          !! Interchange row K with row J.
            do I = 1, N
               Temp = A(K,I)
               A(K,I) = A(J,I)
               A(J,I) = Temp
            end  do
         end  if

         !! Interchange columns.
         I = C(K)
         if (I > K) then                          !! Interchange column K with column I.
            do J = 1, N
               Temp = A(J,K)
               A(J,K) = A(J,I)
               A(J,I) = Temp
            end  do
         end  if

         !! Divide column K by the pivot.
         Akk = A(K,K)
         A( : ,K) = -A( : ,K)/Akk

         do I = 1, N
            if (I /= K) then
               !! Effectively subtract a multiple of the pivotal equation K from Equation I.
               do J = 1, K-1
                  A(I,J) = A(I,K)*A(K,J) + A(I,J)
               end  do
               do J = K+1, N
                  A(I,J) = A(I,K)*A(K,J) + A(I,J)
               end  do
            end  if
         end  do

         !! Divide by the row pivot.
         A(K, : ) = A(K, : )/Akk
         A(K,K) = 1/Akk                           !! Replace the pivot by its reciprocal.

      end  do

      !! Restore rows and/or columns to their original positions.
      do K = N-1, 1, -1
         I = R(K)
         if (I > K) then                          !! Exchange row K with row I.
            TempVec = A( : ,K)
            A( : ,K) = A( : ,I)
            A( : ,I) = TempVec
         end  if
         J = C(K)
         if (J > K) then                          !! Exchange column K with column J.
            TempVec = A(K, : )
            A(K, : ) = A(J, : )
            A(J, : ) = TempVec
         end  if
      end  do

end  subroutine  MATINV

!! Figure 8.12  Subroutine for matrix inversion using the improved Gauss-Jordan method

end  module  GROUP
