module  GROUP
   implicit  none

   public :: PRINT_MATRIX
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This procedure prints the elements of a floating-point array in a table.  The elements
!! are printed in fixed-point form.  The user specifies the number of significant digits
!! to be printed.  This procedure prepares an edit descriptor suitable for printing
!! all the values.  It automatically commences a new line whenever it becomes full.
!! The parameter Line_length is set to 80, but may be changed as required.  The line
!! length may instead be supplied as an argument to the procedure.
subroutine  PRINT_MATRIX (A, No_of_Digits)

      real (kind=Double), dimension ( : , : ), intent (in)  :: A
      integer, intent (in)                                  :: No_of_Digits

      integer, parameter           :: Line_length = 80
      real (kind=Double)           :: X, Scaling
      integer                      :: NNPL
      character (len=10)           :: Edit
      integer                      :: w, d
      integer                      :: Scale_Factor
      integer                      :: J, K, L, M, N

      M = ubound(A, 1)                            !! The number of rows.
      N = ubound(A, 2)                            !! The number of columns.

      X = maxval( abs(A) )                        !! The magnitude of the largest element.

      if (X == 0) then
         L = No_of_Digits
      else
         L = ceiling(log10 (X))
      end  if

      NNPL = (Line_length+1)/(No_of_Digits+4)     !! The number of numbers per line.

      if (No_of_Digits < L) then                  !! The value is large, and needs to be scaled down.
         Scale_Factor = -(L-No_of_Digits)
         print *, "Multiply the following values by 10**", -Scale_Factor
         Scaling = 10.0_Double**Scale_Factor
         L = No_of_Digits
      else                                        !! Normal case.
         Scale_Factor = 0
         Scaling = 1
      end  if

      w = No_of_Digits+3                          !! The number of print columns in the field.
      d = No_of_Digits-L                          !! The number of digits after the decimal point.

      if (d+3 >= w) then                          !! The value is small, and needs to be scaled up.
         Scale_Factor = Scale_Factor + (d - w + 3)
         if (Scale_Factor /= 0) then
            print *, "Divide the following values by 10**", Scale_Factor
         end  if
         Scaling = 10.0_Double**Scale_Factor
         d = w - 3
      end  if

      do J = lbound(A,1), M
         do K = lbound(A, 2), N
            write (unit=Edit, fmt="(a, i3.3, a, i3.3, a)" ) "(f", w, ".", d, ")"
                                                  !! Fabricate an edit descriptor of the form Fw.d
            write (unit=*, fmt=Edit, advance="no" ) A(J, K)*Scaling
            if (modulo (K, NNPL) == 0) then       !! The line is full; start a new line.
               write (unit=*, fmt=*)
            else                                  !! Write out one space . . .
               if (K /= N) then                   !! . . . except when we are at the end of a row.
                  write (unit=*, fmt="(a)", advance="no" ) " "
               end  if
            end  if
         end  do
         write (unit=*, fmt=*)
      end  do

end  subroutine  PRINT_MATRIX

!! Figure 8.13  Printing a matrix

end  module  GROUP
