module  GROUP
   implicit  none

   public :: SIMEQ
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine solves N linear algebraic simultaneous equations using
!! Gauss elimination and backward substitution.
!! This version uses array sections.
subroutine  SIMEQ (A, R, X)                       !!{XE "algorithm: simultaneous equations"}
      !! INCOMING:   A = a set of N coefficients for each of N equations, stored in matrix form.
      !!            The coefficients are changed.
      !!            The coefficients of the first equation are stored in the first row; those of
      !!            the second in row 2, & so on.
      !!       R = a vector of N right-hand sides for the above equations.  The
      !!            right-hand side of the first equation is stored in R(1), etc.
      !!            These also are changed;
      !! OUTGOING:   X = the solutions of the above equations.  The first solution is stored in X(1),
      !!             etc.

      real (kind=Double), dimension ( : , : ), intent (in out) :: A
      real (kind=Double), dimension ( : ), intent (in out)  :: R
      real (kind=Double), dimension ( : ), intent (out)     :: X

      real (kind=Double)           :: Total, Multiplier
      integer                      :: N
      integer                      :: I, J

      !! First check that the arrays are compatible.
      N = ubound(A, 1)                            !! The number of equations.
      if ( (N /= ubound(A, 2)) .or. (N /= ubound (R, 1)) .or. (N /= ubound (X, 1)) ) then
         print *, "**ERROR, the dimensions of the arrays for the " //   &
            " Gauss equation solver are incompatible."
         print *, "The coefficient matrix is of size ", N,           &
            " by ", ubound(A, 2)
         print *, "The vector of right-hand sides is of size ",         &
            ubound (R, 1)
         print *, "The solution vector X is of size ",               &
            ubound (X, 1)
         print *, "The solution is abandoned."
         return
      end  if

      !! The first section performs Gauss elimination,
      !! to reduce the equations to upper triangular form.

      do I = 1, N-1                               !! For each pivotal equation . . .
         do J = I+1, N                            !! For each row below the pivotal equation . . .
            !! Subtract a multiple of row i from row j, thereby setting to zero the coefficients
            !! below the pivot.
            Multiplier = A(J,I) / A(I,I)
            !! For each coefficient in the equation . . .
            A(J,I+1:N) = A(J,I+1:N) - A(I,I+1:N)*Multiplier
            R(J) = R(J) - R(I)*Multiplier         !! . . . and the corresponding right-hand side . . .
         end  do
      end  do

      !! The second section performs backward substitution.

      X(N) = R(N) / A(N,N)                        !! The last solution comes out first.
      do I = N-1, 1, -1

         Total = sum ( A(I,N:I+1:-1)*X(N:I+1:-1) )!! Substitute the solutions X(N) to X(I+1)
                                                                                   !! into Equation I, . . .

         X(I) = (R(I) - Total) / A(I,I)           !! . . . thus giving the unknown X(I).
      end  do

end  subroutine  SIMEQ

!! Figure 8.1a  Gauss elimination with backward substitution for solving simultaneous equations, using array sections

end  module  GROUP