module  GROUP
   implicit  none

   public :: SIMEQ
   integer, parameter, public :: Double=selected_real_kind (14)

contains

   !! This subroutine solves N linear algebraic simultaneous equations using
   !! Gauss elimination and backward substitution.  Partial pivoting (row and column
   !! swapping) is used.
   subroutine  SIMEQ (A, R, X)
      !! INCOMING: A = a set of N coefficients for each of N equations, stored in matrix form.
      !!               The coefficients are changed.
      !!               The first equation is stored in the first row; the second in row 2, & so on.
      !!           R = a vector of N right-hand sides for the above equations.  The
      !!               right-hand side of the first equation is stored in R(1), etc.
      !!               These also are changed.
      !! OUTGOING: X = the solutions of the above equations.

      real (kind=Double), dimension ( : , : ), intent (in out) :: A
      real (kind=Double), dimension ( : ), intent (in out)  :: R
      real (kind=Double), dimension ( : ), intent (out)     :: X

      real (kind=Double), dimension (ubound (X, 1))   :: RHS
      real (kind=Double)                     :: Partial_Sum, Multiplier
      real (kind=Double)                     :: Biggest, Temp
      real (kind=Double), dimension (ubound(A,1)) :: TempCol
      integer, dimension (ubound(A,1))       :: P        !! An array of pointers.
      integer                                :: N
      integer                                :: I, J, K, Row, Col

      !! First check that the arrays are compatible.
      N = ubound(A, 1)
      if (N /= ubound(A, 2) .or. N /= ubound (R, 1) .or. N /= ubound (X, 1) ) then
         print *, "**ERROR, the dimensions of the arrays for the " //   &
            " Gauss equation solver are incompatible."
         print *, "The coefficient matrix is of size ", N,              &
            " by ", ubound(A, 2)
         print *, "The vector of right-hand sides is of size ",         &
            ubound (R, 1)
         print *, "The solution vector X is of size ",                  &
            ubound (X, 1)
         print *, "The solution is abandoned."
         return
      end  if

      !! Initialize:  set pointers containing the positions of columns.
      do I = 1, N
         P(I) = I
      end  do

      !! The first section performs Gauss elimination,
      !! to reduce the equations to upper triangular form.

      do I = 1, N-1                               !! For each pivotal equation . . .
         !! Search for the optimum pivot equation -- search for the row containing the
         !! largest pivot.
         !! During the following loop, Row is the row containing the Biggest pivot.
         Biggest = abs(A(I,I))
         Row = I
         do J = I+1, N                            !! For each row below the pivotal equation . . .
            if (Biggest < abs(A(J,I)) ) then
               if (A(J,J) /= 0) then
                  Biggest = abs(A(J,I))
                  Row = J                         !! Record the row containing the biggest pivot.
               end if
            end  if
         end  do

         if (I /= Row) then                       !! Row Row has the largest pivot element, so . . .
                                                  !! . . . swap row I with row Row.
            do J = I, N
               Temp  = A(I,J)                     !! Swap coefficients in column J (of row I).
               A(I,J)   = A(Row,J)
               A(Row,J) = Temp
            end  do
            Temp  = R(I)                          !! Don't forget the right-hand sides!
            R(I)     = R(Row)
            R(Row)    = Temp
         end  if

         !! Section to search for a column containing an even larger pivot element.
         Col = I
         do J = I+1, N
            if (Biggest < abs (A(I, J)) ) then
               Biggest = abs (A(I,J))
               Col = J
            end  if
         end  do

         if (Col /= I) then                       !! Column Col has a larger element, so . . .
                                                  !! . . . bring it to the pivotal position.
            TempCol = A( : , I)                   !! Exchange column Col with column I.
            A( : ,I) = A( : , Col)
            A( : , Col) = TempCol
            Temp = P(I)                           !! Exchange pointers to columns Col and I.
            P(I) = P(Col)
            P(Col) = Temp
         end  if

         !! Section to perform one step of Gauss Elimination, to eliminate the coefficient below
         !! the pivot element.
         do J = I+1, N                            !! For each row below the pivotal equation . . .
            !! Subtract a multiple of row i from row j, thereby setting to zero the coefficients
            !! below the pivot.
            Multiplier = A(J,I) / A(I,I)
            do K = I+1, N                         !! For each coefficient in the equation . . .
               A(J,K) = A(J,K)- A(I,K)*Multiplier
            end  do
            R(J) = R(J) - R(I)*Multiplier         !! . . . and the corresponding right-hand side.
         end  do
      end  do

      !! The second section performs backward substitution.

      X(N) = R(N) / A(N,N)                        !! The last solution comes out first.
      do I = N-1, 1, -1
         Partial_Sum = 0
         do J = N, I+1, -1                        !! Substitute the solutions X(N) to X(I+1)
            Partial_Sum = Partial_Sum + A(I,J)*X(J)   !! in Equation I, . . .
         end  do
         X(I) = (R(I) - Partial_Sum) / A(I,I)     !! . . . thus giving the unknown X(I).
      end  do

      !! Section to restore the solutions to their correct positions.
      RHS = X
      do J = 1, N
         X(J) = RHS(P(J))
      end  do

   end  subroutine  SIMEQ

!! Figure 8.2  Improved Gauss elimination, with row and column interchanges

end  module  GROUP
