module  GROUP
   implicit  none

   public :: GAUSS_SIEDEL
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine solves a set of linear equations by the Gauss-Siedel iterative process.
subroutine  GAUSS_SIEDEL (A, R, X, NTrips)
      !! INCOMING: A      = a set of N coefficients for N equations;
      !!           R      = a vector of N right-hand sides for the above equations;
      !!           NTrips = maximum number of iterations permitted.
      !! OUTGOING: X      = the solutions of the above equations.

      real (kind=Double), dimension ( : , : ), intent (in out) :: A
      real (kind=Double), dimension ( : ), intent (in out)     :: R
      real (kind=Double), dimension ( : ), intent (out)        :: X
      integer, intent (in)         :: NTrips

      real (kind=Double)           :: NewX, Partial_Sum
      real, parameter              :: Tolerance = 1.0e-12
      integer                      :: J, K, N, NT
      logical                      :: Too_Big

      !! First check that the arrays are compatible.
      N = ubound(A, 1)                            !! The number of equations.
      if (N /= ubound(A, 2) .or. N /= ubound (R, 1) .or. N /= ubound (X, 1) ) then
         print *, "**ERROR, the dimensions of the arrays for the " //   &
            " Gauss-Siedel equation solver are incompatible."
         print *, "The coefficient matrix is of size ", N,              &
            " by ", ubound(A, 2)
         print *, "The vector of right-hand sides is of size ",         &
            ubound (R, 1)
         print *, "The solution vector X is of size ",               &
            ubound (X, 1)
         print *, "The solution is abandoned."
         return
      end  if

      !! PREPARATION SECTION.
      do K = 1, N                                 !! Set up the initial approximations for the
         X(K) = R(K) / A(K,K)                     !! solutions X, obtained by assuming that all
      end  do                                     !! the X-values are zero.

      !! ITERATION SECTION.
      do NT = 1, NTrips
         Too_Big = .false.                        !! Set, hoping that this coming iteration will suffice . . .

         do K = 1, N                              !! This loop derives a solution for the K-th
                                                  !! unknown X(K).
            Partial_Sum = 0

            do J = 1, N                           !! This loop sums the products of coefficients
                                                  !! & unknowns.
               if (J /= K) then
                  Partial_Sum = Partial_Sum + A(K,J) * X(J)
               end  if
            end  do

            NewX = (R(K) - Partial_Sum) / A(K,K)  !! This is the new X(K).

            if (abs(NewX-X(K)) > abs(NewX)*Tolerance) then
               Too_Big = .true.                   !! Solution X(K) is not accurate.  Set to repeat . . .
            end  if
            X(K) = NewX                           !! Update the solution for X(K), to be used in
                                                  !! iterating X(K+1).
         end  do

         if (.not. Too_Big) then
            exit
         end  if
      end  do

      if (Too_Big) then
         print *, "**ERROR, the Gauss-Siedel algorithm did not converge."
      end  if

end  subroutine  GAUSS_SIEDEL

!! Figure 8.3  Solving equations using Gauss-Siedel iteration

end  module  GROUP
