module  GROUP
   implicit  none

   public :: TRI_DIAG
   integer, parameter, public :: Double=selected_real_kind (14)

contains

   !! This subroutine performs Gauss elimination on a system of tri-diagonal equations, to
   !! reduce the coefficients to upper triangular form.
   !! Backward-substitution is then performed to determine the solutions.
   subroutine  TRI_DIAG (Inf, D, S, R)
      !! INCOMING: Inf = the inferior diagonal coefficients (i.e., below the diagonal);
      !!           D   = the diagonal coefficients;
      !!           S   = the superior diagonal coefficients (i.e., above the diagonal);
      !!           R   = the right-hand sides.
      !! OUTGOING: R   = the solutions x(1) through x(n).
      !!           D   :  Note that this vector is changed.

      real, dimension ( : ), intent (in)     :: Inf, S
      real, dimension ( : ), intent (in out) :: D, R

      real (kind=Double)                     :: Multiplier
      integer                                :: I, N

      N = ubound(Inf, 1)                          !! The number of equations.
      if ( (N /= ubound(D, 1)) .or. (N /= ubound(S,1)) .or. (N /= ubound(R,1)) ) then
         print *, "**ERROR, for Gauss elimination of a tri-diagonal system of" //   &
            " simultaneous equations, the array bounds are not all of size ", N
         print *, "The inferior diagonal is of size ", ubound(Inf,1)
         print *, "The diagonal is of size ", ubound(D,1)
         print *, "The superior diagonal is of size ", ubound(S,1)
         print *, "The right-hand side is of size ", ubound(Inf,1)
         return
      end  if

      !! The first section performs Gauss elimination.
      do I = 2, N
         Multiplier = real(Inf(I), kind=Double) / D(I-1)
         D(I) = D(I) - Multiplier*S(I-1)
         R(I) = R(I) - Multiplier*R(I-1)
      end  do

      !! The second section performs backward substitution.
      R(N) = R(N) / D(N)
      do I = N-1, 1, -1
         R(I) = (R(I) - S(I) * R(I+1)) / D(I)
      end  do

   end  subroutine  TRI_DIAG

!! Figure 8.4  Solving a tri-diagonal system of equations

end  module  GROUP
