module GROUP
   implicit none
   public                          :: TRI_DIAG, PRINT_EQUATIONS
   integer, parameter, public      :: Double = selected_real_kind (14)

contains

!!  subroutine TRI_DIAG from Figure 8.4 

   !! This subroutine performs Gauss elimination on a system of tri-diagonal equations, to
   !! reduce the coefficients to upper triangular form.
   !! Backward-substitution is then performed to determine the solutions.
   subroutine  TRI_DIAG (Inf, D, S, R)
      !! INCOMING: Inf = the inferior diagonal coefficients (i.e., below the diagonal);
      !!           D   = the diagonal coefficients;
      !!           S   = the superior diagonal coefficients (i.e., above the diagonal);
      !!           R   = the right-hand sides.
      !! OUTGOING: R   = the solutions x(1) through x(n).
      !!           D   :  Note that this vector is changed.

      real, dimension ( : ), intent (in)     :: Inf, S
      real, dimension ( : ), intent (in out) :: D, R

      real (kind=Double)                     :: Multiplier
      integer                                :: I, N

      N = ubound(Inf, 1)                          !! The number of equations.
      if ( (N /= ubound(D, 1)) .or. (N /= ubound(S,1)) .or. (N /= ubound(R,1)) ) then
         print *, "**ERROR, for Gauss elimination of a tri-diagonal system of" //   &
            " simultaneous equations, the array bounds are not all of size ", N
         print *, "The inferior diagonal is of size ", ubound(Inf,1)
         print *, "The diagonal is of size ", ubound(D,1)
         print *, "The superior diagonal is of size ", ubound(S,1)
         print *, "The right-hand side is of size ", ubound(Inf,1)
         return
      end  if

      !! The first section performs Gauss elimination.
      do I = 2, N
         Multiplier = real(Inf(I), kind=Double) / D(I-1)
         D(I) = D(I) - Multiplier*S(I-1)
         R(I) = R(I) - Multiplier*R(I-1)
      end  do

      !! The second section performs backward substitution.
      R(N) = R(N) / D(N)
      do I = N-1, 1, -1
         R(I) = (R(I) - S(I) * R(I+1)) / D(I)
      end  do

   end  subroutine  TRI_DIAG

   !! This subroutine prints a tri-diagonal matrix.
   subroutine  PRINT_EQUATIONS (Inf, D, S, R)
      !! INCOMING: Inf = the inferior diagonal coefficients (i.e., below the diagonal);
      !!           D   = the diagonal coefficients;
      !!           S   = the superior diagonal coefficients (i.e., above the diagonal);
      !!           R   = the right-hand sides.

      real, dimension ( : ), intent(in)   :: Inf, D, S, R

      integer                             :: J, N

      N = ubound (D,1)
      do J = 1, N
         if (J > 1) then
            write (unit=*, fmt="(f10.4)", advance="no" ) Inf(J)
         end if
         write (unit=*, fmt="(f10.4)", advance="no" ) D(J)
         if (J < N) then
            write (unit=*, fmt="(f10.4)", advance="no" ) S(J)
         end  if
         write (unit=*, fmt="(t35, f10.4)", advance="no" ) R(J)
         print *
      end  do

   end  subroutine  PRINT_EQUATIONS

end  module  GROUP

program  SOLVER
   use  GROUP
   implicit  none
   integer                            :: N
   real, dimension ( : ), allocatable :: Inf,  & !! For the inferior diagonal.
                                         D,    & !! For the diagonal.
                                         S,    & !! For the superior diagonal
                                         R       !! For the right-hand sides, which are overwritten
                                                 !! with the solutions.
   integer                            :: J

   print *, "This program solves a tri-diagonal system of equations."
   print *, "How many equations are there to solve?"
   read *, N

   allocate (Inf(N), D(N), S(N), R(N) )

   do J = 1, N
      print *, "Please type the coefficients for row", J, ":"
      if (J > 1) then
         read *, Inf(J)                           !! The first row has two coefficients.
      end  if
      read *, D(J)
      if (J < N) then
         read *, S(J)                             !! The last row has two coefficients.
      end  if
      print *, "and the right-hand side:"
      read *, R(J)
   end  do

   print *, "The coefficients and right-hand sides are:"
   call PRINT_EQUATIONS (Inf, D, S, R)
   call TRI_DIAG (Inf, D, S, R)

   print *, "The solutions are:"
   do J = 1, N
      print *, "X(", J, ")=", R(J)
   end  do

end  program  SOLVER

!! Figure 8.5  Main program for solving a tri-diagonal system of equations
