module  GROUP
   implicit  none

   public :: SOLVE_BANDED
   integer, parameter, public :: Double=selected_real_kind (14)

contains

!! This subroutine solves N linear banded algebraic simultaneous equations using GAUSS
!! elimination and backward substitution.  It determines the width of the band in the upper
!! triangle, and avoids the zero elements in the lower triangle.
subroutine  SOLVE_BANDED (A, R, X)
      !! INCOMING: A = a set of N coefficients for N equations;
      !!           R = a vector of N right-hand sides for the above equations;
      !! OUTGOING: X = the solutions of the above equations.

      real (kind=Double), dimension ( : , : ), intent (in out) :: A
      real (kind=Double), dimension ( : ), intent (in out)     :: R
      real (kind=Double), dimension ( : ), intent (out)        :: X

      real (kind=Double)               :: Partial_Sum, Multiplier
      integer                      :: N
      integer                      :: I, J, K, P, L

      !! First check that the arrays are compatible.
      N = ubound(A, 1)
      if ( (N /= ubound(A, 2)) .or. (N /= ubound (R, 1)) .or. (N /= ubound (X, 1)) ) then
            print *, "**ERROR, the dimensions of the arrays for the " //   &
               " banded equation solver are incompatible."
            print *, "The coefficient matrix is of size ", N,              &
               " by ", ubound(A, 2)
            print *, "The vector of right-hand sides is of size ",         &
               ubound (R, 1)
            print *, "The solution vector X is of size ",                  &
               ubound (X, 1)
            print *, "The solution is abandoned."
            return
      end  if

      !! The first section determines the extent of the band in the upper triangle.
OUTER: do L = N, 2, -1                            !! Commence at the top right-hand corner.
         P = L
         do J = 1, N-L+1                          !! Work down diagonally.
            if (A(J,P) /= 0) then
               exit OUTER
            end  if
            P = P + 1
         end  do
      end  do  OUTER
      !! When we get here, column L is the last diagonal of the band.

      !! The second section performs Gauss elimination, to reduce
      !! the equations to upper triangular form.

      do I = 1, N-1                               !! For each pivotal equation . . .
         do J = I+1, N                            !! For each row below the pivotal equation . . .
            if (A(J,I) /= 0) then                 !! Only for nonzero elements below the pivot.
               !! Subtract a multiple of the pivotal equation I from row J.
               Multiplier = A(J,I) / A(I,I)
               do K = I+1, min(N, I+L-1)          !! For each coefficient in row J . . .
                  A(J,K) = A(J,K)- A(I,K)*Multiplier
               end  do
               R(J) = R(J) - R(I)*Multiplier      !! . . . and the right-hand side of row J.
            end  if
         end  do
      end  do

      !! The third section performs backward substitution.

      X(N) = R(N) / A(N,N)                        !! The last solution comes out first.
      do I = N-1, 1, -1
         Partial_Sum = 0
         do J = N, I+1, -1                        !! Substitute the solutions into Equation I . . .
            Partial_Sum = Partial_Sum + A(I,J)*X(J)
         end  do
         X(I) = (R(I) - Partial_Sum) / A(I,I)     !! . . . thus giving the unknown X(I).
      end  do

end  subroutine  SOLVE_BANDED

!! Figure 8.6  Solving a banded system of equations

end  module  GROUP
