module GROUP
   implicit none
   public                          :: SOLVE_BANDED, READ_BANDED, PRINT_BANDED
   integer, parameter, public      :: Double = selected_real_kind (14)

contains

!!  subroutine SOLVE_BANDED from Figure 8.6 

!! This subroutine solves N linear banded algebraic simultaneous equations using GAUSS
!! elimination and backward substitution.  It determines the width of the band in the upper
!! triangle, and avoids the zero elements in the lower triangle.
subroutine  SOLVE_BANDED (A, R, X)
      !! INCOMING: A = a set of N coefficients for N equations;
      !!           R = a vector of N right-hand sides for the above equations;
      !! OUTGOING: X = the solutions of the above equations.

      real (kind=Double), dimension ( : , : ), intent (in out) :: A
      real (kind=Double), dimension ( : ), intent (in out)     :: R
      real (kind=Double), dimension ( : ), intent (out)        :: X

      real (kind=Double)           :: Partial_Sum, Multiplier
      integer                      :: N
      integer                      :: I, J, K, P, L

      !! First check that the arrays are compatible.
      N = ubound(A, 1)
      if ( (N /= ubound(A, 2)) .or. (N /= ubound (R, 1)) .or. (N /= ubound (X, 1)) ) then
            print *, "**ERROR, the dimensions of the arrays for the " //   &
               " banded equation solver are incompatible."
            print *, "The coefficient matrix is of size ", N,              &
               " by ", ubound(A, 2)
            print *, "The vector of right-hand sides is of size ",         &
               ubound (R, 1)
            print *, "The solution vector X is of size ",                  &
               ubound (X, 1)
            print *, "The solution is abandoned."
            return
      end  if

      !! The first section determines the extent of the band in the upper triangle.
OUTER: do L = N, 2, -1                            !! Commence at the top right-hand corner.
         P = L
         do J = 1, N-L+1                          !! Work down diagonally.
            if (A(J,P) /= 0) then
               exit OUTER
            end  if
            P = P + 1
         end  do
      end  do  OUTER
      !! When we get here, column L is the last diagonal of the band.

      !! The second section performs Gauss elimination, to reduce
      !! the equations to upper triangular form.

      do I = 1, N-1                               !! For each pivotal equation . . .
         do J = I+1, N                            !! For each row below the pivotal equation . . .
            if (A(J,I) /= 0) then                 !! Only for nonzero elements below the pivot.
               !! Subtract a multiple of the pivotal equation I from row J.
               Multiplier = A(J,I) / A(I,I)
               do K = I+1, min(N, I+L-1)          !! For each coefficient in row J . . .
                  A(J,K) = A(J,K)- A(I,K)*Multiplier
               end  do
               R(J) = R(J) - R(I)*Multiplier      !! . . . and the right-hand side of row J.
            end  if
         end  do
      end  do

      !! The third section performs backward substitution.

      X(N) = R(N) / A(N,N)                        !! The last solution comes out first.
      do I = N-1, 1, -1
         Partial_Sum = 0
         do J = N, I+1, -1                        !! Substitute the solutions into Equation I . . .
            Partial_Sum = Partial_Sum + A(I,J)*X(J)
         end  do
         X(I) = (R(I) - Partial_Sum) / A(I,I)     !! . . . thus giving the unknown X(I).
      end  do

end  subroutine  SOLVE_BANDED

!!  subroutine READ_BANDED from Figure 8.7 

!! This subroutine reads in the coefficients and right-hand sides of a banded system of
!! linear equations.
!! Only elements in the band are read in.  Zero elements in the upper & lower triangles
!! are not read in.
subroutine  READ_BANDED (A, R, NDA, NDD)
      !! OUTGOING: A   = the coefficients, in banded form;
      !!           R   = the right-hand sides;
      !!           NDA = the number of nonzero diagonals along the top row;
      !!           NDD = the number of nonzero diagonals down the left-hand column.
      !! NOTE: The calling program must read in the number of equations.

      real (kind=Double), dimension ( : , : ), intent (out) :: A
      real (kind=Double), dimension ( : ), intent (out)     :: R

      integer, intent (out)        :: NDA,     &
                                      NDD

      integer                      :: NDS,  &     !! NDS = number of nonzero diagonals along
                                                  !! the top row;
                                      NDU         !! NDU = number of nonzero diagonals down
                                                  !! the left-hand column.
      integer                      :: C, I, J, N

      !! Perform compatibility checks on the incoming arrays.
      N = ubound(A, 1)                            !! The number of equations.
      if (N /= ubound(A, 2) .or. N /= ubound (R, 1) ) then
         print *, "**ERROR, the dimensions of the arrays for the " //   &
            " banded equation solver are incompatible."
         print *, "The coefficient matrix is of size ", N,              &
            " by ", ubound(A, 2)
         print *, "The vector of right-hand sides is of size ",         &
            ubound (R, 1)
         print *, "The input is abandoned."
         stop
      end  if

      A = 0                                       !! Outside the band must be zero.
      print *, "Please type the number of nonzero diagonals along the top row:"
      read *, NDA
      NDS = NDA
      print *, "Please type the number of nonzero diagonals down the first column:"
      read *, NDD
      NDU = 2 - NDD

      do I = 1, N                                 !! For each row.
         C = max(1, NDU)
         NDS = min(NDS, N)
         print *, "Please type the coefficients for row", I
         do J = C, NDS                            !! For each column.
            read *, A(I,J)
         end  do
         print *, ". . . and the right-hand side:"
         read *, R(I)                             !! The corresponding right-hand side.
         NDS = NDS + 1
         NDU = NDU + 1
      end  do

end  subroutine  READ_BANDED

!!  subroutine PRINT_BANDED from Figure 8.8 

   !! This subroutine prints the coefficients and right-hand sides of a banded system of
   !! linear equations.
   !! The diagonals are printed vertically.  Zero elements in the upper triangle and lower
   !! triangle are not printed.
   subroutine  PRINT_BANDED (A, R, NDA, NDD)
      !! INCOMING: A   = the coefficients, in banded form;
      !!           R   = the right-hand sides;
      !!           NDA = the number of nonzero diagonals along the top row;
      !!           NDD = the number of nonzero diagonals down the first column.

      real (kind=Double), dimension ( : , : ), intent (in out) :: A
      real (kind=Double), dimension ( : ), intent (in out)     :: R
      integer, intent (in)         :: NDA,  &
                                      NDD

      integer                      :: NDS,  &     !! NDS = number of nonzero diagonals along
                                                  !! the top row;
                                      NDU         !! NDU = number of nonzero diagonals down
                                                  !! the left-hand column.
      integer                      :: C, I, J, N

      !! Perform compatibility checks on the incoming arrays.
      N = ubound(A, 1)                            !! The number of equations.
      if (N /= ubound(A, 2) .or. N /= ubound (R, 1) ) then
         print *, "**ERROR, the dimensions of the arrays for the " //   &
            " banded equation printer are incompatible."
         print *, "The coefficient matrix is of size ", N,              &
            " by ", ubound(A, 2)
         print *, "The vector of right-hand sides is of size ",         &
            ubound (R, 1)
         print *, "The output is abandoned."
         stop
      end  if

      NDS = NDA
      NDU = 2 - NDD

      do I = 1, N                                 !! For each row.
         C = max(1, NDU)
         NDS = min (NDS, N)
         print *, "The coefficients for row", I
         do J = C, NDS                            !! For each column.
            write (unit=*, fmt="(f10.6)", advance="no" ) A(I,J)
         end  do
         write (unit=*, fmt="(f10.6)" ) R(I)      !! The corresponding right-hand side.
         NDS = NDS + 1
         NDU = NDU + 1
      end  do

   end  subroutine  PRINT_BANDED

end  module  GROUP

program  SOLVE
   use  GROUP
   implicit  none
   real (kind=Double), dimension ( : , : ), allocatable  :: A
   real (kind=Double), dimension ( : ), allocatable      :: R, X
   integer                         :: Status
   integer                         :: NDA, NDD
   integer                         :: J

   integer                         :: N

   print *, "This program solves a set of linear simultaneous algebraic" //   &
      " equations presented in banded form."
   print *, "How many equations do you want to solve?"
   read *, N
   print *, "There are", N, " equations to be solved."

   allocate (A(N, N), R(N), X(N), stat=Status)
   if (Status /= 0) then
      print *, "Insufficnent memory to allocate arrays."
      stop
   end  if

   call READ_BANDED (A, R, NDA, NDD)
   call PRINT_BANDED (A, R, NDA, NDD)
   call SOLVE_BANDED (A, R, X)
   print *
   print *, "The solutions are:"
   do J = 1, N
      write (unit=*, fmt="(a, i4, a, f12.6)" ) "X(", J, ")=", X(J)
   end  do

end  program  SOLVE

!! Figure 8.9  Main program for solving a banded system of equations
