!! This program reads in a PCX file in color and black-and-white, and displays it on
!! the screen using Salford Graphics.  It also prints it on an HP DeskJet 500C color printer.
!! The PCX file can be in uncompressed and compressed formats.

!! include "SALFORD.INC"

module PCX_Header

   !! The next set of declarations constitutes the PCX_Header (which is 128 bytes in size).
   type, public                    :: Header
      character (len=1)            :: Identifier  !! File Identification.
      character (len=1)            :: Version     !! Version Number.
      character (len=1)            :: Encoding_format !! 0 = no encoding (uncompressed);
                                                      !! 1 = run-length encoding (compressed).
      character (len=1)            :: Bits_per_Pixel  !! 4 bits per pixel.
                                                  !! Co-ordinates for the start of the image:
      character (len=2)            :: X_Start     !! X co-ordinate of start.
      character (len=2)            :: Y_Start     !! Y co-ordinate of start.
                                                  !! Co-ordinates for the end of the image:
      character (len=2)            :: X_end       !! X co-ordinate of end.
                                                  !! 0:1023 = length of image in pixels.
      character (len=2)            :: Y_end       !! Y co-ordinate of end.
                                                  !! 0:1023 = height of image in pixels.
      character (len=2)            :: Horizontal_Resolution   !! in dots per inch.
      character (len=2)            :: Vertical_Resolution  !! in dots per inch.
      character (len=1), dimension(0:47)  :: Palette  !! The color palette.
      character (len=1)            :: Reserved
      character (len=1)            :: Number_of_Bit_Planes
      character (len=2)            :: Bytes_per_Line  !! # of bytes per line unencoded.
      character (len=2)            :: Palette_type
      character (len=2)            :: Horizontal_Screen_Size
      character (len=2)            :: Vertical_Screen_Size
      character (len=1), dimension(54) :: Reserved2
   end  type  Header                              !! End of declarations for the PCX header.

end  module PCX_Header

module GROUP
   implicit none
   public :: READ_PCX_FILE, READ_BYTE_FROM_GRAPHICS_FILE, READ_PCX_FILE_HEADER,  &
             CHR2int, UNCOMPRESS, UNCOMPRESS_PLANES, PRINT_HP500C, WRITE_BYTES

contains

   !! This procedure reads the header for a picture file in PCX format.
   !! The record size of the input file is 1 byte.

   subroutine  READ_PCX_FILE ( )

      !! A picture file commences with a 128-byte header, which gives video screen
      !! parameters, colors, and size of image.
      !! For details of PCX file formats, see J.D. Murray & W. vanRyper, "Encyclopedia of Graphics
      !! File Formats", O'Reilly, Sebastopol, CA, 1996.

!!    use  SALFORD
      use  PCX_Header

      type (Header)                :: H

      integer                      :: K

      character (len=1), dimension ( : , : ), allocatable   :: Pixel_Image
      !! One row of Pixel_Image holds one line of the image.
      !! Each byte of Pixel_Image holds one pixel of the image, stored as
      !! a 4-bit value representing the color.
      character (len=80)           :: File_Name
      integer                      :: J, Row, Col
      character (len=1)            :: Byte_to_Read
      logical                      :: Status
      character (len=1)            :: Marker

      integer                      :: Lower, Upper

      character (len=128)          :: Buffer
      integer                      :: Bits_per_Pixel,      &
                                                  !! Co-ordinates for the start of the image:
                     X_Start,          &          !! X co-ordinate of start.
                     Y_Start        ,  &          !! Y co-ordinate of start.
                                                  !! Co-ordinates for the end of the image:
                     X_end,            &          !! X co-ordinate of end.
                                                  !! 0:1023 = length of image in pixels.
                     Y_end,            &          !! Y co-ordinate of end.
                                                  !! 0:1023 = height of image in pixels.
                     Horizontal_Resolution,  &    !! in dots per inch.
                     Vertical_Resolution,    &    !! in dots per inch.
                     Number_of_Bit_Planes,   &
                     Bytes_per_Line,         &    !! # of bytes per line unencoded.
                     Palette_type,           &
                     Horizontal_Screen_Size, &
                     Vertical_Screen_Size

      Marker = char(128+64)

      print *, "Please type the file name of the PCX picture you want to read:"
      read *, File_Name
      File_Name = adjustl (File_Name)
      open (unit=9, file=trim(File_Name), access="direct", recl=1, status="old", action="read")

      call READ_PCX_FILE_HEADER(Buffer)           !! Read in the parameters of the picture file.

      !! Section to extract parameters from the PCX file header
      H%Identifier      = Buffer( 1: 1)           !! File Identification.
      H%Version         = Buffer( 2: 2)           !! Version Number.
      H%Encoding_format = Buffer( 3: 3)           !! 0 = no encoding (uncompressed);
      H%Bits_per_Pixel  = Buffer( 4: 4)
      H%X_Start         = Buffer( 5: 6)           !! X co-ordinate of start.
      H%Y_Start         = Buffer( 7: 8)           !! Y co-ordinate of start.
      H%X_end           = Buffer( 9:10)           !! X co-ordinate of end.
      H%Y_end           = Buffer(11:12)           !! Y co-ordinate of end.
      H%Horizontal_Resolution = Buffer(13:14)     !! in dots per inch.
      H%Vertical_Resolution   = Buffer(15:16)     !! in dots per inch.
      H%Number_of_Bit_Planes = Buffer(66:66)
      H%Bytes_per_Line  = Buffer(67:68)           !! # of bytes per line unencoded.
      H%Palette_type    = Buffer(69:70)
      H%Horizontal_Screen_Size = Buffer(71:72)
      H%Vertical_Screen_Size  = Buffer(73:74)

      !! Section to convert from CHARACTER to INTEGER:
      Bits_per_Pixel = ichar(H%Bits_per_Pixel)
      X_Start = CHR2int(H%X_Start)
      Y_Start = CHR2int(H%Y_Start)
      X_end   = CHR2int(H%X_end)
      Y_end   = CHR2int(H%Y_end)
      Horizontal_Resolution = CHR2int(H%Horizontal_Resolution)
      Vertical_Resolution   = CHR2int(H%Vertical_Resolution)
      Number_of_Bit_Planes = ichar(H%Number_Of_Bit_Planes)
      Bytes_per_Line = CHR2int(H%Bytes_per_Line)
      Palette_type = CHR2int(H%Palette_type)
      Horizontal_Screen_Size = CHR2int(H%Horizontal_Screen_Size)
      Vertical_Screen_Size   = CHR2int(H%Vertical_Screen_Size)

      print *, "PCX file parameters:"
      print *, "Identifier=", ichar(H%Identifier)
      print *, "Version=", ichar(H%Version)
      print *, "Encoding format=", ichar(H%Encoding_Format)
      print *, "Bits per pixel=", Bits_per_Pixel
      print *, "X-start=", X_Start, ", Y-start=", Y_Start
      print *, "X-end=", X_end, ", Y-end=", Y_end
      print *, "Horizontal Resolution=", Horizontal_Resolution
      print *, "Vertical Resolution=", Vertical_Resolution
      print *, "Number of bit planes=", Number_of_Bit_Planes
      print *, "Bytes per line=", Bytes_per_Line
      print *, "Palette type=", Palette_type
      print *, "Horizontal screen size=", Horizontal_Screen_Size
      print *, "Vertical screen size=", Vertical_Screen_Size

      !! Validate the file as a PCX file.
      if (H%Identifier /= char(10)) then
         print *, "This is not a PCX file.  Display abandoned."
         stop
      end  if

      print *, "X_Limit=",             &
         Bytes_per_Line*Number_of_Bit_Planes*8/Bits_per_Pixel-1

      print *, "Enter an integer when you are ready to read in the picture:"
      read *, Row                                 !! a pause only.

      !! This section reads in the image part of the file.

      allocate (Pixel_Image (0:Y_end, 0:X_end) )

      Pixel_Image = char(0)                       !! Clear storage for building the image.

      if (H%Encoding_format == char(0)) then      !! the file is not compressed.

         select  case(Bits_per_Pixel)
            case (1)
               print *, "Displaying uncompressed black & white graphics file"
               do Row = 0, Y_end
                  do Col = 0, X_end, 8
                     call READ_BYTE_FROM_GRAPHICS_FILE(Byte_to_Read, Status)
                     if (Status) then
                        print *, &
                        "Unexpected end of file encountered when reading pixel row", &
                           Row
                        stop
                     end  if
                                                  !! Each byte contains 8 pixel elements.  Unpack them.
                     J = ichar (Byte_to_Read)
                     do K = 7, 0, -1
                        Pixel_Image(Row, Col+K) = char( iand(J, 1) )
                        J = ishft(J, -1)          !! A right shift of 1 place.
                     end  do
                  end  do
               end  do

            case (4)
               print *, "Displaying uncompressed color graphics file"
               do Row = 0, Y_end
                  do Col = 0, X_end, 2
                     call READ_BYTE_FROM_GRAPHICS_FILE (Byte_to_Read, Status)
                     if (Status) then
                        print *, &
                        "Unexpected end of file encountered when reading pixel row", &
                           Row
                        stop
                     end  if
                                                  !! Each byte read in contains 2 pixel elements.
                     Upper = ishft( ichar(Byte_to_Read), -4)
                     Lower = iand ( ichar(Byte_to_Read), 15)
                     Pixel_Image(Row, Col) = char(Upper)
                     Pixel_Image(Row, Col+1) = char(Lower)
                  end  do
               end  do

            case default
               print *, "Unable to display graphic image."
               print *, "Unsupported number of bits per pixel,", Bits_per_Pixel
               stop
         end  select

      else                                        !! The file is compressed.  Uncompress it.
         if (Bits_per_Pixel == 0) then
            print *, "Unable to display graphic image."
            print *, "Unsupported number of bits per pixel,", Bits_per_Pixel
            print *, "Attempting to continue with 1 bit per pixel."
            Bits_per_Pixel = 1
         end  if

         if (Number_of_Bit_Planes == 1) then
            call UNCOMPRESS (X_end, Y_end, Pixel_Image, Number_of_Bit_Planes, &
               Bytes_per_Line, Bits_per_Pixel)

         else
            call UNCOMPRESS_PLANES (X_end, Y_end, Pixel_Image, Number_of_Bit_Planes,   &
               Bytes_per_Line, Bits_per_Pixel)
         end  if
      end  if

      !! Check whether there is a color palette attached to the end of the file.

      print *, "Ready to display image:"
!!    call GET_KEY@ (Row)

      !! Display the image.
!!    call VGA@
      !! PCX files treat the top left-hand corner of the screen as the origin (0, 0).
      do Row = 0, min(Y_end, 440)                 !! Work along the Y direction.
         do Col = 0, min(X_end, 600)              !! Work along the X direction.
            K = ichar(Pixel_Image(Row, Col) )
!!          call SET_PIXEL@ (Col, Row, K)
         end  do
      end  do

!!    call GET_KEY@ (Row)
!!    call TEXT_MODE@
      call PRINT_HP500C (Pixel_Image)             !! To print the image.

   end  subroutine  READ_PCX_FILE

   !! These procedures read a compressed picture file in PCX format, &
   !! display it on the screen.
   !! The record size of the input file is 1 byte.

   subroutine  READ_BYTE_FROM_GRAPHICS_FILE (Byte_to_Read, Status)

      character (len=*), intent (out) :: Byte_to_Read
      logical, intent (out)           :: Status

      integer, save                   :: Rec_No = 0
      integer                         :: File_Status

      Rec_No = Rec_No + 1
      read (unit=9, rec=Rec_No, iostat=File_Status) Byte_to_Read

      if (File_Status < 0) then                   !! Seen end of file.
         Status = .true.
         Byte_to_Read = char(0)
      else
         Status = .false.
      end  if
   end  subroutine  READ_BYTE_FROM_GRAPHICS_FILE

   !! Reads in the 128-byte header of a PCX file.
   subroutine READ_PCX_FILE_HEADER (Buffer)

      character (len=*), intent (out) :: Buffer

      character (len=1)               :: Byte_to_Read
      logical                         :: Status
      integer                         :: K

      do K = 1, 128
         call READ_BYTE_FROM_GRAPHICS_FILE (Byte_to_Read, Status)
         if (Status) then
            print *, "Encountered premature end of file while reading the 128-byte header."
            print *, "The read was abandoned."
            stop
         end  if
         Buffer(K:K) = Byte_to_Read
      end  do

   end  subroutine  READ_PCX_FILE_HEADER

   !! This function converts a 2-byte CHARACTER data to an integer.
   !! As the right-hand byte is the more significant, it is multiplied by 256
   !! to obtain the required weight.
   function CHR2int (String) result (Whole_Number)
      character (len=*), intent (in) :: String
      integer                        :: Whole_Number

      Whole_Number = ishft(ichar(String(2:2)), 8) + ichar(String(1:1))
   end  function CHR2int

   !! This subroutine uncompresses a PCX picture file presented as a single plane.
   subroutine  UNCOMPRESS (X_end, Y_end,  Pixel_Image, Number_of_Bit_Planes,  &
                     Bytes_per_Line, Bits_per_Pixel)
      !! INCOMING: X_end                = The X co-ordinate of the RH end of the picture;
      !!           Y_end                = The Y co-ordinate of the bottom of the picture;
      !!           Number_of_Bit_Planes = The number of bit planes;
      !!           Bytes_per_line       = The number of bytes for each scan line;
      !!           Bits_per_Pixel       = The number of bytes that each pixel occupies.
      !! OUTGOING: Pixel_Image          = Receives the image of the PCX file that was read in.

      integer, intent (in)         :: X_end       !! The address of the RHS column of the image.
      integer, intent (in)         :: Y_end       !! The address of the bottom row of the image.
      character (len=*), dimension (0: , 0: ), intent (out) :: Pixel_Image
      integer, intent (in)         :: Number_of_Bit_Planes, Bytes_per_Line,  &
                                      Bits_per_Pixel
      character (len=1)            :: Byte_to_Read
      integer                      :: I, K, L, CP, Run_Length
      integer                      :: B8
      integer                      :: Bit8
      logical, save                :: Status = .false.

      integer                      :: X_Limit
      integer, parameter           :: Marker32 = 128+64

      print *, X_end, Y_end
      X_Limit = Bytes_per_Line*Number_of_Bit_Planes*8/Bits_per_Pixel - 1
      print *, X_Limit

READING_FILE_LOOP:   &
      do I = 0, Y_end                             !! For each of the scan lines . . .
         print *, "Reading single-plane compressed Scan line", I
         CP = 0
         do                                       !! Forever.
!!          IF (Status)  EXIT READING_FILE_LOOP
            call READ_BYTE_FROM_GRAPHICS_FILE (Byte_to_Read, Status)
            B8 = ichar(Byte_to_Read)
            if (iand(B8, Marker32) == Marker32) then !! We have a run of similar bytes.
               !! Expand the next two bytes.
               Run_Length = iand(B8, not(Marker32))
!!             IF (Status) EXIT READING_FILE_LOOP
               call READ_BYTE_FROM_GRAPHICS_FILE (Byte_to_Read, Status)
                                                  !! The next byte is a data byte.
               B8 = ichar (Byte_to_Read)          !! Save & unpack the pair of pixels at the
                                                  !! start of a potential run.

UNPACK_RUN:    &
               do K = 1, Run_Length
                  do L = 1, 8/Bits_per_Pixel      !! Unpack one byte into 8/Bits_per_Pixel.
                     if (CP <= min (X_Limit, X_end-1)) then
                        Bit8 = iand(Bit8, 15)
                        select case (Bits_per_Pixel)
                           case (1, 2, 3, 4)
                              Pixel_Image(I, CP) = char(Bit8)
                           case (5, 6, 7, 8)
                              Pixel_Image(I, CP) = char(Bit8)
                           case default
                        end  select
                     end  if
                     CP = CP + 1
                     if (CP > X_Limit .and. K < Run_Length) then
                        print *, "***Warning: scan line ", I,        &
                              " exceeds the declared length by ", &
                              Run_Length-K+1, " bytes, while uncompressing."
                        print *, "The PCX file is possibly damaged."
                        print *, "These bytes have been ignored."
                        exit  UNPACK_RUN
                     end  if
                  end  do
               end  do  UNPACK_RUN
            else                                  !! It's a single (uncompressed) byte.
               do L = 1, 8/Bits_per_Pixel         !! Unpack one byte into 8/Bits_per_Pixel.
                  if (CP <= min (X_Limit, X_end)) then
                     Bit8 = ishft(B8, -8+L*Bits_per_Pixel)  !! A right shift.
                     Bit8 = iand(Bit8, 15)
                     select case (Bits_per_Pixel)
                        case (1, 2, 3, 4)
                           Pixel_Image(I, CP) = char(Bit8)
                        case (5, 6, 7, 8)
                           Pixel_Image(I, CP) = char(Bit8)
                        case default
                     end  select
                  end  if
                  CP = CP + 1
               end  do
            end  if
            if (CP > X_Limit) then
               exit
            end  if
         end  do                                  !! of DO FOREVER.
      end  do  READING_FILE_LOOP                  !! of do I = . . .

      if (Status) then
         print *, "Unexpected end of file while uncompressing."
      end  if
   end  subroutine  UNCOMPRESS

   !! This subroutine uncompresses a picture file presented as bit planes, that is,
   !! an image consisting of single bits.  It assumes that the planes are
   !! consecutive for each line.
   !! For EGA, a screen line of <n> pixels will be presented as <n> bits
   !! representing red, immediately followed by <n> bits representing green,
   !! immediately followed by <n> bits representing blue, and immediately
   !! followed by <n> intensity bits.

   subroutine  UNCOMPRESS_PLANES (X_end, Y_end, Pixel_Image, Number_of_Bit_Planes,  &
                         Bytes_per_Line, Bits_per_Pixel)
      !! INCOMING: X_end                = The X co-ordinate of the RH end of the picture;
      !!           Y_end                = The Y co-ordinate of the bottom of the picture;
      !!           Number_of_Bit_Planes = The number of bit planes;
      !!           Bytes_per_line       = The number of bytes for each scan line;
      !!           Bits_per_Pixel       = The number of bytes that each pixel occupies.
      !! OUTGOING: Pixel_Image          = Receives the image of the PCX file that was read in.

      integer, intent (in)         :: X_end       !! The address of the RHS column of the image.
      integer, intent (in)         :: Y_end       !! The address of the bottom row of the image.
      character (len=*), dimension (0: , 0: ), intent (out) :: Pixel_Image
      integer, intent (in)         :: Number_of_Bit_Planes, Bytes_per_Line,  &
                                      Bits_per_Pixel

      character (len=1)            :: Byte_to_Read
      integer                      :: I, K, L, No_Bits_per_Line, CP,  Run_Length
      integer                      :: B8
      logical, save                :: Status = .false.

      integer                      :: X_Limit, Plane
      integer, parameter           :: Marker32 = 128+64
      integer                      :: Ktemp, Ltemp

      Status = .false.

      X_Limit = Bytes_per_Line*Number_of_Bit_Planes*8/Bits_per_Pixel - 1
      No_Bits_per_Line = 8*Bytes_per_Line         !! The number of bits in a scan line.
      print *, "X-limit=", X_Limit, ", No. of bits per line=", No_Bits_per_Line

      print *, "Please type an integer to continue:"
      read *, I

READING_FILE_LOOP2:   &
       do I = 0, Y_end                            !! For each of the scan lines . . .
         print *, "Reading compressed multi-plane scan line", I
         CP = 0
         do                                       !! FOREVER -- For each line (for all the bit planes).
            if (Status) then
               exit READING_FILE_LOOP2
            end  if
            call READ_BYTE_FROM_GRAPHICS_FILE (Byte_to_Read, Status)
            B8 = ichar(Byte_to_Read)
            if (iand(B8, Marker32) == Marker32) then !! We have a run of similar bytes.
               !! Expand the next two bytes.
               Run_Length = iand(B8, not(Marker32))
               if (Status) then
                  exit READING_FILE_LOOP2
               end  if
               call READ_BYTE_FROM_GRAPHICS_FILE (Byte_to_Read, Status)
                                                  !! The next byte is a data byte.
               B8 = ichar (Byte_to_Read)          !! Save & unpack the pair of pixels at the
                                                  !! start of a potential run.
UNPACK_RUN2:    &
               do K = 1, Run_Length
                  do L = 1, 8                     !! For each bit in the byte . . .
                     Plane = CP/(No_Bits_per_Line+1)
                     if (CP <= X_Limit) then      !! OR a bit of the plane into the image.
                        Ktemp = iand(ishft(B8, -(8-L)), 1) !! A right shift.  Result is 0 or 1.
                        Ltemp = ichar( Pixel_Image(I,modulo(CP, No_Bits_per_Line)) )
                        Ktemp = ishft(Ktemp, Plane)    !! A left shift of "Plane" positions.
                        if (Ktemp /= 0) then      !! conditionally, to save time.
                           Pixel_Image(I,modulo(CP, No_Bits_per_Line)) =   &
                              char(ior(Ltemp, Ktemp) )
                        end  if
                                                  !! Store the pixel.
                     end  if
                     CP = CP + 1
                  end do
                  if (CP > X_Limit .and. K < Run_Length) then
                     print *, "***Warning: scan line ", I,        &
                           " exceeds the declared length by ", &
                           Run_Length-K+1, " bytes, while uncompressing."
                     print *, "The PCX file is possibly damaged."
                     print *, "These bytes have been ignored."
                     exit UNPACK_RUN2
                  end  if
               end  do  UNPACK_RUN2
            else                                  !! It's a single (uncompressed) byte.
               do L = 1, 8                        !! For each bit in the byte . . .
                  Plane = CP/(No_Bits_per_Line+1)
                  if (CP <= X_Limit) then
                     Ktemp = iand(ishft(B8, -(8-L)), 1) !! A right shift.  Result is 0 or 1.
                     Ltemp = ichar( Pixel_Image(I,modulo(CP, No_Bits_per_Line)) )
                     Ktemp = ishft(Ktemp, Plane)  !! A left shift of "Plane" positions.
                     if (Ktemp /= 0) then         !! conditionally, to save time.
                        Pixel_Image(I,modulo(CP, No_Bits_per_Line)) =   &
                           char(ior(Ltemp, Ktemp) )
                     end  if
                     Pixel_Image(I,modulo(CP, No_Bits_per_Line)) = &
                        char(ior(Ltemp, Ktemp) )  !! Store the pixel.
                  end  if
                  CP = CP + 1
               end  do
            end  if

            if (CP > X_Limit) then
               exit
            end  if
         end  do                                  !! of DO FOREVER.

      end  do  READING_FILE_LOOP2                 !! of do I = . . .

      if (Number_of_Bit_Planes == 4) then
         Pixel_Image = char(iand(ishft(ichar(Pixel_Image), -1), 7))
                                                  !! To remove the intensity bit, which is the least sign
                                                  !! bit.  Shift the remaining bits to the right, using
                                                  !! ISHFT.  IAND is used to remove any high-order bits.
      end  if

      if (Status) then
         print *, "Unexpected end of file while uncompressing."
      end  if
   end  subroutine  UNCOMPRESS_PLANES

   !! This subroutine prints a color graphics image on an HP DeskJet 500C color printer.
   subroutine  PRINT_HP500C (Pixel_Image)
      !! INCOMING: Pixel_Image = the picture file to be printed.
      character (len=*), dimension (0: , 0: ), intent (in)       :: Pixel_Image

      character (len=(size(Pixel_Image,2)+7)/8), dimension (0:2) :: Three_Plane_Line
                                                  !! /8 to convert # bits to # bytes.

      character (len=1), parameter :: ESC = char(27)
      character (len=2), parameter :: Reset_Printer  = ESC // "E"      !! Reset printer.
      character (len=5), parameter :: Start_Graphics = ESC // "*r0A"   !! Start graphics mode.
      character (len=7), parameter :: Resolution     = ESC // "*t150R"
                                                  !! Resolution of 150 dots per inch.
                                                  !! (300 dpi gives high resolution, 150 dpi gives a
                                                  !! less-accurate picture, but twice the size).
      character (len=8), save      :: Line_Width     = ESC // "*r0000S"   !! Line Width in pixels.
      character (len=6), parameter :: Three_Planes   = ESC // "*r-3U" !! 3 planes, CYM palette.
      character (len=8), save      :: One_Plane_Data = ESC // "*b0000V"
                                                  !! Transmit data by plane (for planes 1 and 2 only).
      character (len=8), save      :: Third_Plane_Data = ESC // "*b0000W"
                                                  !! Transmit third data plane.
      character (len=5), parameter :: Compression = ESC // "*b0M"
      character (len=5), parameter :: Advance  = ESC // "*b1Y"      !! Specify advance of 1 pixel
                                                                    !! in the Y direction.
      character (len=5), parameter :: Stop_Graphics = ESC // "*rbC" !! End of graphics.
      integer                      :: Line_Size   !! in number of pixels in a line of the image.
      integer                      :: Bit_Counter, Planes, CP, B8, J, K
      integer                      :: Bytes_per_Line

      !! Set up the Line_Width parameter for the HP DeskJet 500C.
      Line_Size = ubound(Pixel_Image, 2) + 1      !! Number of pixels in a line (4 bits per pixel).
      write (unit=Line_Width(4:7), fmt="(i4.4)" ) Line_Size

      Bytes_per_Line = (Line_Size+7)/8

      !! Set up the Transmit data code.
      write (unit=One_Plane_Data(4:7), fmt="(i4.4)" ) Bytes_per_Line
      write (unit=Third_Plane_Data(4:7), fmt="(i4.4)" ) Bytes_per_Line

      print *, "Line size=", Line_Size, " pixels,", (Line_Size*3)/8, " data bytes."
      print *, "Writing file for HP DeskJet 500C"

      open (unit=10, file="HP500C.OUT", access="direct", action="write",   &
           status="new", recl= 1)

      call WRITE_BYTES (Reset_Printer)            !! Reset Printer.
      call WRITE_BYTES (Line_Width)               !! Must come before Start Graphics command.
      call WRITE_BYTES (Resolution)               !! Ditto.
      call WRITE_BYTES (Three_Planes)             !! Specify 3-plane color raster data.
      call WRITE_BYTES (Start_Graphics)
      call WRITE_BYTES (Compression)

      !! This loop converts a pixel image (4 bits per byte) to one
      !! of planes.  Each line of the image is converted to 3 planes.
      do J = 0, ubound (Pixel_Image, 1)           !! For each row of the pixel image . . .
         do Planes = 0, 2
            Bit_Counter = 0                       !! Tells us when a byte is full.
            CP = 0                                !! A column counter for each plane image.
            do K = 0, ubound(Pixel_Image, 2)
               B8 = ishft(B8, 1) + iand (ishft (ichar (Pixel_Image(J,K)), -Planes), 1)
               Bit_Counter = Bit_Counter + 1
               if (Bit_Counter >= 8) then
                  CP = CP + 1
                  Three_Plane_Line(Planes)(CP:CP) = char (B8)
                  Bit_Counter = 0
               end  if
            end  do
         end  do

         if (Bit_Counter > 0) then                !! Write out an incomplete byte.
            Three_Plane_Line(CP+1:CP+1) = char ( ishft(B8, 8-Bit_Counter) )
         end  if

         call WRITE_BYTES (One_Plane_Data)
         call WRITE_BYTES (Three_Plane_Line(0))

         call WRITE_BYTES (One_Plane_Data)
         call WRITE_BYTES (Three_Plane_Line(1))
         call WRITE_BYTES (Third_Plane_Data)
         call WRITE_BYTES (Three_Plane_Line(2))

      end  do

      call WRITE_BYTES (Stop_Graphics)
      call WRITE_BYTES (Reset_Printer)            !! Reset Printer.
      close (unit=10)
      print *, "Graphics file for HP DeskJet 500C completed."

   end  subroutine  PRINT_HP500C

   !! This subroutine writes out a byte at a time.
   subroutine WRITE_BYTES (Data)
      !! INCOMING: Data= a string.
      character (len=*), intent (in) :: Data

      integer, save                :: Rec_No = 0
      integer                      :: J

      !! Write the string into the file:
      do J = 1, len (Data)
         Rec_No = Rec_No + 1
         write (unit=10, rec=Rec_No) Data (J:J)
      end  do

   end  subroutine  WRITE_BYTES

end  module  GROUP

program  READ_PCX
   use  GROUP
   implicit  none

   call READ_PCX_FILE ( )

end  program  READ_PCX

!! Figure 9.10  Reading a PCX file
