module  GLOBAL
   !! Graph holds the artwork, which is 1024 pixels wide and 1024 pixels deep.
   !! Each element of Graph holds one pixel row of the finished artwork.
   !! A one-bit inserted in Graph will be displayed as a black dot.
   character (len=128), dimension (1024), public   :: Graph
   !! Each bit is initialized to zero, representing an untarnished (white) image.

   real, public                    :: x0, y0,              &
                                      x1, y1
   real, public                    :: theta

end  module  GLOBAL

module GROUP
   implicit none
   public                          :: DRAW, DRAW_LINE, PRINT_GRAPHICS

contains

!!  INCLUDE  "DRAWSAL.INC"        !! For Salford graphics.
!!  include subroutine DRAWLINE from Figure 9.4 
!!  include subroutine PRINT_GRAPHICS from Figure 9.4 

   !! This subroutine draws a straight line from the starting point (Xs,Ys) to the
   !! finishing point (Xf,Yf).  It also performs scaling of the artwork according to
   !! the value of <Scaling>, and gives it an <Offset>, for centering on the page.
   subroutine  DRAW_LINE (Xs, Ys, Xf, Yf)
      !! INCOMING: (Xs, Ys) are the co-ordinates of the starting point;
      !!       (Xf, Yf) are the co-ordinates of the finish point.
      use  GLOBAL

      real, intent (in)            :: Xs, Ys,        &
                                      Xf, Yf

      integer, parameter           :: Scaling = 100
      integer, parameter           :: Offset = 512
      real                         :: Row, Col
      integer                      :: X_Increment,   &
                                      Y_Increment
      integer                      :: Byte_Address, Bit_Address, Byte

      if (Ys > Yf) then                           !! "Y_Increment" is 1 Pixel in the Y direction.
         Y_Increment = -1
      else
         Y_Increment = 1
      end  if
      if (Xs > Xf) then                           !! "X_Increment" is 1 Pixel in the X direction.
         X_Increment = -1
      else
         X_Increment = 1
      end  if

      if (int(Xf*Scaling) == int(Xs*Scaling) ) then !! Only the Y value changes.
         Byte_Address = ishft (int(Xs*Scaling)+Offset, -3)
         Bit_Address = iand ( not ( int(Xs*Scaling)+Offset), 7)

         Row = Ys*Scaling
         do
            if (Row > Yf*Scaling) then
               exit
            end  if
            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph(int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Row = Row + Y_Increment
         end  do
      else if (abs(Xs-Xf) >= abs(Ys-Yf) ) then    !! The slope of the line is <= 45 degrees.
                                                  !! The X and Y values change, or just X changes.

         Col = Xs*Scaling
         do
            if (Col > Xf*Scaling) then
               exit
            end  if
            Row = (Yf-Ys)*(Col-Xs*Scaling)/(Xf - Xs) + Ys*Scaling

            Byte_Address = ishft (int(Col)+Offset, -3)
            Bit_Address = iand ( not ( int(Col)+Offset), 7)

            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph( int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Col = Col + X_Increment
         end  do
      else                                        !! The slope of the line is > 45 degrees.
         Row = Ys*Scaling
         do
            if (Row > Yf*Scaling) then
               exit
            end  if
            Col = (Xf-Xs)*(Row-Ys*Scaling)/(Yf - Ys) + Xs*Scaling

            Byte_Address = ishft (int(Col)+Offset, -3)
            Bit_Address = iand ( not ( int(Col)+Offset), 7)
            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph( int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Row = Row + Y_Increment
         end  do
      end  if

   end  subroutine  DRAW_LINE

   !! This subroutine prints the finished artwork on an HP DeskJet 500 series inkjet printer.
   subroutine  PRINT_GRAPHICS ( )
      use  GLOBAL

      integer                      :: Row
      character (len=1), parameter :: ESC = char(27) !!{XE "LEN: in CHARACTER declaration"}
      character (len=7), parameter :: Resolution = ESC // "*t300R"  !! Control codes for the HP 500
      character (len=5), parameter :: Line_Spacing = ESC // "*b1Y"
      character (len=3), parameter :: Start = ESC // "*b" !! Defines a transmission of bytes for HP 500
      integer                      :: No_Bytes      !!
      character (len=1), parameter :: Ending = "W"  !!

      character (len=1), parameter :: Empty = char (0)
      character (len=128)          :: Line_Raster
      character (len=135)          :: Line        !! 7 control characters for the HP 500, and 128 image
      integer                      :: Rec_No

      open (unit=9, file="GRAPHICS.OUT", access="direct", action="write",  &
          form="unformatted", status="new", recl=135)

      Line(1:7) = Resolution                      !! The first record contains control codes for the
      Line(8:12) = Line_Spacing                   !! HP DeskJet 500 to start up in graphics mode.

      Line(13:15) = Start                         !! Codes to write out an empty line, to fill up Line.
      write ( unit=Line(16:18), fmt="(i3.3)" ) len(Line) - 19
      Line(19:19) = Ending
      Line(20: ) = repeat (Empty, len(Line)-19)

      Rec_No = 1
      write (unit=9, rec=Rec_No) Line

      Row = 1
!!    do Row = 1, ubound(Graph, 1)        !! A loop to ignore initial empty pixel rows.
!!       if (Graph(Row) /= repeat(Empty, len(Graph(1)) ) ) then
!!          exit
!!       end  if
!!    end  do

      do Row = Row, ubound(Graph, 1)              !! For each pixel row in the image.
         Line_Raster = Graph(Row)
         !! Truncate zero bytes on the right.
         do No_Bytes = len (Line_Raster), 2, -1
            if (Line_Raster(No_Bytes:No_Bytes) /= Empty) then
               exit
            end  if
         end  do
         Line(1:3) = Start
         write ( unit=Line(4:6), fmt="(i3.3)" ) No_Bytes
         Line(7:7) = Ending
         Line(8: ) = Line_Raster(1:128)           !! (1: No_Bytes)
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Line
      end  do

      close (unit=9)

   end  subroutine  PRINT_GRAPHICS
   !! This procedure draws a straight line between the previous point (x0, y0) and the current
   !! point (x1, y1).  The current point then becomes the old point, in readiness for the next
   !! call of this subroutine.
   subroutine  DRAW ( )
      use  GLOBAL

      y1 = 200 - y1                               !! Invert the image.
      if (theta /= 0) then
         call DRAW_LINE (x0, y0, x1, y1)
      end  if
      x0 = x1
      y0 = y1
      return
   end  subroutine  DRAW

end  module  GROUP

!! This program draws a Scaphites seashell.
!! INCLUDE  "SALFORD.INC"              !! Salford graphics interface.  (Refer to Appendix D.)
program  SCAPHITES
!! use  SALFORD                        !! Supply colors and Interfaces.
   use  GLOBAL
   use  GROUP
   implicit  none
   real                            :: c, s, e, e2
   real                            :: c4, s4,              &
                                      a1, b, Angle_of_view, L,   &
                                      x, y, z, e5
   real                            :: p1, p2, p3,          &
                                      q1, q2, q3,          &
                                      r1, r2, r3,          &
                                      L1, L2, L3,          &
                                      m1, m2, m3,          &
                                      n1, n2, n3,          &
                                      u, v, w,             &
                                      u1,u2, u3,           &
                                      v1, v2, v3,          &
                                      w1, w2, w3

   real                            :: phi
   integer                         :: Viewing_angle
   integer                         :: J           !! Salford.

   real, parameter                 :: pi = 3.14180

   Graph = repeat (char(0), len(Graph) )          !! Initialize the image to zero bits.

   print *, "What aspect would you like to view?"
   print *, "For side view, angle = 0 degrees,"
   print *, "For the end view, angle = 90 degrees,"
   print *, "For the opposite end view, angle = -90 degrees."
   print *, "Please type the angle of view:"
   read *, Viewing_angle
   print *, "The viewing angle is ", Viewing_angle, " degrees."

!! CALL VGA@

   Angle_of_view = Viewing_angle * 3.14159 /180   !! Convert degrees to radians.

   b = 0.45
         a1 = 0.5
               L = -0.7

   phi = -10
   do
      if (phi > -1.88) then
         exit
      end  if
      e  = exp (a1*phi)
      e2 = exp (-phi*(a1+L))
      c  = cos(phi)
      s  = sin(phi)

      p1 = 0.5*(e + e2)*c
      p2 = 0.5*(e - e2)*s
      p3 = 0

      q1 = ((a1*e - (a1+L)*e2)*c - (e + e2)*s)/2
      L1 = q1
      q2 = ((a1*e + (a1+L)*e2)*s + (e - e2)*c)/2
      L2 = q2
      L3 = 0
      q3 = 0

      r1 = ((a1**2-1)*e + ((a1 + L)**2-1)*e2)*c/2 - (a1*e - (a1+L)*e2)*s
      r2 = (a1*e + (a1+L)*e2)*c + ((a1**2-1)*e - ((a1+L)**2-1)*e2)*s/2
      r3 = 0

      n1 = q2*r3 - q3*r2
      n2 = q3*r1 - q1*r3
      n3 = q1*r2 - q2*r1

      m1 = n2*L3 - n3*L2
      m2 = n3*L1 - n1*L3
      m3 = n1*L2 - n2*L1

      u = sqrt (L1**2 + L2**2 + L3**2)
      v = sqrt (m1**2 + m2**2 + m3**2)
      w = sqrt (n1**2 + n2**2 + n3**2)

      u1 = L1/u
            u2 = L2/u
                  u3 = L3/u
      v1 = m1/v
            v2 = m2/v
                  v3 = m3/v
      w1 = n1/w
            w2 = n2/w
                  w3 = n3/w

      theta = 0
      do
         if (theta > 2*pi) then
            exit
         end  if
         c4 = cos(theta)
         s4 = sin(theta)
         e5 = (e + e2)/2
         x = p1 + e5*b*(-c4*v1 + s4*w1)
         y = p2 + e5*b*(-c4*v2 + s4*w2)
         z = p3 + e5*b*(-c4*v3 + s4*w3)
         x = x*cos(Angle_of_view) - z*sin(Angle_of_view)
         z = x*sin(Angle_of_view) + z*cos(Angle_of_view)
         x1 = (x + 0.65)*600                      !! Scale to fit the screen -- 600x180 pixels.
         y1 = (y + 0.65)*180
         call DRAW ( )
         theta = theta + pi/10.01
      end  do
      phi = phi + pi/30.01
   end  do

!! call GET_KEY@(J)                 !! For Salford Fortran 90.  Pause, to admire the display;
!!                                  !! press any key to continue.
!! call TEXT_MODE@                  !! For Salford Fortran 90.
   print *, "Completed generation of seashell"
   print *, "Now writing binary file"
   call PRINT_GRAPHICS ( )

end  program  SCAPHITES

!! Figure 9.2  Program to draw a Scaphites seashell
