module  GLOBAL
   !! Graph holds the artwork, which is 1024 pixels wide and 1024 pixels deep.
   !! Each element of Graph holds one pixel row of the finished artwork.
   !! A one-bit inserted in Graph will be displayed as a black dot.
   character (len=128), dimension (1024), public   :: Graph
   !! Each bit is initialized to zero, representing an untarnished (white) image.
end  module  GLOBAL

module GROUP
   implicit none
   public                          :: DRAW_LINE, PRINT_GRAPHICS

contains

   !! This subroutine draws a straight line from the starting point (Xs,Ys) to the
   !! finishing point (Xf,Yf).  It also performs scaling of the artwork according to
   !! the value of <Scaling>, and gives it an <Offset>, for centering on the page.
   subroutine  DRAW_LINE (Xs, Ys, Xf, Yf)
      !! INCOMING: (Xs, Ys) are the co-ordinates of the starting point;
      !!           (Xf, Yf) are the co-ordinates of the finish point.
      use  GLOBAL

      real, intent (in)            :: Xs, Ys,        &
                                      Xf, Yf

      integer, parameter           :: Scaling = 100
      integer, parameter           :: Offset = 512
      real                         :: Row, Col
      integer                      :: X_Increment,   &
                                      Y_Increment
      integer                      :: Byte_Address, Bit_Address, Byte

      if (Ys > Yf) then                           !! "Y_Increment" is 1 Pixel in the Y direction.
         Y_Increment = -1
      else
         Y_Increment = 1
      end  if
      if (Xs > Xf) then                           !! "X_Increment" is 1 Pixel in the X direction.
         X_Increment = -1
      else
         X_Increment = 1
      end  if

      if (int(Xf*Scaling) == int(Xs*Scaling) ) then !! Only the Y value changes.
         Byte_Address = ishft (int(Xs*Scaling)+Offset, -3)
         Bit_Address = iand ( not ( int(Xs*Scaling)+Offset), 7)

         Row = Ys*Scaling
         do
            if (Row > Yf*Scaling) then
               exit
            end  if
            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph(int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Row = Row + Y_Increment
         end  do
      else if (abs(Xs-Xf) >= abs(Ys-Yf) ) then    !! The slope of the line is <= 45 degrees.
                                                  !! The X and Y values change, or just X changes.

         Col = Xs*Scaling
         do
            if (Col > Xf*Scaling) then
               exit
            end  if
            Row = (Yf-Ys)*(Col-Xs*Scaling)/(Xf - Xs) + Ys*Scaling

            Byte_Address = ishft (int(Col)+Offset, -3)
            Bit_Address = iand ( not ( int(Col)+Offset), 7)

            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph( int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Col = Col + X_Increment
         end  do
      else                                        !! The slope of the line is > 45 degrees.
         Row = Ys*Scaling
         do
            if (Row > Yf*Scaling) then
               exit
            end  if
            Col = (Xf-Xs)*(Row-Ys*Scaling)/(Yf - Ys) + Xs*Scaling

            Byte_Address = ishft (int(Col)+Offset, -3)
            Bit_Address = iand ( not ( int(Col)+Offset), 7)
            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph( int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Row = Row + Y_Increment
         end  do
      end  if

   end  subroutine  DRAW_LINE

   !! This subroutine prints the finished artwork on an HP DeskJet 500 series inkjet printer.
   subroutine  PRINT_GRAPHICS ( )
      use  GLOBAL

      integer                      :: Row
      character (len=1), parameter :: ESC = char(27) !!{XE "LEN: in CHARACTER declaration"}
      character (len=7), parameter :: Resolution = ESC // "*t300R"  !! Control codes for the HP 500
      character (len=5), parameter :: Line_Spacing = ESC // "*b1Y"
      character (len=3), parameter :: Start = ESC // "*b" !! Defines a transmission of bytes for HP 500
      integer                      :: No_Bytes    !!
      character (len=1), parameter :: Ending = "W" !!

      character (len=1), parameter :: Empty = char (0)
      character (len=128)          :: Line_Raster
      character (len=135)          :: Line        !! 7 control characters for the HP 500, and 128 image
      integer                      :: Rec_No

      open (unit=9, file="GRAPHICS.OUT", access="direct", action="write",  &
          form="unformatted", status="new", recl=135)

      Line(1:7) = Resolution                      !! The first record contains control codes for the
      Line(8:12) = Line_Spacing                   !! HP DeskJet 500 to start up in graphics mode.

      Line(13:15) = Start                         !! Codes to write out an empty line, to fill up Line.
      write ( unit=Line(16:18), fmt="(i3.3)" ) len(Line) - 19
      Line(19:19) = Ending
      Line(20: ) = repeat (Empty, len(Line)-19)

      Rec_No = 1
      write (unit=9, rec=Rec_No) Line

      Row = 1
!!    do Row = 1, ubound(Graph, 1)        !! A loop to ignore initial empty pixel rows.
!!       if (Graph(Row) /= repeat(Empty, len(Graph(1)) ) ) then
!!          exit
!!       end  if
!!    end  do

      do Row = Row, ubound(Graph, 1)              !! For each pixel row in the image.
         Line_Raster = Graph(Row)
         !! Truncate zero bytes on the right.
         do No_Bytes = len (Line_Raster), 2, -1
            if (Line_Raster(No_Bytes:No_Bytes) /= Empty) then
               exit
            end  if
         end  do
         Line(1:3) = Start
         write ( unit=Line(4:6), fmt="(i3.3)" ) No_Bytes
         Line(7:7) = Ending
         Line(8: ) = Line_Raster(1:128)           !! (1: No_Bytes)
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Line
      end  do

      close (unit=9)

   end  subroutine  PRINT_GRAPHICS

end  module  GROUP

!! INCLUDE  "SALFORD.INC"              !! Salford graphics interface.  (Refer to Appendix D.)

!! This procedure generates the graphics display XYLEM II.
program  XYLEM_II
!! USE  SALFORD                  !! Supply colors and Interfaces.
   use  GLOBAL
   use  GROUP
   implicit  none
   real                            :: A, T
   real                            :: B, D
   real                            :: Theta, R
   real                            :: X, Y
   integer                         :: J
   real                            :: X_old, Y_old
   integer, parameter              :: X_Scale = 30, Y_Scale = 30, X_offset = 300, Y_offset = 180

!! CALL EGA@                     !! For Salford Fortran 90.

   Graph = repeat (char(0), len(Graph) )          !! Initialize the image to zero bits.

   A = 1.0
   do
      if (A > 3.0) then
         exit
      end  if
      X_old = 0
            Y_old = 0

      T = 0
      do
         if (T > 6.28) then
            exit
         end  if
         R = A*sin(4*T)
         X = R*cos(T)
               Y = R*sin(T)
         call DRAW_LINE (X_old, Y_old, X, Y)
!!       Salford Fortran 90 Graphics
!!       CALL DRAW_LINE@ (INT(X_old*X_Scale+X_offset), INT(Y_old*Y_Scale+Y_offset), &
!!          INT(X*X_Scale+X_offset), INT(Y*Y_Scale+Y_offset), Color_Red)
         X_old = X
               Y_old = Y
         T = T + 0.02
      end  do
      A = A + 0.2
   end  do

   do J = 1, 8
print *, J
      A = 1.0
      do
         if (A > 3.0) then
            exit
         end  if
         B = atan (A/sqrt(25-A**2))/4
         select case (J)
            case (1)
                  Theta = B
                                 D = 0.78540 - B
            case (2)
                  Theta = 0.78540 + B
                                 D = 1.57080 - B
            case (3)
                  Theta = 1.57080 + B
                                 D = 2.35619 - B
            case (4)
                  Theta = 2.35619 + B
                                 D = 3.14159 - B
            case (5)
                  Theta = 3.14159 + B
                                 D = 3.92699 - B
            case (6)
                  Theta = 3.92699 + B
                                 D = 4.71239 - B
            case (7)
                  Theta = 4.71239 + B
                                 D = 5.49779 - B
            case (8)
                  Theta = 5.49779 + B
                                 D = 6.28319 - B
         end  select

         X_old = 5*cos(Theta)
                        Y_old = 5*sin(Theta)

         T = Theta
         do
            if (T > D) then
               exit
            end  if
            R = A/sin(4*T)
            X = R*cos(T)
                        Y = R*sin(T)
            call DRAW_LINE (X_old, Y_old, X, Y)
!!          Salford Fortran 90 Graphics
!!          CALL DRAW_LINE@ (INT(X_old*X_Scale+X_offset), INT(Y_old*Y_Scale+Y_offset), &
!!             INT(X*X_Scale+X_offset), INT(Y*Y_Scale+Y_offset), Color_Yellow)
            X_old = X
                  Y_old = Y
            T = T + 0.02
         end  do
         A = A + 0.02
      end  do
   end  do

!! CALL GET_KEY@(J)                 !! For Salford Fortran 90.  Pause, to admire the display;
!!                                  !! press any key to continue.
!! CALL TEXT_MODE@                  !! For Salford Fortran 90.

   print *, "Completed generation of XYLEM II."
   print *, "Writing graphics file"

   call PRINT_GRAPHICS ( )

end  program  XYLEM_II

!! Figure 9.4  Program to generate and print Xylem II
