module  GLOBAL

   !! Graph holds the artwork, which is 1024 pixels wide and 1024 pixels deep.
   !! Each element of Graph holds one pixel row of the finished artwork.
   !! A one-bit inserted in Graph will be displayed as a black dot.
   character (len=128), dimension (1024), public   :: Graph
   !! Each bit is initialized to zero, representing an untarnished (white) image.

   real, public                    :: x1, y1

   real, public                    :: x0, z0

end  module  GLOBAL

module GROUP
   implicit none
   public                          :: TILT, PLOT, DRAW_LINE

contains

   !! This subroutine tilts the chalice through an angle BETA about the X axis,
   !! that is, it tilts the chalice away from the viewer when beta is negative,
   !! and tilts it towards the viewer when beta is positive.
   subroutine  TILT (y, z, beta)

      real, intent (in out)        :: y, z
      real, intent (in)            :: beta

      real                         :: y1, s, c

      if (beta == 0) then
         return
      end  if
      s = sin(beta)
            c = cos(beta)
      y1 = y*c - z*s
      z   = y*s + z*c
      y   = y1
   end  subroutine  TILT

   !! This procedure draws a straight line between the previous point (x0, z0) and the current
   !! point (x, z).  The current point then becomes the old point, in readiness for the next
   !! call of this subroutine.
   subroutine  PLOT (x, y, z, theta)
      use  GLOBAL

      real, intent (in out)        :: x, y, z
      integer, intent (in)         :: theta

      real, parameter              :: Rmin=-50, Rmax=50,   &
                                      Zmin=-15, Zmax=100

      x = (x - Rmin)/(Rmax-Rmin)*600              !! Scale the diagram to fit the screen.
      y = (y - Rmin)/(Rmax-Rmin)*600
      z = (z - Zmin)/(Zmax-Zmin)*440
      z = 180 - z                                 !! Invert the image.
      if (theta /= 1) then                        !! Except for the first point . . .
         call DRAW_LINE (x0, z0, x, z)
!!       CALL DRAW_LINE@ (INT(x0), 300+INT(z0), INT(x), 300+INT(z))  !! For Salford graphics.
      end  if

   end  subroutine  PLOT

!!  include "DRAWLINE.INC" 

   !! This subroutine draws a straight line from the starting point (Xs,Ys) to the
   !! finishing point (Xf,Yf).  It also performs scaling of the artwork according to
   !! the value of <Scaling>, and gives it an <Offset>, for centering on the page.
   subroutine  DRAW_LINE (Xs, Ys, Xf, Yf)
      !! INCOMING: (Xs, Ys) are the co-ordinates of the starting point;
      !!           (Xf, Yf) are the co-ordinates of the finish point.
      use  GLOBAL

      real, intent (in)            :: Xs, Ys,        &
                                      Xf, Yf

      integer, parameter           :: Scaling = 100
      integer, parameter           :: Offset = 512
      real                         :: Row, Col
      integer                      :: X_Increment,   &
                                      Y_Increment
      integer                      :: Byte_Address, Bit_Address, Byte

      if (Ys > Yf) then                           !! "Y_Increment" is 1 Pixel in the Y direction.
         Y_Increment = -1
      else
         Y_Increment = 1
      end  if
      if (Xs > Xf) then                           !! "X_Increment" is 1 Pixel in the X direction.
         X_Increment = -1
      else
         X_Increment = 1
      end  if

      if (int(Xf*Scaling) == int(Xs*Scaling) ) then !! Only the Y value changes.
         Byte_Address = ishft (int(Xs*Scaling)+Offset, -3)
         Bit_Address = iand ( not ( int(Xs*Scaling)+Offset), 7)

         Row = Ys*Scaling
         do
            if (Row > Yf*Scaling) then
               exit
            end  if
            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph(int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Row = Row + Y_Increment
         end  do
      else if (abs(Xs-Xf) >= abs(Ys-Yf) ) then    !! The slope of the line is <= 45 degrees.
                                                  !! The X and Y values change, or just X changes.

         Col = Xs*Scaling
         do
            if (Col > Xf*Scaling) then
               exit
            end  if
            Row = (Yf-Ys)*(Col-Xs*Scaling)/(Xf - Xs) + Ys*Scaling

            Byte_Address = ishft (int(Col)+Offset, -3)
            Bit_Address = iand ( not ( int(Col)+Offset), 7)

            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph( int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Col = Col + X_Increment
         end  do
      else                                        !! The slope of the line is > 45 degrees.
         Row = Ys*Scaling
         do
            if (Row > Yf*Scaling) then
               exit
            end  if
            Col = (Xf-Xs)*(Row-Ys*Scaling)/(Yf - Ys) + Xs*Scaling

            Byte_Address = ishft (int(Col)+Offset, -3)
            Bit_Address = iand ( not ( int(Col)+Offset), 7)
            Byte = ichar(Graph(int(Row)+Offset)(Byte_Address : Byte_Address) )
            Byte = ibset (Byte, Bit_Address)
            Graph( int(Row)+Offset)(Byte_Address : Byte_Address) = char(Byte)
            Row = Row + Y_Increment
         end  do
      end  if

   end  subroutine  DRAW_LINE

end  module  GROUP

!! INCLUDE  "SALFORD.INC"              !! Salford graphics interface.  (Refer to Appendix D.)
!! This program displays a line drawing of Uccello's chalice.
program  CHALICE
!! use  SALFORD                        !! Supply colors and Interfaces.
   use  GLOBAL
   use  GROUP
   implicit  none
   real                            :: beta, phi
   integer                         :: theta, Angle_of_tilt
   real                            :: x, y, zz
   !! Establish the profile of the chalice.  Co-ordinates are measured from the base of the
   !! chalice.  Each R(i) and z(i) form a pair of co-ordinates, where R(i) is the radius of the
   !! chalice at point z(i).
   real, dimension (54), parameter  :: R = (/                           &
                           37.0, 37.0, 32.0, 30.0, 25.0, 26.0, 25.0,    &
                           25.0, 23.0, 28.0, 28.0, 23.0, 18.0, 16.0,    &
                           14.5, 13.0, 12.0, 11.0, 10.5, 10.0,  9.7,    &
                            9.3,  9.0,  8.8,  8.8, 20.0, 19.5, 18.5,    &
                           17.5, 16.0, 14.0, 12.0, 10.0, 11.0, 10.0,    &
                           13.0, 16.0, 19.0, 22.5, 25.0, 27.0, 28.5,    &
                           30.0, 31.0, 33.0, 33.0, 30.0, 32.0, 32.0,    &
                           37.0, 40.0, 40.0, 37.0, 32.0 /)
   real, dimension (54), parameter  :: z = (/                           &
                            1.0,  2.0,  3.0,  6.0,  7.0,  8.0,  9.0,    &
                           11.0, 14.0, 16.0, 21.0, 23.0, 21.0, 23.0,    &
                           25.0, 27.0, 29.0, 31.0, 33.0, 35.0, 37.0,    &
                           39.0, 41.0, 43.0, 46.0, 46.0, 48.0, 50.0,    &
                           52.0, 54.0, 55.5, 57.0, 58.0, 60.0, 62.0,    &
                           63.0, 64.0, 65.0, 67.0, 69.0, 71.0, 73.0,    &
                           75.0, 77.0, 77.0, 78.0, 80.0, 83.0, 85.0,    &
                           85.0, 88.0, 93.0, 96.0, 96.0 /)

   real, parameter                 :: pi = 3.14180

   x0 = 0
   z0 = 0

   print *, "This program displays a chalice."
   print *, "Please give the angle of tilt, in degrees."
   print *, "Typing 30 will tilt the chalice towards the viewer by 30 degrees."
   print *, "Typing -30 will tilt the chalice away from the viewer by 30 degrees."
   print *, "To show the chalice upright, type 0."
   read *, Angle_of_tilt
   print *, "You requested an angle of tilt of ", Angle_of_tilt, " degrees."

!! CALL VGA@

   beta = Angle_of_tilt * 3.14159 / 180           !! Convert degrees to radians.

   phi = 0
   do                                             !! For each phi, draw one vertical contour line from
      if (phi < -pi) then                         !! bottom to top.  Each line (at 6 degree intervals)
         exit                                     !! defines the shape of the chalice.
      end  if
      do theta = 1, ubound(R, 1)                  !! Work through the 54 data points [z, R], in a
                                                  !! vertical plane to draw a straight line from each
                                                  !! point to the next.
         x = R(theta) * cos(phi)
         y = R(theta) * sin(phi)
         zz = z(theta)
         call TILT (y, zz, beta)
         call PLOT (x, y, zz, theta)
         x0 = x                                   !! Save x, zz for the next CALL PLOT.
         z0 = zz
      end  do
      phi = phi - pi/30
   end  do

!! CALL GET_KEY@(J)                 !! For Salford Fortran 90.  Pause, to admire the display;
!!                                  !! press any key to continue.
!! CALL TEXT_MODE@                  !! For Salford Fortran 90.

end  program  CHALICE

!! Figure 9.6  Program to draw a chalice
