!! These procedures (with accompanying modules) write out a TIFF file containing
!! a black & white graphic image.

module  GLOBAL

   character (len=1), dimension (0:1023, 0:1023), public :: Pixel_Image

end  module  GLOBAL

module  GLOBAL2
   character (len=128), public     :: Record
   integer, save, public           :: Rec_No=0, L=0
end  module  GLOBAL2

module  GROUP

   implicit none
   public :: WRITE_BYTE_INTO_BUFFER, WRITE_WORD, MAKE_TAG, MAKE_TIFF_FILE, &
             MAKE_STRING

contains

   !! This subroutine writes one byte into the output buffer "Record".
   !! When the buffer is full, it is written to the graphics file.
   subroutine WRITE_BYTE_INTO_BUFFER (Byte)
      use  GLOBAL2

      character (len=*), intent (in)   :: Byte

      L = L + 1
      if (L > len(Record) ) then
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Record
         L = 1
      end  if
      Record(L:L) = Byte

   end  subroutine  WRITE_BYTE_INTO_BUFFER

   !! This subroutine breaks an integer into 4 bytes (or 2 bytes), and sends them
   !! one-at-a-time to WRITE_BYTE_INTO_BUFFER, least significant byte first.
   subroutine  WRITE_WORD (Word, Length)
      !! INCOMING: Word   = an integer to be written into the buffer as 4 or 2 bytes;
      !!           Length = the number of bytes to be written.  This argument is optional.  If
      !!                    absent, 4 bytes are written; if present, <Length> bytes are written.

      integer, intent (in)           :: Word
      integer, intent (in), optional :: Length

      integer                        :: L_Word      !! A local version of dummy argument "Word".
      integer                        :: L_Length
      integer                        :: J
      character (len=1)              :: Ch

      if (present (Length)) then
         L_Length = Length
      else
         L_Length = 4
      end  if

      L_Word = Word
      do J = 1, L_Length                          !! Write out 4 or <Length> bytes of Word,
         Ch = char (iand (L_Word, 255) )          !! least significant byte first.
         call WRITE_BYTE_INTO_BUFFER (Ch)
         L_Word = ishft (L_Word, -8)
      end do

   end  subroutine  WRITE_WORD

   !! This subroutine prepares to write out the four components of an IFD:
   !! namely, the Tag ID (2 bytes);
   !! the Data Type (2 bytes);
   !! the Count [the number of data items] (4 bytes); and
   !! the Offset [the address of the data item] (4 bytes).
   subroutine  MAKE_TAG (Number, Tag_ID, Data_Type, Counter, Offset)

      integer, intent (in)         :: Number, Tag_ID, Data_Type, Counter, Offset
      !! <Number> is not used by this subroutine -- it is for cosmetic purposes only.

      call WRITE_WORD (Tag_ID, 2)                 !! Write out 2 bytes of Tag_ID,
                                                  !! least significant byte first.
      call WRITE_WORD (Data_Type, 2)              !! Write out 2 bytes of Data Type,
                                                  !! least-significant byte first.
      call WRITE_WORD (Counter)                   !! Write out 4 bytes of Counter,
                                                  !! least-significant byte first.
      call WRITE_WORD (Offset)                    !! Write out 4 bytes of Offset,
                                                  !! least-significant byte first.
   end  subroutine  MAKE_TAG

   !! This subroutine writes out a TIFF file containing a black & white graphic image.
   subroutine  MAKE_TIFF_FILE ( )
      use  GLOBAL
      use  GLOBAL2

      character (len=128)          :: Line_Raster

      integer                      :: I, Row

      open (unit=9, file="GRAPHICS.TIF", access="direct", action="write",  &
          form="unformatted", status="new", recl=128)

      !! Create the Image File Header (IFH) (8 bytes).
      !! Contains: Byte Ordering:   Intel byte ordering (2 bytes of "II");
      !!       Identification:   2 bytes, integer value 42;
      !!       Address of first IFD (4 bytes).
                                                  !! Header.
      call WRITE_BYTE_INTO_BUFFER ("I")           !! The header for Intel format (Little Endian).
      call WRITE_BYTE_INTO_BUFFER ("I")
      call WRITE_WORD (42, 2)                     !! Version Number, always 42.
      call WRITE_WORD ( 10 )                      !! The address of the first IFD, which is
                                                  !! four bytes altogether.
      call WRITE_WORD ( 0, 2 )                    !! Two bytes unused.

      !! Create the Image File Directory (IFD) Header:
      !! 2 bytes:  Number of entries in the IFD.
      !! 12 bytes: For each Tag in the IFD: Tag_ID (2 bytes),
      !!                                    Data Type (2 bytes)
      !!                                    Number of items (4 bytes)
      !!                                    Address of item (4 bytes)
      !!                                    (If the No. of items is 1, and the data type is 4 or less,
      !!                                    this field contains the actual data.)
      !! 4 bytes: Trailer: A link to other IFDs, or zero if there is no more.

      call WRITE_BYTE_INTO_BUFFER (char(14))      !! The number of entries in the IFD,
      call WRITE_BYTE_INTO_BUFFER (char( 0))      !! which starts at address 10.

      !! The following should be in increasing order of TAG ID.
      call MAKE_TAG (1, 254, 4, 1, 1)             !! Tag for new sub-file type.
      call MAKE_TAG (2, 256, 3, 1, 1024)          !! Tag for LINE WIDTH in pixels.
      call MAKE_TAG (3, 257, 3, 1, 1024)          !! Tag for IMAGE DEPTH in pixels.
      call MAKE_TAG (4, 258, 3, 1, 1)             !! Tag for Number of Bits per Sample.
      call MAKE_TAG (5, 259, 3, 1, 1)             !! Tag for Compression: 1=uncompressed.
      call MAKE_TAG (6, 262, 3, 1, 0)             !! Photometric Interpretation:
                                                  !! 0 means render pixel 0 as white, pixel 1 as black
                                                  !! 1 means render pixel 0 as black, pixel 1 as white
                                                  !! 3 means that the image is color.
      call MAKE_TAG (7, 273, 4, 1024, 256)        !! Strip Offsets: This tag points to a table of
                                                  !! addresses.  The table contains 1024 entries, one
                                                  !! for each row of the image (and each entry
                                                  !! occupies 4 bytes).
      call MAKE_TAG (8, 277, 3, 1, 1)             !! Tag for Samples per pixel.
      call MAKE_TAG (9, 278, 3, 1, 1)             !! Tag for Rows per strip.
      call MAKE_TAG (10, 279, 3, 1, 128)          !! Tag for Strip byte count.
      call MAKE_TAG (11, 282, 5, 1, 184)          !! Tag for address of X resolution.
      call MAKE_TAG (12, 283, 5, 1, 192)          !! Tag for address of Y resolution.
      call MAKE_TAG (13, 284, 3, 1, 1)            !! Tag for Planar configuration.
      call MAKE_TAG (14, 296, 3, 1, 3)            !! Tag for Resolution.
                                                  !! 1 = undefined, 2 = cm, 3 = inches.

      call WRITE_WORD (0)                         !! 4 bytes of zeros to terminate the IFD chain.

      call WRITE_WORD (300)                       !! The X resolution.
      call WRITE_WORD (1)
      call WRITE_WORD (300)                       !! The Y resolution.
      call WRITE_WORD (1)

      do I = L+1, 128                             !! Write the remaining (short) record.
         call  WRITE_BYTE_INTO_BUFFER (char(0))   !! Zero fill.
      end  do

      !! SECTION TO CREATE THE StripAddressArray, CONTAINING 1024 ADDRESSES,
      !! EACH BEING A POINTER TO ONE PIXEL LINE OF THE IMAGE.
      do Row = 0, 1023
         call WRITE_WORD ( 4352 + Row*128 )
      end  do
      call WRITE_BYTE_INTO_BUFFER ( char(0) )     !! To force out the last record.

      !! SECTION TO WRITE OUT THE PIXEL IMAGES.
      do Row = 0, 1023
         Line_Raster = MAKE_STRING(Pixel_Image(Row, : ))
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Line_Raster
      end  do

      print *, "TIFF file created."

   end  subroutine  MAKE_TIFF_FILE

   !! This function packs 1024 pixels from one row of the pixel image
   !! (one bit per pixel) into a 128-byte raster line.
   function  MAKE_STRING (Pixel_Line) result (String_Result)

      character (len=*), dimension (0: ), intent (in) :: Pixel_Line
      character (len=128)          :: String_Result

      integer                      :: J, K, Byte, Col

      J = 1
      K = 1
      Byte = 0
      do Col = 0, 1023
         Byte = Byte*2                            !! Shift the bits to the left.
         if (Pixel_Line(Col) == "1") then
            Byte = Byte + 1                       !! Convert a character into a bit.
         end  if
         if (J >= 8) then                         !! When we have 8 pixels,
            String_Result(K:K) = char(Byte)       !! store them in a character.
            J = 1
            K = K + 1
            Byte = 0
         else
            J = J + 1
         end  if
      end  do

   end  function  MAKE_STRING

end  module  GROUP

!! Figure 9.7  Procedure to produce a black-and-white picture file in TIFF file format

program test

end  program test
