module  GLOBAL

   character (len=1), dimension (0:1023, 0:1023), public :: Pixel_Image

end  module  GLOBAL

!! This module defines intensities for use with the TIFF color palette.
module  INTENSITIES
   implicit  none
   character (len=2), parameter, public   :: &
      Full = char(255) // char(255),         &
      Tqtr = char(204) // char(0),           &
      Half = char(127) // char(255),         &
      Qtr  = char(128) // char(0),           &
      Nil  = char(0)   // char(0),           &
      Special = char(192) // char(0)              !! '110 000 000 000 000'B
end  module  INTENSITIES

module  Directory_Type

   type, public                    :: Image_File_Directory         !! An Image File Directory.
      integer                      :: Tag_ID
      integer                      :: Data_Type
      integer                      :: Counter
      integer                      :: Offset
   end  type  Image_File_Directory

end  module  Directory_Type

module GROUP
   implicit none
   public                          :: MAKE_TIFF_FILE
   private                         :: WRITE_BUFFER, MAKE_TAG

contains

   !! This subroutine is used by MAKE_TIFF_FILE to check whether the buffer
   !! is full.  When the buffer is full, it is written to the graphics file.
   subroutine WRITE_BUFFER (Rec, Rec_No, I)
      integer, dimension ( : ), intent (in)  :: Rec
      integer, intent (in out)         :: Rec_No, I

      I = I + 1
      if (I > 32) then
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Rec
         I = 1
      end  if
   end  subroutine  WRITE_BUFFER

   !! This subroutine makes a TIFF tag at position "P" in IFD.
   subroutine  MAKE_TAG (P, IFD, TID, DT, Cnt, Off)
      use  Directory_Type
      type (Image_File_Directory), dimension (0: ), intent (in out)  :: IFD
      integer, intent (in)         :: P, TID, DT, Cnt, Off

      IFD(P)%Tag_ID = TID
               IFD(P)%Data_Type = DT
                        IFD(P)%Counter = Cnt
                              IFD(P)%Offset = Off

   end  subroutine  MAKE_TAG

    !! This procedure writes out a TIFF file containing a color image.
    subroutine  MAKE_TIFF_FILE ( )
      use  GLOBAL
      use  INTENSITIES
      use  Directory_Type

      type (Image_File_Directory), dimension (0:17)   :: IFD

      integer, dimension (32)       :: Rec
      character (len=2), dimension (0:63) :: Colors
      character (len=128)           :: Line
      character (len=512)           :: Pixel_Line

      character (len=2), dimension (0:47), parameter  :: Color_Palette = (/   &
               Full, Full, Full,      &          !!  0 = White.
               Full, Nil,  Nil,       &          !!  1 = Red.
               Full, Full, Nil,       &          !!  2 = Yellow.
               Full, Nil,  Full,      &          !!  3 = Purple.
               Tqtr, Full, Tqtr,      &          !!  4 = Light Green
               Nil,  Full, Full,      &          !!  5 = Light blue.
               Nil,  Nil,  Full,      &          !!  6 = Dark blue.
               Full, Tqtr, Nil,       &          !!  7 = Orange.
               Nil,  Full, Nil,       &          !!  8 = Green.
               Full, Half, Full,      &          !!  9 = Pale purple.
               Full, Full, Half,      &          !! 10 = Lemon.
               Nil,  Half, Full,      &          !! 11 = Medium blue.
               Full, Special, Nil,    &          !! 12 = Light brown.
               Nil,  Nil,  Nil,       &          !! 13 = Black.
               Nil,  Nil,  Nil,       &          !! 14 = Black.
               Nil,  Nil,  Nil /)                !! 15 = Black.

      character (len=2), parameter     :: Empty = char(0) // char(0)
      integer                      :: I, J, K, N, Row, Col
      integer                      :: Rec_No

      open (unit=9, file="TIFF_FILE.TIF", access="direct", action="write",    &
           form="unformatted", status="new", recl=128)

      !! Create the Image File Header (IFH) (8 bytes).
      !! Contains: Byte Ordering:  Intel byte ordering (2 bytes of "II");
      !!           Identification: 2 bytes, integer value 42;
      !!           Address of the first IFD (4 bytes).
      IFD(0)%Tag_ID = ichar("I")*256+ichar("I")
                        IFD(0)%Data_Type = 42
                              IFD(0)%Counter = 10
                                                  !! Header.

      !! Create the Image File Directory (IFD) Header:
      !! 2 bytes:  Number of entries in the IFD.
      !! 12 bytes: For each Tag in the IFD: Tag_ID (2 bytes),
      !!                                    Data Type (2 bytes)
      !!                                    Number of items (4 bytes)
      !!                                    Address of item (4 bytes)
      !!                                    (If the No. of items is 1,
      !!                                    and the data type is 4 or less,
      !!                                    this field contains the actual data.)
      !! 4 bytes: Trailer: A link to other IFDs, or zero if there is no more.
      IFD(0)%Offset = 15*65536                    !! No. of IFD tags*65536.

      !! Following should be in increasing order of TAG ID.
      call MAKE_TAG (1, IFD, 254, 4, 1, 1)        !! Tag for new sub-file type.
      call MAKE_TAG (2, IFD, 256, 3, 1, 1024)     !! Tag for LINE WIDTH in pixels.
      call MAKE_TAG (3, IFD, 257, 3, 1, 1024)     !! Tag for IMAGE DEPTH in pixels.
      call MAKE_TAG (4, IFD, 258, 3, 1, 4)        !! Tag for Number of Bits per Sample.
      call MAKE_TAG (5, IFD, 259, 3, 1, 1)        !! Tag for Compression: 1=uncompressed.
      call MAKE_TAG (6, IFD, 262, 3, 1, 3)        !! Photometric Interpretation:
                                                  !! 0 means render pixel 0 as white, pixel 1 as black
                                                  !! 1 means render pixel 0 as black, pixel 1 as white
                                                  !! 3 means that the image is color.
      call MAKE_TAG (7, IFD, 273, 4, 1024, 384)   !! Strip Offsets: This tag points to a table of
                                                  !! addresses.  The table contains 1024 entries, one
                                                  !! for each row of the image (and each entry
                                                  !! occupies 4 bytes).
      call MAKE_TAG (8, IFD, 277, 3, 1, 1)        !! Tag for Samples per pixel.
      call MAKE_TAG (9, IFD, 278, 3, 1, 1)        !! Tag for Rows per strip.
      call MAKE_TAG (10, IFD, 279, 3, 1, 128)     !! Tag for Strip byte count.
      call MAKE_TAG (11, IFD, 282, 5, 1, 196)     !! Tag for address of X resolution.
      call MAKE_TAG (12, IFD, 283, 5, 1, 204)     !! Tag for address of Y resolution.
      call MAKE_TAG (13, IFD, 284, 3, 1, 1)       !! Tag for Planar configuration.
      call MAKE_TAG (14, IFD, 296, 3, 1, 3)       !! Tag for Resolution:
                                                  !! 1 = undefined, 2 = cm, 3 = inches.
      call MAKE_TAG (15, IFD, 320, 3, 48, 256)    !! Tag Points to the Color Palette.

      J = 16
      IFD(J)%Tag_ID = 0
               IFD(J)%Data_Type = 0               !! 4 bytes of zeros to terminate the IFD chain.

      !! The next few entries are parameters of the picture.
      IFD(J)%Counter = 300
               IFD(J)%Offset = 1                  !! X resolution.
      J = J + 1
      IFD(J)%Tag_ID = 300
               IFD(J)%Data_Type = 0
                        IFD(J)%Counter = 1        !! Y resolution.
      IFD(J)%Offset = 0                           !! unused area.

      N = J                                       !! the total number of records.

      !! Write the Header, the IFD tags, & the trailer, to the graphic file.
      !! Each IFD tag is 12 bytes (2 x 16-bit integers & 2 x 32-bit integers).
      !! Records are 128 bytes.  32-bit integers are copied from structure
      !! IFD until the buffer REC is full, then they are written out.
      Rec_No = 0
      I = 1
      do J = 0, N
         Rec(I) = IFD(J)%Tag_ID + IFD(J)%Data_Type*65536
         call WRITE_BUFFER (Rec, Rec_No, I)

         Rec(I) = IFD(J)%Counter
         call WRITE_BUFFER (Rec, Rec_No, I)

         Rec(I) = IFD(J)%Offset
         call WRITE_BUFFER (Rec, Rec_No, I)
      end  do

      if (I > 1) then                             !! Write the remaining (short) record.
         Rec(I:32) = 0                            !! Zero fill the remainder of the record.
         Rec_No = Rec_No + 1
         write (unit=9, rec=Rec_No) Rec
      end  if

      !! SECTION TO WRITE OUT THE COLOR PALETTE.
      Colors( 0:15) = Color_Palette(0:45:3)       !! Collect the reds;{XE "subscript triplet"}
      Colors(16:31) = Color_Palette(1:46:3)       !! Collect the greens;{XE "array section"}
      Colors(32:47) = Color_Palette(2:47:3)       !! Collect the blues.
      Colors(48:63) = Empty

      Rec_No = Rec_No + 1
      write (unit=9, rec=Rec_No) Colors

      !! SECTION TO CREATE THE StripAddressArray, CONTAINING 1024 ADDRESSES,
      !! EACH BEING A POINTER TO ONE PIXEL LINE OF THE IMAGE.
      K = 1
      do Row = 0, 1023
         Rec(K) = 4480 + Row*512
         K = K + 1
         if (K > 32) then
            Rec_No = Rec_No + 1
            write (unit=9, rec=Rec_No) Rec
            K = 1
         end  if
      end  do

      !! SECTION TO WRITE OUT THE PIXEL IMAGES
      do Row = 0, 1023
         !! Prepare one scan line.
         do Col = 0, 1023, 2                      !! pack two pixels (4 bits each) into a byte.
            Pixel_Line(Col/2+1:Col/2+1) = char (      &
               ichar(Pixel_Image(Row,Col))*16 + ichar(Pixel_Image(Row,Col+1)) )
         end  do
         !! Write out a line.
         do K = 1, 512, 128                       !! The line is written in four chunks.
            Line = Pixel_Line(K:K+127)
            Rec_No = Rec_No + 1
            write (unit=9, rec=Rec_No) Line
         end  do
      end  do

      print *, "TIFF file created."

   end  subroutine  MAKE_TIFF_FILE

end  module  GROUP

!! Figure 9.9  This procedure writes out a color picture file in TIFF format
